const char config_utils_log_lua[] =
"-- Logger wrapper with a few enhancements.\n"
"--\n"
"-- 1. Encode tables into JSON.\n"
"-- 2. Enable all the messages on TT_CONFIG_DEBUG=1.\n"
"--\n"
"-- Hopefully, the wrapper can be dropped when the following logger\n"
"-- problems will be solved.\n"
"--\n"
"-- * JSONify tables at logging (gh-8611)\n"
"-- * Follow TT_LOG_LEVEL before log.cfg()/box.cfg() (gh-8092)\n"
"--\n"
"-- A solution of the following problems may simplify the code of\n"
"-- this wrapper.\n"
"--\n"
"-- * mylog.cfg() configures the main logger (gh-8610)\n"
"-- * No easy way to determine if a message should go to the log\n"
"--   (gh-8730)\n"
"\n"
"local json_noexc = require('json').new()\n"
"json_noexc.cfg({encode_use_tostring = true})\n"
"\n"
"local logger_name = 'tarantool.config'\n"
"local log = require('log').new(logger_name)\n"
"\n"
"local func2level = {\n"
"    [log.error] = 2,\n"
"    [log.warn] = 3,\n"
"    [log.info] = 5,\n"
"    [log.verbose] = 6,\n"
"    [log.debug] = 7,\n"
"}\n"
"\n"
"local func2prefix = {\n"
"    [log.error] = 'E> ',\n"
"    [log.warn] = 'W> ',\n"
"    [log.info] = 'I> ',\n"
"    [log.verbose] = 'V> ',\n"
"    [log.debug] = 'D> ',\n"
"}\n"
"\n"
"local str2level = {\n"
"    ['fatal'] = 0,\n"
"    ['syserror'] = 1,\n"
"    ['error'] = 2,\n"
"    ['warn'] = 3,\n"
"    ['crit'] = 4,\n"
"    ['info'] = 5,\n"
"    ['verbose'] = 6,\n"
"    ['debug'] = 7,\n"
"}\n"
"\n"
"-- Accept false/true case insensitively.\n"
"--\n"
"-- Accept 0/1 as boolean values.\n"
"--\n"
"-- If anything else is given, just return the default.\n"
"local function boolean_fromenv(raw_value, default)\n"
"    if raw_value == nil or raw_value == '' then\n"
"        return default\n"
"    end\n"
"    if raw_value:lower() == 'false' or raw_value == '0' then\n"
"        return false\n"
"    end\n"
"    if raw_value:lower() == 'true' or raw_value == '1' then\n"
"        return true\n"
"    end\n"
"    return default\n"
"end\n"
"\n"
"local function say_closure(log_f)\n"
"    local prefix = ''\n"
"\n"
"    -- Enable logging of everything if an environment variable is\n"
"    -- set.\n"
"    --\n"
"    -- Useful for debugging.\n"
"    --\n"
"    -- Just setting of...\n"
"    --\n"
"    -- ```\n"
"    -- TT_LOG_MODULES='{\"tarantool.config\": \"debug\"}'\n"
"    -- ```\n"
"    --\n"
"    -- ...is not suitable due to gh-8092: messages before first\n"
"    -- box.cfg() are not shown.\n"
"    --\n"
"    -- Explicit calling of...\n"
"    --\n"
"    -- ```\n"
"    -- log.cfg({modules = {['tarantool.config'] = 'debug'}})\n"
"    -- ```\n"
"    --\n"
"    -- is not suitable as well, because it makes the logger\n"
"    -- already configured and non-dynamic options like\n"
"    -- `box_cfg.log_nonblock` can't be applied at first box.cfg()\n"
"    -- invocation.\n"
"    --\n"
"    -- So just prefix our log messages and use log.info().\n"
"    --\n"
"    -- The prefix represents the original log level of the\n"
"    -- message.\n"
"    if boolean_fromenv(os.getenv('TT_CONFIG_DEBUG'), false) then\n"
"        prefix = func2prefix[log_f]\n"
"        log_f = log.info\n"
"    end\n"
"\n"
"    return function(fmt, ...)\n"
"        -- Skip logging based on the log level before performing\n"
"        -- the encoding into JSON.\n"
"        --\n"
"        -- log.new(<...>).cfg is the main logger configuration,\n"
"        -- the same as log.cfg. Extract the current log level from\n"
"        -- log.cfg.modules if it is set specifically for the config\n"
"        -- module. See gh-8610.\n"
"        local level = log.cfg.modules and\n"
"            log.cfg.modules[logger_name] or\n"
"            log.cfg.level\n"
"        -- The level is either a string or a number. Transform it\n"
"        -- to a number. See gh-8730.\n"
"        level = str2level[level] or level\n"
"        assert(type(level) == 'number')\n"
"        if func2level[log_f] > level then\n"
"            return\n"
"        end\n"
"\n"
"        -- Micro-optimization: don't create a temporary table if\n"
"        -- it is not needed.\n"
"        local argc = select('#', ...)\n"
"        if argc == 0 then\n"
"            log_f(prefix .. fmt)\n"
"            return\n"
"        end\n"
"\n"
"        -- Encode tables into JSON.\n"
"        --\n"
"        -- Ignores presence of __serialize and __tostring in the\n"
"        -- metatatable. It is suitable for the config module needs.\n"
"        local args = {...}\n"
"        for i = 1, argc do\n"
"            if type(args[i]) == 'table' then\n"
"                args[i] = json_noexc.encode(args[i])\n"
"            end\n"
"        end\n"
"\n"
"        -- Pass the result to the logger function.\n"
"        log_f(prefix .. fmt, unpack(args, 1, argc))\n"
"    end\n"
"end\n"
"\n"
"return {\n"
"    error = say_closure(log.error),\n"
"    warn = say_closure(log.warn),\n"
"    info = say_closure(log.info),\n"
"    verbose = say_closure(log.verbose),\n"
"    debug = say_closure(log.debug),\n"
"}\n"
""
;
