const char config_utils_schema_lua[] =
"-- Schema-aware data manipulations.\n"
"--\n"
"-- The following annotations have some meaning for the module\n"
"-- itself (affects work of the schema methods):\n"
"--\n"
"-- * allowed_values\n"
"-- * validate\n"
"-- * default\n"
"-- * apply_default_if\n"
"--\n"
"-- Others are just stored.\n"
"\n"
"local fun = require('fun')\n"
"local json = require('json')\n"
"\n"
"local methods = {}\n"
"local schema_mt = {}\n"
"\n"
"local scalars = {}\n"
"\n"
"-- {{{ Walkthrough helpers\n"
"\n"
"-- Create walkthrough context.\n"
"local function walkthrough_start(self, params)\n"
"    local ctx = {path = {}, name = rawget(self, 'name')}\n"
"    for k, v in pairs(params or {}) do\n"
"        ctx[k] = v\n"
"    end\n"
"    return ctx\n"
"end\n"
"\n"
"-- Step down to a field.\n"
"local function walkthrough_enter(ctx, name)\n"
"    table.insert(ctx.path, name)\n"
"end\n"
"\n"
"-- Step up from the last field.\n"
"local function walkthrough_leave(ctx)\n"
"    table.remove(ctx.path)\n"
"end\n"
"\n"
"-- Construct a string that describes the current path.\n"
"local function walkthrough_path(ctx)\n"
"    local res = ''\n"
"    for _, name in ipairs(ctx.path) do\n"
"        if type(name) == 'number' then\n"
"            res = res .. ('[%d]'):format(name)\n"
"        else\n"
"            res = res .. '.' .. name\n"
"        end\n"
"    end\n"
"    if res:startswith('.') then\n"
"        res = res:sub(2)\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"-- Generate a prefix for an error message based on the given\n"
"-- walkthrough context.\n"
"local function walkthrough_error_prefix(ctx)\n"
"    if ctx.path == nil or next(ctx.path) == nil then\n"
"        return ('[%s] '):format(ctx.name)\n"
"    end\n"
"    return ('[%s] %s: '):format(ctx.name, walkthrough_path(ctx))\n"
"end\n"
"\n"
"-- Generate an error supplemented by details from the given\n"
"-- walkthrough context.\n"
"local function walkthrough_error(ctx, message, ...)\n"
"    local error_prefix = walkthrough_error_prefix(ctx)\n"
"    error(('%s%s'):format(error_prefix, message:format(...)), 0)\n"
"end\n"
"\n"
"-- Return a function that raises an error with a prefix formed\n"
"-- from the given context.\n"
"--\n"
"-- The context information is gathered before the capturing.\n"
"local function walkthrough_error_capture(ctx)\n"
"    local error_prefix = walkthrough_error_prefix(ctx)\n"
"    return function(message, ...)\n"
"        error(('%s%s'):format(error_prefix, message:format(...)), 0)\n"
"    end\n"
"end\n"
"\n"
"-- Verify that the data is a table and, if it is not so, produce a\n"
"-- nice schema-aware error.\n"
"--\n"
"-- Applicable for a record, a map, an array.\n"
"--\n"
"-- Useful as part of validation, but also as a lightweight\n"
"-- consistency check.\n"
"local function walkthrough_assert_table(ctx, schema, data)\n"
"    assert(schema.type == 'record' or schema.type == 'map' or\n"
"        schema.type == 'array')\n"
"\n"
"    if type(data) == 'table' then\n"
"        return\n"
"    end\n"
"\n"
"    local article = schema.type == 'array' and 'an' or 'a'\n"
"    walkthrough_error(ctx, 'Unexpected data type for %s %s: %q', article,\n"
"        schema.type, type(data))\n"
"end\n"
"\n"
"-- }}} Walkthrough helpers\n"
"\n"
"-- {{{ Scalar definitions\n"
"\n"
"-- A scalar definition:\n"
"--\n"
"-- {\n"
"--     -- How the scalar is named.\n"
"--     type = <string>,\n"
"--     -- Check given data against the type constraints.\n"
"--\n"
"--     -> true (means the data is valid)\n"
"--     -> false, err (otherwise)\n"
"--     validate_noexc = <function>,\n"
"--\n"
"--     -- Parse data originated from an environment variable.\n"
"--     --\n"
"--     -- Should raise an error it is not possible to interpret\n"
"--     -- the data as a value of the given scalar type.\n"
"--     --\n"
"--     -- Return a parsed/typecasted value of the given scalar\n"
"--     -- type otherwise.\n"
"--     fromenv = <function>,\n"
"-- }\n"
"\n"
"-- Verify whether the given value (data) has expected type and\n"
"-- produce a human readable error message otherwise.\n"
"local function validate_type_noexc(data, exp_type)\n"
"    -- exp_type is like {'string', 'number'}.\n"
"    if type(exp_type) == 'table' then\n"
"        local found = false\n"
"        for _, exp_t in ipairs(exp_type) do\n"
"            if type(data) == exp_t then\n"
"                found = true\n"
"                break\n"
"            end\n"
"        end\n"
"        if not found then\n"
"            local exp_type_str = ('\"%s\"'):format(table.concat(exp_type, '\", \"'))\n"
"            local err = ('Expected one of %s, got %q'):format(exp_type_str,\n"
"                type(data))\n"
"            return false, err\n"
"        end\n"
"        return true\n"
"    end\n"
"\n"
"    -- exp_type is a Lua type like 'string'.\n"
"    assert(type(exp_type) == 'string')\n"
"    if type(data) ~= exp_type then\n"
"        local err = ('Expected %q, got %q'):format(exp_type, type(data))\n"
"        return false, err\n"
"    end\n"
"    return true\n"
"end\n"
"\n"
"scalars.string = {\n"
"    type = 'string',\n"
"    validate_noexc = function(data)\n"
"        return validate_type_noexc(data, 'string')\n"
"    end,\n"
"    fromenv = function(_env_var_name, raw_value)\n"
"        return raw_value\n"
"    end,\n"
"}\n"
"\n"
"scalars.number = {\n"
"    type = 'number',\n"
"    validate_noexc = function(data)\n"
"        -- TODO: Should we accept cdata<int64_t> and\n"
"        -- cdata<uint64_t> here\?\n"
"        return validate_type_noexc(data, 'number')\n"
"    end,\n"
"    fromenv = function(env_var_name, raw_value)\n"
"        -- TODO: Accept large integers and return cdata<int64_t>\n"
"        -- or cdata<uint64_t>\?\n"
"        local res = tonumber(raw_value)\n"
"        if res == nil then\n"
"            error(('Unable to decode a number value from environment ' ..\n"
"                'variable %q, got %q'):format(env_var_name, raw_value), 0)\n"
"        end\n"
"        return res\n"
"    end,\n"
"}\n"
"\n"
"-- TODO: This hack is needed until a union schema node will be\n"
"-- implemented.\n"
"scalars['string, number'] = {\n"
"    type = 'string, number',\n"
"    validate_noexc = function(data)\n"
"        return validate_type_noexc(data, {'string', 'number'})\n"
"    end,\n"
"    fromenv = function(_env_var_name, raw_value)\n"
"        return tonumber(raw_value) or raw_value\n"
"    end,\n"
"}\n"
"scalars['number, string'] = {\n"
"    type = 'number, string',\n"
"    validate_noexc = function(data)\n"
"        return validate_type_noexc(data, {'string', 'number'})\n"
"    end,\n"
"    fromenv = function(_env_var_name, raw_value)\n"
"        return tonumber(raw_value) or raw_value\n"
"    end,\n"
"}\n"
"\n"
"scalars.integer = {\n"
"    type = 'integer',\n"
"    validate_noexc = function(data)\n"
"        -- TODO: Accept cdata<int64_t> and cdata<uint64_t>.\n"
"        local ok, err = validate_type_noexc(data, 'number')\n"
"        if not ok then\n"
"            return false, err\n"
"        end\n"
"        if data - math.floor(data) ~= 0 then\n"
"            -- NB: %s is chosen deliberately: it formats a\n"
"            -- floating-point number in a more human friendly way\n"
"            -- than %f. For example, 5.5 vs 5.500000.\n"
"            local err = ('Expected number without a fractional part, ' ..\n"
"                'got %s'):format(data)\n"
"            return false, err\n"
"        end\n"
"        return true\n"
"    end,\n"
"    fromenv = function(env_var_name, raw_value)\n"
"        local res = tonumber64(raw_value)\n"
"        if res == nil then\n"
"            error(('Unable to decode an integer value from environment ' ..\n"
"                'variable %q, got %q'):format(env_var_name, raw_value), 0)\n"
"        end\n"
"        return res\n"
"    end,\n"
"}\n"
"\n"
"scalars.boolean = {\n"
"    type = 'boolean',\n"
"    validate_noexc = function(data)\n"
"        return validate_type_noexc(data, 'boolean')\n"
"    end,\n"
"    fromenv = function(env_var_name, raw_value)\n"
"        -- Accept false/true case insensitively.\n"
"        --\n"
"        -- Accept 0/1 as boolean values.\n"
"        if raw_value:lower() == 'false' or raw_value == '0' then\n"
"            return false\n"
"        end\n"
"        if raw_value:lower() == 'true' or raw_value == '1' then\n"
"            return true\n"
"        end\n"
"\n"
"        error(('Unable to decode a boolean value from environment ' ..\n"
"            'variable %q, got %q'):format(env_var_name, raw_value), 0)\n"
"    end,\n"
"}\n"
"\n"
"scalars.any = {\n"
"    type = 'any',\n"
"    validate_noexc = function(_data)\n"
"        -- No validation.\n"
"        return true\n"
"    end,\n"
"    fromenv = function(env_var_name, raw_value)\n"
"        -- Don't autoguess type. Accept JSON only.\n"
"        local ok, res = pcall(json.decode, raw_value)\n"
"        if not ok then\n"
"            error(('Unable to decode JSON data in environment ' ..\n"
"                'variable %q: %s'):format(env_var_name, res), 0)\n"
"        end\n"
"        return res\n"
"    end,\n"
"}\n"
"\n"
"local function is_scalar(schema)\n"
"    return scalars[schema.type] ~= nil\n"
"end\n"
"\n"
"-- }}} Scalar definitions\n"
"\n"
"-- {{{ Schema node constructors: scalar, record, map, array\n"
"\n"
"-- A schema node:\n"
"--\n"
"-- {\n"
"--     -- One of scalar types, 'record', 'map' or 'array'.\n"
"--     type = <string>,\n"
"--     -- For a record.\n"
"--     fields = <table>,\n"
"--     -- For a map.\n"
"--     key = <table>,\n"
"--     value = <table>,\n"
"--     -- For an array.\n"
"--     items = <table>,\n"
"--     -- Arbitrary user specified annotations.\n"
"--     <..annotations..>\n"
"-- }\n"
"\n"
"-- Create a scalar.\n"
"--\n"
"-- Example:\n"
"--\n"
"-- schema.scalar({\n"
"--     type = 'string',\n"
"--     <..annotations..>,\n"
"-- })\n"
"local function scalar(scalar_def)\n"
"    assert(scalar_def.type ~= nil)\n"
"    assert(is_scalar(scalar_def))\n"
"    return scalar_def\n"
"end\n"
"\n"
"-- Create a record.\n"
"--\n"
"-- A record node describes an object with the following properties:\n"
"--\n"
"-- * string keys\n"
"-- * certain keys (listed)\n"
"-- * certain value types (listed)\n"
"--\n"
"-- Example:\n"
"--\n"
"-- schema.record({\n"
"--     foo = <schema node>,\n"
"--     bar = <schema node>,\n"
"-- }, {\n"
"--     <..annotations..>\n"
"-- })\n"
"local function record(fields, annotations)\n"
"    local res = {\n"
"        type = 'record',\n"
"        fields = fields or {},\n"
"    }\n"
"    for k, v in pairs(annotations or {}) do\n"
"        assert(k ~= 'type' and k ~= 'fields')\n"
"        res[k] = v\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"-- Create a map.\n"
"--\n"
"-- A map node describes an object with the following properties:\n"
"--\n"
"-- * arbitrary keys\n"
"-- * all keys have the same certain type\n"
"-- * all values have the same certain type\n"
"--\n"
"-- Example:\n"
"--\n"
"-- schema.map({\n"
"--     key = <schema node>,\n"
"--     value = <schema node>,\n"
"--     <..annotations..>\n"
"-- })\n"
"local function map(map_def)\n"
"    assert(map_def.key ~= nil)\n"
"    assert(map_def.value ~= nil)\n"
"    assert(map_def.type == nil)\n"
"    local res = table.copy(map_def)\n"
"    res.type = 'map'\n"
"    return res\n"
"end\n"
"\n"
"-- Create an array.\n"
"--\n"
"-- Example:\n"
"--\n"
"-- schema.array({\n"
"--     items = <schema node>,\n"
"--     <..annotations..>\n"
"-- })\n"
"local function array(array_def)\n"
"    assert(array_def.items ~= nil)\n"
"    assert(array_def.type == nil)\n"
"    local res = table.copy(array_def)\n"
"    res.type = 'array'\n"
"    return res\n"
"end\n"
"\n"
"-- }}} Schema node constructors: scalar, record, map, array\n"
"\n"
"-- {{{ Derived schema node type constructors: enum, set\n"
"\n"
"local function validate_no_repeat(data, w)\n"
"    local visited = {}\n"
"    for _, item in ipairs(data) do\n"
"        assert(type(item) == 'string')\n"
"        if visited[item] then\n"
"            w.error('Values should be unique, but %q appears at ' ..\n"
"                'least twice', item)\n"
"        end\n"
"        visited[item] = true\n"
"    end\n"
"end\n"
"\n"
"-- Shortcut for a string scalar with given allowed values.\n"
"local function enum(allowed_values, annotations)\n"
"    local scalar_def = {\n"
"        type = 'string',\n"
"        allowed_values = allowed_values,\n"
"    }\n"
"    for k, v in pairs(annotations or {}) do\n"
"        assert(k ~= 'type' and k ~= 'allowed_values')\n"
"        scalar_def[k] = v\n"
"    end\n"
"    return scalar(scalar_def)\n"
"end\n"
"\n"
"-- Shortcut for array of unique string values from the given list\n"
"-- of allowed values.\n"
"local function set(allowed_values, annotations)\n"
"    local array_def = {\n"
"        items = enum(allowed_values),\n"
"        validate = validate_no_repeat,\n"
"    }\n"
"    for k, v in pairs(annotations or {}) do\n"
"        assert(k ~= 'type' and k ~= 'items' and k ~= 'validate')\n"
"        array_def[k] = v\n"
"    end\n"
"    return array(array_def)\n"
"end\n"
"\n"
"-- }}} Derived schema node type constructors: enum, set\n"
"\n"
"-- {{{ <schema object>:validate()\n"
"\n"
"-- Verify that the given table adheres array requirements.\n"
"--\n"
"-- It accepts an array without holes.\n"
"--\n"
"-- Strictly speaking,\n"
"--\n"
"-- * If the table is empty it is OK.\n"
"-- * If the table is non-empty, the constraints are the following:\n"
"--   * all keys are numeric, without a fractional part\n"
"--   * the lower key is 1\n"
"--   * the higher key is equal to the number of items\n"
"local function validate_table_is_array(data, ctx)\n"
"    assert(type(data) == 'table')\n"
"\n"
"    -- Check that all the keys are numeric.\n"
"    local key_count = 0\n"
"    local min_key = 1/0  -- +inf\n"
"    local max_key = -1/0 -- -inf\n"
"    for k, _ in pairs(data) do\n"
"        if type(k) ~= 'number' then\n"
"            walkthrough_error(ctx, 'An array contains a non-numeric ' ..\n"
"                'key: %q', k)\n"
"        end\n"
"        if k - math.floor(k) ~= 0 then\n"
"            walkthrough_error(ctx, 'An array contains a non-integral ' ..\n"
"                'numeric key: %s', k)\n"
"        end\n"
"        key_count = key_count + 1\n"
"        min_key = math.min(min_key, k)\n"
"        max_key = math.max(max_key, k)\n"
"    end\n"
"\n"
"    -- An empty array is a valid array.\n"
"    if key_count == 0 then\n"
"        return\n"
"    end\n"
"\n"
"    -- Check that the array starts from 1 and has no holes.\n"
"    if min_key ~= 1 then\n"
"        walkthrough_error(ctx, 'An array must start from index 1, ' ..\n"
"            'got min index %d', min_key)\n"
"    end\n"
"\n"
"    -- Check that the array has no holes.\n"
"    if max_key ~= key_count then\n"
"        walkthrough_error(ctx, 'An array must not have holes, got ' ..\n"
"            'a table with %d integer fields with max index %d', key_count,\n"
"            max_key)\n"
"    end\n"
"end\n"
"\n"
"-- Verify the given data against the `allowed_values` annotation\n"
"-- in the schema node (if present).\n"
"local function validate_by_allowed_values(schema, data, ctx)\n"
"    if schema.allowed_values == nil then\n"
"        return\n"
"    end\n"
"\n"
"    assert(type(schema.allowed_values) == 'table')\n"
"    local found = false\n"
"    for _, v in ipairs(schema.allowed_values) do\n"
"        if data == v then\n"
"            found = true\n"
"            break\n"
"        end\n"
"    end\n"
"    if not found then\n"
"        walkthrough_error(ctx, 'Got %s, but only the following values ' ..\n"
"            'are allowed: %s', data, table.concat(schema.allowed_values, ', '))\n"
"    end\n"
"end\n"
"\n"
"-- Call schema node specific validation function.\n"
"local function validate_by_node_function(schema, data, ctx)\n"
"    if schema.validate == nil then\n"
"        return\n"
"    end\n"
"\n"
"    assert(type(schema.validate) == 'function')\n"
"    local w = {\n"
"        schema = schema,\n"
"        -- ctx.path is modified during the traversal, so an\n"
"        -- attempt to store it and use later would give\n"
"        -- an unexpected result. Let's copy it to avoid the\n"
"        -- confusion.\n"
"        path = table.copy(ctx.path),\n"
"        error = walkthrough_error_capture(ctx),\n"
"    }\n"
"    schema.validate(data, w)\n"
"end\n"
"\n"
"local function validate_impl(schema, data, ctx)\n"
"    if is_scalar(schema) then\n"
"        local scalar_def = scalars[schema.type]\n"
"        assert(scalar_def ~= nil)\n"
"\n"
"        local ok, err = scalar_def.validate_noexc(data)\n"
"        if not ok then\n"
"            walkthrough_error(ctx, 'Unexpected data for scalar %q: %s',\n"
"                schema.type, err)\n"
"        end\n"
"    elseif schema.type == 'record' then\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"\n"
"        for field_name, field_def in pairs(schema.fields) do\n"
"            walkthrough_enter(ctx, field_name)\n"
"            local field = data[field_name]\n"
"            -- Assume fields as non-required.\n"
"            if field ~= nil then\n"
"                validate_impl(field_def, field, ctx)\n"
"            end\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"\n"
"        -- Walk over the data to catch unknown fields.\n"
"        for field_name, _ in pairs(data) do\n"
"            local field_def = schema.fields[field_name]\n"
"            if field_def == nil then\n"
"                walkthrough_error(ctx, 'Unexpected field %q', field_name)\n"
"            end\n"
"        end\n"
"    elseif schema.type == 'map' then\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"\n"
"        for field_name, field_value in pairs(data) do\n"
"            walkthrough_enter(ctx, field_name)\n"
"            validate_impl(schema.key, field_name, ctx)\n"
"            validate_impl(schema.value, field_value, ctx)\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"    elseif schema.type == 'array' then\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"        validate_table_is_array(data, ctx)\n"
"\n"
"        for i, v in ipairs(data) do\n"
"            walkthrough_enter(ctx, i)\n"
"            validate_impl(schema.items, v, ctx)\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"\n"
"    validate_by_allowed_values(schema, data, ctx)\n"
"\n"
"    -- Call schema node specific validation function.\n"
"    --\n"
"    -- Important: it is called when all the type validation is\n"
"    -- already done, including nested nodes.\n"
"    validate_by_node_function(schema, data, ctx)\n"
"end\n"
"\n"
"-- Validate the given data against the given schema.\n"
"--\n"
"-- Nuances:\n"
"--\n"
"-- * `schema.new('<...>', schema.scalar(<...>))` doesn't accept\n"
"--   `nil` and `box.NULL`. However,\n"
"-- * All fields in a record are optional: they accept `nil` and\n"
"--   `box.NULL`.\n"
"-- * The record/map/array determination is purely schema based.\n"
"--   mt.__serialize marks in the data are not involved anyhow.\n"
"-- * An array shouldn't have any holes (nil values in a middle).\n"
"--\n"
"-- Annotations taken into accounts:\n"
"--\n"
"-- * allowed_values (table) -- whitelist of values\n"
"-- * validate (function) -- schema node specific validator\n"
"--\n"
"--   validate = function(data, w)\n"
"--       -- w.schema -- current schema node\n"
"--       -- w.path -- path to the node\n"
"--       -- w.error -- function that prepends a caller provided\n"
"--       --            error message with context information;\n"
"--       --            use it for nice error messages\n"
"--   end\n"
"function methods.validate(self, data)\n"
"    local ctx = walkthrough_start(self)\n"
"    validate_impl(rawget(self, 'schema'), data, ctx)\n"
"end\n"
"\n"
"-- }}} <schema object>:validate()\n"
"\n"
"-- {{{ :get()/:set() helpers\n"
"\n"
"-- The path can be passed as a string in the dot notation or as a\n"
"-- table representing an array of components. This function\n"
"-- converts the path into the array if necessary.\n"
"local function normalize_path(path, error_f)\n"
"    if type(path) ~= 'string' and type(path) ~= 'table' then\n"
"        return error_f()\n"
"    end\n"
"\n"
"    -- Dot notation/JSON path alike.\n"
"    --\n"
"    -- TODO: Support numeric indexing: 'foo[1]' and `{'foo', 1}`.\n"
"    if type(path) == 'string' then\n"
"        -- NB: string.split('') returns {''}.\n"
"        if path == '' then\n"
"            path = {}\n"
"        else\n"
"            path = path:split('.')\n"
"        end\n"
"    end\n"
"\n"
"    return path\n"
"end\n"
"\n"
"-- }}} :get()/:set() helpers\n"
"\n"
"-- {{{ <schema object>:get()\n"
"\n"
"local function get_usage()\n"
"    error('Usage: schema:get(data: <as defined by the schema>, ' ..\n"
"        'path: nil/string/table)', 0)\n"
"end\n"
"\n"
"local function get_impl(schema, data, ctx)\n"
"    -- The journey is finished. Return what is under the feet.\n"
"    if #ctx.journey == 0 then\n"
"        return data\n"
"    end\n"
"\n"
"    -- There are more steps in the journey (at least one).\n"
"    -- Let's dive deeper and process it per schema node type.\n"
"\n"
"    local requested_field = ctx.journey[1]\n"
"    assert(requested_field ~= nil)\n"
"\n"
"    if is_scalar(schema) then\n"
"        walkthrough_error(ctx, 'Attempt to index a scalar value of type %s ' ..\n"
"            'by field %q', schema.type, requested_field)\n"
"    elseif schema.type == 'record' then\n"
"        walkthrough_enter(ctx, requested_field)\n"
"        local field_def = schema.fields[requested_field]\n"
"        if field_def == nil then\n"
"            walkthrough_error(ctx, 'No such field in the schema')\n"
"        end\n"
"\n"
"        -- Even if there is no such field in the data, continue\n"
"        -- the descending to validate the path against the schema.\n"
"        local field_value\n"
"        if data ~= nil then\n"
"            field_value = data[requested_field]\n"
"        end\n"
"\n"
"        table.remove(ctx.journey, 1)\n"
"        return get_impl(field_def, field_value, ctx)\n"
"    elseif schema.type == 'map' then\n"
"        walkthrough_enter(ctx, requested_field)\n"
"        local field_def = schema.value\n"
"\n"
"        -- Even if there is no such field in the data, continue\n"
"        -- the descending to validate the path against the schema.\n"
"        local field_value\n"
"        if data ~= nil then\n"
"            field_value = data[requested_field]\n"
"        end\n"
"\n"
"        table.remove(ctx.journey, 1)\n"
"        return get_impl(field_def, field_value, ctx)\n"
"    elseif schema.type == 'array' then\n"
"        -- TODO: Support 'foo[1]' and `{'foo', 1}` paths. See the\n"
"        -- normalize_path() function.\n"
"        walkthrough_error(ctx, 'Indexing an array is not supported yet')\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"-- Get nested data that is pointed by the given path.\n"
"--\n"
"-- Important: the data is assumed as already validated against\n"
"-- the given schema.\n"
"--\n"
"-- The indexing is performed in the optional chaining manner\n"
"-- ('foo.bar' works like foo\?.bar in TypeScript).\n"
"--\n"
"-- The function checks the path against the schema: it doesn't\n"
"-- allow to use a non-existing field or index a scalar value.\n"
"--\n"
"-- The path is either array-like table or a string in the dot\n"
"-- notation.\n"
"--\n"
"-- local data = {foo = {bar = 'x'}}\n"
"-- schema:get(data, 'foo.bar') -> 'x'\n"
"-- schema:get(data, {'foo', 'bar'}) -> 'x'\n"
"--\n"
"-- local data = {}\n"
"-- schema:get(data, 'foo.bar') -> nil\n"
"--\n"
"-- Nuances:\n"
"--\n"
"-- * Array indexing is not supported yet.\n"
"-- * A scalar of the 'any' type can't be indexed, even when it is\n"
"--   a table. It is OK to acquire the whole value of the 'any'\n"
"--   type.\n"
"function methods.get(self, data, path)\n"
"    local schema = rawget(self, 'schema')\n"
"\n"
"    -- It is easy to forget about the `data` argument or misorder\n"
"    -- the `data` and the `path` arguments. Let's add a fast check\n"
"    -- for the most common scenario: a record schema and a string\n"
"    -- path.\n"
"    if schema.type == 'record' and type(data) ~= 'table' then\n"
"        return get_usage()\n"
"    end\n"
"\n"
"    if path ~= nil then\n"
"        path = normalize_path(path, get_usage)\n"
"    end\n"
"\n"
"    if path == nil or next(path) == nil then\n"
"        return data\n"
"    end\n"
"\n"
"    local ctx = walkthrough_start(self, {\n"
"        -- The `path` field is already in the context and it means\n"
"        -- the passed path. Let's name the remaining path as\n"
"        -- `journey`.\n"
"        journey = path,\n"
"    })\n"
"    return get_impl(schema, data, ctx)\n"
"end\n"
"\n"
"-- }}} <schema object>:get()\n"
"\n"
"-- {{{ <schema object>:set()\n"
"\n"
"local function set_usage()\n"
"    error('Usage: schema:set(data: <as defined by the schema>, ' ..\n"
"        'path: string/table, rhs: <as defined by the schema>)', 0)\n"
"end\n"
"\n"
"local function set_impl(schema, data, rhs, ctx)\n"
"    -- The journey is finished. Validate and return the new value.\n"
"    if #ctx.journey == 0 then\n"
"        -- Call validate_impl() directly to don't construct a\n"
"        -- schema object.\n"
"        validate_impl(schema, rhs, ctx)\n"
"        return rhs\n"
"    end\n"
"\n"
"    local requested_field = ctx.journey[1]\n"
"    assert(requested_field ~= nil)\n"
"\n"
"    if is_scalar(schema) then\n"
"        walkthrough_error(ctx, 'Attempt to index a scalar value of type %s ' ..\n"
"            'by field %q', schema.type, requested_field)\n"
"    elseif schema.type == 'record' then\n"
"        walkthrough_enter(ctx, requested_field)\n"
"        local field_def = schema.fields[requested_field]\n"
"        if field_def == nil then\n"
"            walkthrough_error(ctx, 'No such field in the schema')\n"
"        end\n"
"\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"        local field_value = data[requested_field] or {}\n"
"        table.remove(ctx.journey, 1)\n"
"        data[requested_field] = set_impl(field_def, field_value, rhs, ctx)\n"
"        return data\n"
"    elseif schema.type == 'map' then\n"
"        walkthrough_enter(ctx, requested_field)\n"
"        local field_def = schema.value\n"
"\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"        local field_value = data[requested_field] or {}\n"
"        table.remove(ctx.journey, 1)\n"
"        data[requested_field] = set_impl(field_def, field_value, rhs, ctx)\n"
"        return data\n"
"    elseif schema.type == 'array' then\n"
"        -- TODO: Support 'foo[1]' and `{'foo', 1}` paths. See the\n"
"        -- normalize_path() function.\n"
"        walkthrough_error(ctx, 'Indexing an array is not supported yet')\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"-- Set the given `rhs` value at the given path in the `data`.\n"
"--\n"
"-- Important: `data` is assumed as already validated against the\n"
"-- given schema, but `rhs` is validated by the method before the\n"
"-- assignment.\n"
"--\n"
"-- The function checks the path against the schema: it doesn't\n"
"-- allow to use a non-existing field or index a scalar value.\n"
"--\n"
"-- The path is either array-like table or a string in the dot\n"
"-- notation.\n"
"--\n"
"-- local data = {}\n"
"-- schema:set(data, 'foo.bar', 42)\n"
"-- print(data.foo.bar) -- 42\n"
"--\n"
"-- local data = {}\n"
"-- schema:set(data, {'foo', 'bar'}, 42)\n"
"-- print(data.foo.bar) -- 42\n"
"--\n"
"-- Nuances:\n"
"--\n"
"-- * A root node (pointed by the empty path) can't be set using\n"
"--   this method.\n"
"-- * Array indexing is not supported yet.\n"
"-- * A scalar of the 'any' type can't be indexed, even when it is\n"
"--   a table. It is OK to set the whole value of the 'any'\n"
"--   type.\n"
"function methods.set(self, data, path, rhs)\n"
"    local schema = rawget(self, 'schema')\n"
"\n"
"    -- Detect `data` and `path` misordering for the most common\n"
"    -- scenario: a record schema and a string path.\n"
"    if schema.type == 'record' and type(data) ~= 'table' then\n"
"        return set_usage()\n"
"    end\n"
"\n"
"    if path ~= nil then\n"
"        path = normalize_path(path, set_usage)\n"
"    end\n"
"\n"
"    if path == nil or next(path) == nil then\n"
"        error('schema:set: empty path', 0)\n"
"    end\n"
"\n"
"    local ctx = walkthrough_start(self, {\n"
"        -- The `path` field is already in the context and it means\n"
"        -- the passed path. Let's name the remaining path as\n"
"        -- `journey`.\n"
"        journey = path,\n"
"    })\n"
"    return set_impl(schema, data, rhs, ctx)\n"
"end\n"
"\n"
"-- }}} <schema object>:set()\n"
"\n"
"-- {{{ <schema object>:filter()\n"
"\n"
"local function filter_impl(schema, data, f, ctx)\n"
"    local w = {\n"
"        path = table.copy(ctx.path),\n"
"        schema = schema,\n"
"        data = data,\n"
"    }\n"
"    if f(w) then\n"
"        table.insert(ctx.acc, w)\n"
"    end\n"
"\n"
"    -- The exit condition is after the table.insert(), because a\n"
"    -- caller may want to handle box.NULL values somehow.\n"
"    if data == nil then\n"
"        return\n"
"    end\n"
"\n"
"    -- luacheck: ignore 542 empty if branch\n"
"    if is_scalar(schema) then\n"
"        -- Nothing to do.\n"
"    elseif schema.type == 'record' then\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"\n"
"        for field_name, field_def in pairs(schema.fields) do\n"
"            walkthrough_enter(ctx, field_name)\n"
"            local field_value = data[field_name]\n"
"            -- Step down if the field exists in the data. box.NULL\n"
"            -- field value is interpreted here as an existing\n"
"            -- field.\n"
"            if type(field_value) ~= 'nil' then\n"
"                filter_impl(field_def, field_value, f, ctx)\n"
"            end\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"    elseif schema.type == 'map' then\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"\n"
"        for field_name, field_value in pairs(data) do\n"
"            walkthrough_enter(ctx, field_name)\n"
"            filter_impl(schema.key, field_name, f, ctx)\n"
"            filter_impl(schema.value, field_value, f, ctx)\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"    elseif schema.type == 'array' then\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"\n"
"        for i, v in ipairs(data) do\n"
"            walkthrough_enter(ctx, i)\n"
"            filter_impl(schema.items, v, f, ctx)\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"-- Filter data based on the schema annotations.\n"
"--\n"
"-- Important: the data is assumed as already validated against\n"
"-- the given schema. (A fast type check is performed on composite\n"
"-- types, but it is not recommended to lean on it.)\n"
"--\n"
"-- The user-provided filter function `f` receives the following\n"
"-- table as the argument:\n"
"--\n"
"-- w = {\n"
"--     path = <array-like table>,\n"
"--     schema = <schema node>,\n"
"--     data = <data at the given path>,\n"
"-- }\n"
"--\n"
"-- And returns a boolean value that is interpreted as 'accepted'\n"
"-- or 'not accepted'.\n"
"--\n"
"-- The user-provided function `f` is called for each schema node,\n"
"-- including ones that have box.NULL value (but not nil). A node\n"
"-- of a composite type (record/map/array) is not traversed down\n"
"-- if it has nil or box.NULL value.\n"
"--\n"
"-- The `:filter()` function returns a luafun iterator by all `w`\n"
"-- values accepted by the `f` function.\n"
"--\n"
"-- A composite node that is not accepted still traversed down.\n"
"--\n"
"-- Examples:\n"
"--\n"
"-- -- Do something for each piece of data that is marked by the\n"
"-- -- given annotation.\n"
"-- s:filter(function(w)\n"
"--     return w.schema.my_annotation ~= nil\n"
"-- end):each(function(w)\n"
"--     do_something(w.data)\n"
"-- end)\n"
"--\n"
"-- -- Group data by a value of an annotation.\n"
"-- local group_by_my_annotation = s:filter(function(w)\n"
"--     return w.schema.my_annotation ~= nil\n"
"-- end):map(function(w)\n"
"--     return w.schema.my_annotation, w.data\n"
"-- end):tomap()\n"
"--\n"
"-- Nuances:\n"
"--\n"
"-- * box.NULL is assumed as an existing value, so the\n"
"--   user-provided filter function `f` is called for it. However,\n"
"--   it is not called for `nil` values. See details below.\n"
"-- * While it is technically possible to pass information about\n"
"--   a field name for record/map field values and about an item\n"
"--   index for an array item value, it is not implemented for\n"
"--   simplicity (and because it is not needed in config's code).\n"
"-- * `w.path` for a map key and a map value are the same. It\n"
"--   seems, we should introduce some syntax to point a key in a\n"
"--   map, but the config module doesn't need it, so it is not\n"
"--   implemented.\n"
"--\n"
"-- nil/box.NULL nuances explanation\n"
"-- --------------------------------\n"
"--\n"
"-- Let's assume that a record defines three scalar fields: 'foo',\n"
"-- 'bar' and 'baz'. Let's name a schema object that wraps the\n"
"-- record as `s`.\n"
"--\n"
"-- * `s:filter(nil, f)` calls `f` only for the record itself.\n"
"-- * `s:filter(box.NULL, f` works in the same way.\n"
"-- * `s:filter({foo = box.NULL, bar = nil}, f)` calls `f` two\n"
"--   times: for the record and for the 'foo' field.\n"
"--\n"
"-- This behavior is needed to provide ability to handle box.NULL\n"
"-- values in the data somehow. It reflects the pairs() behavior on\n"
"-- a usual table, so it looks quite natural.\n"
"function methods.filter(self, data, f)\n"
"    local ctx = walkthrough_start(self, {acc = {}})\n"
"    filter_impl(rawget(self, 'schema'), data, f, ctx)\n"
"    return fun.iter(ctx.acc)\n"
"end\n"
"\n"
"-- }}} <schema object>:filter()\n"
"\n"
"-- {{{ <schema object>:map()\n"
"\n"
"local function map_impl(schema, data, f, ctx)\n"
"    if is_scalar(schema) then\n"
"        -- We're reached a scalar: let's call the user-provided\n"
"        -- transformation function.\n"
"        local w = {\n"
"            schema = schema,\n"
"            path = table.copy(ctx.path),\n"
"            error = walkthrough_error_capture(ctx),\n"
"        }\n"
"        return f(data, w, ctx.f_ctx)\n"
"    elseif schema.type == 'record' then\n"
"        -- Traverse record's fields unconditionally: if the record\n"
"        -- itself is nil/box.NULL, if the fields are nil/box.NULL.\n"
"        --\n"
"        -- Missed values may be transformed to something non-nil.\n"
"        if data ~= nil then\n"
"            walkthrough_assert_table(ctx, schema, data)\n"
"        end\n"
"\n"
"        -- Collect the new field values.\n"
"        local res = {}\n"
"        for field_name, field_def in pairs(schema.fields) do\n"
"            walkthrough_enter(ctx, field_name)\n"
"\n"
"            local field\n"
"            if data ~= nil then\n"
"                field = data[field_name]\n"
"            end\n"
"            res[field_name] = map_impl(field_def, field, f, ctx)\n"
"\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"\n"
"        -- If the original value is nil/box.NULL and all the new\n"
"        -- fields are nil, let's preserve the original value as\n"
"        -- is: return nil/box.NULL instead of an empty table.\n"
"        if next(res) == nil and data == nil then\n"
"            return data -- nil or box.NULL\n"
"        end\n"
"        return res\n"
"    elseif schema.type == 'map' then\n"
"        -- If a map is nil/box.NULL, there is nothing to traverse.\n"
"        --\n"
"        -- Let's just return the original value.\n"
"        if data == nil then\n"
"            return data -- nil or box.NULL\n"
"        end\n"
"\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"\n"
"        local res = {}\n"
"        for field_name, field_value in pairs(data) do\n"
"            walkthrough_enter(ctx, field_name)\n"
"            local new_field_name = map_impl(schema.key, field_name, f, ctx)\n"
"            local new_field_value = map_impl(schema.value, field_value, f, ctx)\n"
"            res[new_field_name] = new_field_value\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"        return res\n"
"    elseif schema.type == 'array' then\n"
"        -- If an array is nil/box.NULL, there is nothing to\n"
"        -- traverse.\n"
"        --\n"
"        -- Just return the original value.\n"
"        if data == nil then\n"
"            return data -- nil or box.NULL\n"
"        end\n"
"\n"
"        walkthrough_assert_table(ctx, schema, data)\n"
"\n"
"        local res = {}\n"
"        for i, v in ipairs(data) do\n"
"            walkthrough_enter(ctx, i)\n"
"            local new_item_value = map_impl(schema.items, v, f, ctx)\n"
"            res[i] = new_item_value\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"        return res\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"-- Transform data by the given function.\n"
"--\n"
"-- Leave the shape of the data unchanged.\n"
"--\n"
"-- Important: the data is assumed as already validated against\n"
"-- the given schema. (A fast type check is performed on composite\n"
"-- types, but it is not recommended to lean on it.)\n"
"--\n"
"-- The user-provided transformation function receives the\n"
"-- following three arguments in the given order:\n"
"--\n"
"-- * data -- value at the given path\n"
"-- * w -- walkthrough node, described below\n"
"-- * ctx -- user-provided context for the transformation function\n"
"--\n"
"-- The walkthrough node `w` has the following fields:\n"
"--\n"
"-- * w.schema -- schema node at the given path\n"
"-- * w.path -- path to the schema node\n"
"-- * w.error -- function that prepends a caller provided error\n"
"--   message with context information; use it for nice error\n"
"--   messages\n"
"--\n"
"-- An example of the mapping function:\n"
"--\n"
"-- local function f(data, w, ctx)\n"
"--     if w.schema.type == 'string' and data ~= nil then\n"
"--         return data:gsub('{{ *foo *}}', ctx.foo)\n"
"--     end\n"
"--     return data\n"
"-- end\n"
"--\n"
"-- The :map() method is recursive with certain rules:\n"
"--\n"
"-- * All record fields are traversed unconditionally, including\n"
"--   ones with nil/box.NULL values. Even if the record itself is\n"
"--   nil/box.NULL, its fields are traversed down (assuming their\n"
"--   values as nil).\n"
"--\n"
"--   It is important when the original data should be extended\n"
"--   using some information from the schema: say, default values.\n"
"-- * It is not the case for a map and an array: nil/box.NULL\n"
"--   fields and items are preserved as is, they're not traversed\n"
"--   down. If the map/the array itself is nil/box.NULL, it is\n"
"--   preserved as well.\n"
"--\n"
"--   A map has no list of fields in the schema, so it is not\n"
"--   possible to traverse it down. Similarly, an array has no\n"
"--   items count in the schema.\n"
"--\n"
"-- The method attempts to preserve the original shape of values\n"
"-- of a composite type:\n"
"--\n"
"-- * nil/box.NULL record is traversed down, but if all the new\n"
"--   field values are nil, the return value is the original one\n"
"--   (nil/box.NULL), not an empty table.\n"
"-- * nil/box.NULL values for a map and an array are preserved.\n"
"--\n"
"-- Nuances:\n"
"--\n"
"-- * The user-provided transformation function is called only for\n"
"--   scalars.\n"
"-- * nil/box.NULL handling for composite types. Described above.\n"
"-- * `w.path` for a map key and a map value are the same. It\n"
"--   seems, we should introduce some syntax to point a key in a\n"
"--   map, but the config module doesn't need it, so it is not\n"
"--   implemented.\n"
"function methods.map(self, data, f, f_ctx)\n"
"    local ctx = walkthrough_start(self, {f_ctx = f_ctx})\n"
"    return map_impl(rawget(self, 'schema'), data, f, ctx)\n"
"end\n"
"\n"
"-- }}} <schema object>:map()\n"
"\n"
"-- {{{ <schema object>:apply_default()\n"
"\n"
"local function apply_default_f(data, w)\n"
"    -- The value is present, keep it.\n"
"    --\n"
"    -- box.NULL is assumed as a missed value.\n"
"    if data ~= nil then\n"
"        return data\n"
"    end\n"
"\n"
"    -- Don't replace box.NULL in the original data with nil if\n"
"    -- there is no 'default' annotation.\n"
"    if type(w.schema.default) == 'nil' then\n"
"        return data\n"
"    end\n"
"\n"
"    -- Determine whether to apply the default.\n"
"    --\n"
"    -- Perform the apply if the apply_default_if annotation\n"
"    -- returns true or if there is no such an annotation.\n"
"    --\n"
"    -- Keep the original value otherwise.\n"
"    local apply_default = true\n"
"    if w.schema.apply_default_if ~= nil then\n"
"        assert(type(w.schema.apply_default_if) == 'function')\n"
"        apply_default = w.schema.apply_default_if(data, w)\n"
"    end\n"
"\n"
"    if apply_default then\n"
"        return w.schema.default\n"
"    end\n"
"\n"
"    return data\n"
"end\n"
"\n"
"-- Apply default values from the schema.\n"
"--\n"
"-- Important: the data is assumed as already validated against\n"
"-- the given schema. (A fast type check is performed on composite\n"
"-- types, but it is not recommended to lean on it.)\n"
"--\n"
"-- Nuances:\n"
"--\n"
"-- * Defaults are taken into account only for scalars.\n"
"--\n"
"-- Annotations taken into accounts:\n"
"--\n"
"-- * default -- the value to be placed instead of a missed one\n"
"-- * apply_default_if (function) -- whether to apply the default\n"
"--\n"
"--   apply_default_if = function(data, w)\n"
"--       -- w.schema -- current schema node\n"
"--       -- w.path -- path to the node\n"
"--       -- w.error -- for nice error messages\n"
"--   end\n"
"--\n"
"--   If there is no apply_default_if annotation, the default is\n"
"--   assumed as to be applied.\n"
"function methods.apply_default(self, data)\n"
"    return self:map(data, apply_default_f)\n"
"end\n"
"\n"
"-- }}} <schema object>:apply_default()\n"
"\n"
"-- {{{ <schema object>:merge()\n"
"\n"
"local function merge_impl(schema, a, b, ctx)\n"
"    -- Prefer box.NULL over nil.\n"
"    if a == nil and b == nil then\n"
"        if type(a) == 'nil' then\n"
"            return b\n"
"        else\n"
"            return a\n"
"        end\n"
"    end\n"
"\n"
"    -- Prefer X ~= nil over nil/box.NULL.\n"
"    if a == nil then\n"
"        return b\n"
"    elseif b == nil then\n"
"        return a\n"
"    end\n"
"\n"
"    assert(a ~= nil and b ~= nil)\n"
"\n"
"    -- Scalars and arrays are not to be merged.\n"
"    --\n"
"    -- At this point neither a == nil, nor b == nil, so return the\n"
"    -- preferred value, `b`.\n"
"    if is_scalar(schema) then\n"
"        return b\n"
"    elseif schema.type == 'array' then\n"
"        walkthrough_assert_table(ctx, schema, a)\n"
"        walkthrough_assert_table(ctx, schema, b)\n"
"\n"
"        return b\n"
"    end\n"
"\n"
"    -- `a` and `b` are both non-nil non-box.NULL records or maps.\n"
"    -- Perform the deep merge.\n"
"    if schema.type == 'record' then\n"
"        walkthrough_assert_table(ctx, schema, a)\n"
"        walkthrough_assert_table(ctx, schema, b)\n"
"\n"
"        local res = {}\n"
"        for field_name, field_def in pairs(schema.fields) do\n"
"            walkthrough_enter(ctx, field_name)\n"
"            local a_field = a[field_name]\n"
"            local b_field = b[field_name]\n"
"            res[field_name] = merge_impl(field_def, a_field, b_field, ctx)\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"        return res\n"
"    elseif schema.type == 'map' then\n"
"        walkthrough_assert_table(ctx, schema, a)\n"
"        walkthrough_assert_table(ctx, schema, b)\n"
"\n"
"        local res = {}\n"
"        for field_name, a_field in pairs(a) do\n"
"            walkthrough_enter(ctx, field_name)\n"
"            local b_field = b[field_name]\n"
"            res[field_name] = merge_impl(schema.value, a_field, b_field, ctx)\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"        -- NB: No error is possible, so let's skip\n"
"        -- walkthrough_enter()/walkthrough_leave().\n"
"        for field_name, b_field in pairs(b) do\n"
"            if type(a[field_name]) == 'nil' then\n"
"                res[field_name] = b_field\n"
"            end\n"
"        end\n"
"        return res\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"-- Merge two hierarical values (prefer the latter).\n"
"--\n"
"-- Important: the data is assumed as already validated against\n"
"-- the given schema. (A fast type check is performed on composite\n"
"-- types, but it is not recommended to lean on it.)\n"
"--\n"
"-- box.NULL is preferred over nil, any X where X ~= nil is\n"
"-- preferred over nil/box.NULL.\n"
"--\n"
"-- Records and maps are deeply merged. Scalars and arrays are\n"
"-- all-or-nothing: the right hand one is chosen if both are\n"
"-- not nil/box.NULL.\n"
"--\n"
"-- The formal rules are below.\n"
"--\n"
"-- Let's define the merge result for nil and box.NULL values:\n"
"--\n"
"-- 1. merge(nil, nil) -> nil\n"
"-- 2. merge(nil, box.NULL) -> box.NULL\n"
"-- 3. merge(box.NULL, nil) -> box.NULL\n"
"-- 4. merge(box.NULL, box.NULL) -> box.NULL\n"
"--\n"
"-- Let's define X as a value that is not nil and is not box.NULL.\n"
"--\n"
"-- 5. merge(X, nil) -> X\n"
"-- 6. merge(X, box.NULL) -> X\n"
"-- 7. merge(nil, X) -> X\n"
"-- 8. merge(box.NULL, X) -> X\n"
"--\n"
"-- If the above conditions are not meet, the following type\n"
"-- specific rules are is effect.\n"
"--\n"
"-- 9. merge(<scalar A>, <scalar B>) -> <scalar B>\n"
"-- 10. merge(<array A>, <array B>) -> <array B>\n"
"-- 11. merge(<record A>, <record B>) -> deep-merge(A, B)\n"
"-- 12. merge(<map A>, <map B>) -> deep-merge(A, B)\n"
"--\n"
"-- For each key K in A and each key K in B: deep-merge(A, B)[K] is\n"
"-- merge(A[K], B[K]).\n"
"function methods.merge(self, a, b)\n"
"    local ctx = walkthrough_start(self)\n"
"    return merge_impl(rawget(self, 'schema'), a, b, ctx)\n"
"end\n"
"\n"
"-- }}} <schema object>:merge()\n"
"\n"
"-- {{{ <schema object>:pairs()\n"
"\n"
"local function schema_pairs_append_node(schema, ctx)\n"
"    table.insert(ctx.acc, {\n"
"        path = table.copy(ctx.path),\n"
"        schema = schema,\n"
"    })\n"
"end\n"
"\n"
"local function schema_pairs_impl(schema, ctx)\n"
"    if is_scalar(schema) then\n"
"        schema_pairs_append_node(schema, ctx)\n"
"    elseif schema.type == 'record' then\n"
"        for k, v in pairs(schema.fields) do\n"
"            walkthrough_enter(ctx, k)\n"
"            schema_pairs_impl(v, ctx)\n"
"            walkthrough_leave(ctx)\n"
"        end\n"
"    elseif schema.type == 'map' then\n"
"        schema_pairs_append_node(schema, ctx)\n"
"    elseif schema.type == 'array' then\n"
"        schema_pairs_append_node(schema, ctx)\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"-- Walk over the schema and return scalar, array and map schema\n"
"-- nodes.\n"
"--\n"
"-- Usage example:\n"
"--\n"
"-- for _, w in schema:pairs() do\n"
"--     local path = w.path\n"
"--     local schema = w.schema\n"
"--     <...>\n"
"-- end\n"
"--\n"
"-- TODO: Rewrite it without collecting a list beforehand.\n"
"function methods.pairs(self)\n"
"    local ctx = walkthrough_start(self, {acc = {}})\n"
"    schema_pairs_impl(rawget(self, 'schema'), ctx)\n"
"    return fun.iter(ctx.acc)\n"
"end\n"
"\n"
"-- }}} <schema object>:pairs()\n"
"\n"
"-- {{{ Schema preprocessing\n"
"\n"
"-- Step down to a schema node.\n"
"--\n"
"-- The function performs annotations tracking.\n"
"--\n"
"-- The annotations are tracked for all the nodes up to the root\n"
"-- forming a stack of annotations. It allows to step down to a\n"
"-- child node using preprocess_enter(), step up to the original\n"
"-- node using preprocess_leave() and step down again into another\n"
"-- child node.\n"
"--\n"
"-- At each given point the top annotation frame represents\n"
"-- annotations merged from the root node down to the given one.\n"
"-- If there are same named annotations on the path, then one from\n"
"-- a descendant node is preferred.\n"
"local function preprocess_enter(ctx, schema)\n"
"    assert(ctx.annotation_stack ~= nil)\n"
"\n"
"    -- These keys are part of the schema node tree structure\n"
"    -- itself. See the 'Schema nodes constructors' section in this\n"
"    -- file for details about the schema node structure.\n"
"    local non_annotation_keys = {\n"
"        type = true,\n"
"        fields = true,\n"
"        key = true,\n"
"        value = true,\n"
"        items = true,\n"
"    }\n"
"\n"
"    -- There are known annotations that barely has any sense in\n"
"    -- context of descendant schema nodes. Don't track them.\n"
"    local ignored_annotations = {\n"
"        allowed_values = true,\n"
"        validate = true,\n"
"        default = true,\n"
"        apply_default_if = true,\n"
"    }\n"
"\n"
"    local frame = table.copy(ctx.annotation_stack[#ctx.annotation_stack] or {})\n"
"    for k, v in pairs(schema) do\n"
"        if not non_annotation_keys[k] and not ignored_annotations[k] then\n"
"            frame[k] = v\n"
"        end\n"
"    end\n"
"    table.insert(ctx.annotation_stack, frame)\n"
"end\n"
"\n"
"-- Step up from a schema node.\n"
"--\n"
"-- Returns the computed fields for the node that we're leaving.\n"
"--\n"
"-- See preprocess_enter() for details.\n"
"local function preprocess_leave(ctx)\n"
"    return {\n"
"        annotations = table.remove(ctx.annotation_stack),\n"
"    }\n"
"end\n"
"\n"
"-- The function prepares the given schema node tree in the\n"
"-- following way.\n"
"--\n"
"-- * The schema node is copied.\n"
"-- * All its descendant nodes are prepared using this algorithm\n"
"--   (recursively).\n"
"-- * Computed fields are calculated and stored in the copied node.\n"
"-- * The copied node is the result.\n"
"--\n"
"-- The sketchy structure of the returned schema node is the\n"
"-- following.\n"
"--\n"
"-- {\n"
"--     <..fields..>\n"
"--     computed = {\n"
"--         annotations = <...>,\n"
"--     },\n"
"-- }\n"
"--\n"
"-- At now there is only one computed field: `annotations`.\n"
"--\n"
"-- The `annotations` field contains all the annotations merged\n"
"-- from the root schema node down to the given one. If the same\n"
"-- annotation is present in an ancestor node and in an descendant\n"
"-- node, the latter is preferred.\n"
"--\n"
"-- Design details\n"
"-- --------------\n"
"--\n"
"-- The copying is performed to decouple schema node definitions\n"
"-- provided by a caller from the schema nodes stored in the schema\n"
"-- object. It allows to modify them without modifying caller's own\n"
"-- objects.\n"
"--\n"
"-- It is especially important if parts of one schema are passed\n"
"-- as schema node definitions to create another schema.\n"
"--\n"
"-- This case also motivates to store all the computed fields in\n"
"-- the `computed` field. This way it is easier to guarantee that\n"
"-- all the computed fields are stripped down from the original\n"
"-- schema nodes and don't affect the new schema anyhow.\n"
"local function preprocess_schema(schema, ctx)\n"
"    -- A schema node from another (already constructed) schema may\n"
"    -- be used in the schema that is currently under construction.\n"
"    --\n"
"    -- Since we're going to modify the schema node, we should copy\n"
"    -- it beforehand. Our modifications must not affect another\n"
"    -- schema.\n"
"    local res = table.copy(schema)\n"
"\n"
"    -- Eliminate previously computed values if any.\n"
"    --\n"
"    -- The past values were computed against node's place in\n"
"    -- another schema, so it must not influence the node in its\n"
"    -- current place in the given schema.\n"
"    --\n"
"    -- This field is rewritten at end of the function, but it\n"
"    -- anyway should be stripped before going to\n"
"    -- preprocess_enter(). Otherwise it would be taken as an\n"
"    -- annotation.\n"
"    --\n"
"    -- (It can be just ignored in preprocess_enter(), but dropping\n"
"    -- it beforehand looks less error-prone.)\n"
"    res.computed = nil\n"
"\n"
"    preprocess_enter(ctx, res)\n"
"\n"
"    -- luacheck: ignore 542 empty if branch\n"
"    if is_scalar(schema) then\n"
"        -- Nothing to do.\n"
"    elseif schema.type == 'record' then\n"
"        local fields = {}\n"
"        for field_name, field_def in pairs(schema.fields) do\n"
"            fields[field_name] = preprocess_schema(field_def, ctx)\n"
"        end\n"
"        res.fields = fields\n"
"    elseif schema.type == 'map' then\n"
"        res.key = preprocess_schema(schema.key, ctx)\n"
"        res.value = preprocess_schema(schema.value, ctx)\n"
"    elseif schema.type == 'array' then\n"
"        res.items = preprocess_schema(schema.items, ctx)\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"\n"
"    res.computed = preprocess_leave(ctx)\n"
"\n"
"    return res\n"
"end\n"
"\n"
"-- }}} Schema preprocessing\n"
"\n"
"-- {{{ Schema object constructor: new\n"
"\n"
"-- Define a field lookup function on a schema object.\n"
"--\n"
"-- `<schema object>.foo` performs the following:\n"
"--\n"
"-- * search for a user-provided method\n"
"-- * search for a method defined in this module\n"
"-- * if 'name', 'schema' or 'methods' -- return the given field\n"
"-- * otherwise return nil\n"
"function schema_mt.__index(self, key)\n"
"    local instance_methods = rawget(self, 'methods')\n"
"    if instance_methods[key] ~= nil then\n"
"        return instance_methods[key]\n"
"    end\n"
"    if methods[key] ~= nil then\n"
"        return methods[key]\n"
"    end\n"
"    return rawget(self, key)\n"
"end\n"
"\n"
"-- Create a schema object.\n"
"--\n"
"-- Unlike a schema node it has a name, has methods defined in this\n"
"-- module and user-provided methods.\n"
"local function new(name, schema, opts)\n"
"    local opts = opts or {}\n"
"    local instance_methods = opts.methods or {}\n"
"\n"
"    assert(type(name) == 'string')\n"
"    assert(type(schema) == 'table')\n"
"\n"
"    local ctx = {\n"
"        annotation_stack = {},\n"
"    }\n"
"    local preprocessed_schema = preprocess_schema(schema, ctx)\n"
"\n"
"    return setmetatable({\n"
"        name = name,\n"
"        schema = preprocessed_schema,\n"
"        methods = instance_methods,\n"
"    }, schema_mt)\n"
"end\n"
"\n"
"-- }}} Schema object constructor: new\n"
"\n"
"-- {{{ schema.fromenv()\n"
"\n"
"-- Forward declaration.\n"
"local fromenv\n"
"\n"
"-- Decode a map value from an environment variable data.\n"
"--\n"
"-- Accepts two formats:\n"
"--\n"
"-- 1. JSON format (if data starts from \"{\").\n"
"-- 2. Simple foo=bar,baz=fiz object format (otherwise).\n"
"--\n"
"-- The simple format is applicable for a map with string keys and\n"
"-- scalar values of any type except 'any'.\n"
"local function map_from_env(env_var_name, raw_value, schema)\n"
"    local can_have_simple_format = schema.key.type == 'string' and\n"
"        is_scalar(schema.value) and schema.value.type ~= 'any'\n"
"\n"
"    local recommendation = can_have_simple_format and\n"
"        'Use either the simple \"foo=bar,baz=fiz\" object format or the JSON ' ..\n"
"        'object format (starts from \"{\").' or\n"
"        'Use the JSON object format (starts from \"{\").'\n"
"\n"
"    -- JSON object -> try to decode.\n"
"    if raw_value:startswith('{') then\n"
"        local ok, res = pcall(json.decode, raw_value)\n"
"        if not ok then\n"
"            error(('Unable to decode JSON data in environment ' ..\n"
"                'variable %q: %s'):format(env_var_name, res), 0)\n"
"        end\n"
"        return res\n"
"    end\n"
"\n"
"    -- JSON array -> error.\n"
"    if raw_value:startswith('[') then\n"
"        error(('A JSON array is provided for environment variable %q of ' ..\n"
"            'type map, an object is expected. %s'):format(env_var_name,\n"
"            recommendation), 0)\n"
"    end\n"
"\n"
"    -- Check several prerequisites for the simple foo=bar,baz=fiz\n"
"    -- object format. If any check is not passed, suggest to use\n"
"    -- the JSON object format and describe why.\n"
"\n"
"    -- Allow only string keys.\n"
"    if schema.key.type ~= 'string' then\n"
"        error(('Use the JSON object format for environment variable %q: the ' ..\n"
"            'keys are supposed to have a non-string type (%q) that is not ' ..\n"
"            'supported by the simple \"foo=bar,baz=fiz\" object format. ' ..\n"
"            'A JSON object value starts from \"{\".'):format(env_var_name,\n"
"            schema.key.type), 0)\n"
"    end\n"
"\n"
"    -- Allow only scalar field values. Forbid composite ones.\n"
"    if not is_scalar(schema.value) then\n"
"        error(('Use the JSON object format for environment variable %q: the ' ..\n"
"            'field values are supposed to have a composite type (%q) that ' ..\n"
"            'is not supported by the simple \"foo=bar,baz=fiz\" object ' ..\n"
"            'format. A JSON object value starts from \"{\".'):format(env_var_name,\n"
"            schema.value.type), 0)\n"
"    end\n"
"\n"
"    -- Forbid scalar field values of the 'any' type.\n"
"    if schema.value.type == 'any' then\n"
"        error(('Use the JSON object format for environment variable %q: the ' ..\n"
"            'field values are supposed to have an arbitrary type (\"any\") ' ..\n"
"            'that is not supported by the simple \"foo=bar,baz=fiz\" object ' ..\n"
"            'format. A JSON object value starts from \"{\".'):format(\n"
"            env_var_name), 0)\n"
"    end\n"
"\n"
"    -- Assume the simple foo=bar,baz=fiz object format.\n"
"    local err_msg_prefix = ('Unable to decode data in environment variable ' ..\n"
"        '%q assuming the simple \"foo=bar,baz=fiz\" object format'):format(\n"
"        env_var_name)\n"
"    local res = {}\n"
"    for _, v in ipairs(raw_value:split(',')) do\n"
"        local eq = v:find('=')\n"
"        if eq == nil then\n"
"            error(('%s: no \"=\" is found in a key-value pair. %s'):format(\n"
"                err_msg_prefix, recommendation), 0)\n"
"        end\n"
"        local lhs = string.sub(v, 1, eq - 1)\n"
"        local rhs = string.sub(v, eq + 1)\n"
"\n"
"        if lhs == '' then\n"
"            error(('%s: no value before \"=\" is found in a key-value pair. ' ..\n"
"                '%s'):format(err_msg_prefix, recommendation), 0)\n"
"        end\n"
"        local subname = ('%s.%s'):format(env_var_name, lhs)\n"
"        res[lhs] = fromenv(subname, rhs, schema.value)\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"-- Decode an array from an environment variable data.\n"
"--\n"
"-- Accepts two formats:\n"
"--\n"
"-- 1. JSON format (if data starts from \"[\").\n"
"-- 2. Simple foo,bar,baz array format (otherwise).\n"
"--\n"
"-- The simple format is applicable for an array with scalar item\n"
"-- values of any type except 'any'.\n"
"local function array_from_env(env_var_name, raw_value, schema)\n"
"    local can_have_simple_format = is_scalar(schema.items) and\n"
"        schema.items.type ~= 'any'\n"
"\n"
"    local recommendation = can_have_simple_format and\n"
"        'Use either the simple \"foo,bar,baz\" array format or the JSON ' ..\n"
"        'array format (starts from \"[\").' or\n"
"        'Use the JSON array format (starts from \"[\").'\n"
"\n"
"    -- JSON array -> try to decode.\n"
"    if raw_value:startswith('[') then\n"
"        local ok, res = pcall(json.decode, raw_value)\n"
"        if not ok then\n"
"            error(('Unable to decode JSON data in environment ' ..\n"
"                'variable %q: %s'):format(env_var_name, res), 0)\n"
"        end\n"
"        return res\n"
"    end\n"
"\n"
"    -- JSON object -> error.\n"
"    if raw_value:startswith('{') then\n"
"        error(('A JSON object is provided for environment variable %q of ' ..\n"
"            'type array, an array is expected. %s'):format(env_var_name,\n"
"            recommendation), 0)\n"
"    end\n"
"\n"
"    -- Check several prerequisites for the simple foo,bar,baz\n"
"    -- array format. If any check is not passed, suggest to use\n"
"    -- the JSON array format and describe why.\n"
"\n"
"    -- Allow only scalar item values. Forbid composite ones.\n"
"    if not is_scalar(schema.items) then\n"
"        error(('Use the JSON array format for environment variable %q: the ' ..\n"
"            'item values are supposed to have a composite type (%q) that ' ..\n"
"            'is not supported by the simple \"foo,bar,baz\" array format. ' ..\n"
"            'A JSON array value starts from \"[\".'):format(env_var_name,\n"
"            schema.items.type), 0)\n"
"    end\n"
"\n"
"    -- Forbid scalar field values of the 'any' type.\n"
"    if schema.items.type == 'any' then\n"
"        error(('Use the JSON array format for environment variable %q: the ' ..\n"
"            'item values are supposed to have an arbitrary type (\"any\") ' ..\n"
"            'that is not supported by the simple \"foo,bar,baz\" array ' ..\n"
"            'format. A JSON array value starts from \"[\".'):format(\n"
"            env_var_name), 0)\n"
"    end\n"
"\n"
"    -- Assume the simple foo,bar,baz array format.\n"
"    local res = {}\n"
"    for i, v in ipairs(raw_value:split(',')) do\n"
"        local subname = ('%s[%d]'):format(env_var_name, i)\n"
"        res[i] = fromenv(subname, v, schema.items)\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"-- Parse data from an environment variable as a value of the given\n"
"-- type.\n"
"--\n"
"-- Important: the result is not necessarily valid against the given\n"
"-- schema node. It should be validated using the\n"
"-- <schema object>:validate() method before further processing.\n"
"fromenv = function(env_var_name, raw_value, schema)\n"
"    if raw_value == nil or raw_value == '' then\n"
"        return nil\n"
"    end\n"
"\n"
"    if is_scalar(schema) then\n"
"        local scalar_def = scalars[schema.type]\n"
"        assert(scalar_def ~= nil)\n"
"        return scalar_def.fromenv(env_var_name, raw_value)\n"
"    elseif schema.type == 'record' then\n"
"        -- TODO: It is technically possible to implement parsing\n"
"        -- of records similarly how it is done for maps, but it is\n"
"        -- not needed for the config module and left unimplemented\n"
"        -- for now.\n"
"        error(('Attempt to parse environment variable %q as a ' ..\n"
"            'record value: this is not supported yet and likely ' ..\n"
"            'caused by an internal error in the config module'):format(\n"
"            env_var_name), 0)\n"
"    elseif schema.type == 'map' then\n"
"        return map_from_env(env_var_name, raw_value, schema)\n"
"    elseif schema.type == 'array' then\n"
"        return array_from_env(env_var_name, raw_value, schema)\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"-- }}} schema.fromenv()\n"
"\n"
"return {\n"
"    -- Schema node constructors.\n"
"    scalar = scalar,\n"
"    record = record,\n"
"    map = map,\n"
"    array = array,\n"
"\n"
"    -- Constructors for 'derived types'.\n"
"    --\n"
"    -- It produces a scalar, record, map or array, but annotates\n"
"    -- it in some specific way to, say, impose extra constraint\n"
"    -- rules at validation.\n"
"    enum = enum,\n"
"    set = set,\n"
"\n"
"    -- Schema object constructor.\n"
"    new = new,\n"
"\n"
"    -- Parse data from an environment variable.\n"
"    fromenv = fromenv,\n"
"}\n"
""
;
