/*
** 2000-05-29
**
** The author disclaims copyright to this source code.  In place of
** a legal notice, here is a blessing:
**
**    May you do good and not evil.
**    May you find forgiveness for yourself and forgive others.
**    May you share freely, never taking more than you give.
**
*************************************************************************
** Driver template for the LEMON parser generator.
**
** The "lemon" program processes an LALR(1) input grammar file, then uses
** this template to construct a parser.  The "lemon" program inserts text
** at each "%%" line.  Also, any "P-a-r-s-e" identifer prefix (without the
** interstitial "-" characters) contained in this template is changed into
** the value of the %name directive from the grammar.  Otherwise, the content
** of this template is copied straight through into the generate parser
** source file.
**
** The following is the concatenation of all %include directives from the
** input grammar file:
*/
#include <stdio.h>
#include <stdbool.h>
/************ Begin %include sections from the grammar ************************/
#line 56 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

#include "sqlInt.h"

/*
** Disable all error recovery processing in the parser push-down
** automaton.
*/
#define YYNOERRORRECOVERY 1

/*
** Indicate that sqlParserFree() will never be called with a null
** pointer.
*/
#define YYPARSEFREENEVERNULL 1

/*
 * Stop the parser if an error occurs. This macro adds an
 * additional check that allows the parser to be stopped if any
 * error was noticed.
 */
#define PARSER_ERROR_CHECK && ! pParse->is_aborted

/*
** An instance of this structure holds information about the
** LIMIT clause of a SELECT statement.
*/
struct LimitVal {
  Expr *pLimit;    /* The LIMIT expression.  NULL if there is no limit */
  Expr *pOffset;   /* The OFFSET expression.  NULL if there is none */
};

/*
** An instance of the following structure describes the event of a
** TRIGGER.  "a" is the event type, one of TK_UPDATE, TK_INSERT,
** TK_DELETE, or TK_INSTEAD.  If the event is of the form
**
**      UPDATE ON (a,b,c)
**
** Then the "b" IdList records the list "a,b,c".
*/
struct TrigEvent { int a; IdList * b; };

/*
** Disable lookaside memory allocation for objects that might be
** shared across database connections.
*/
static void disableLookaside(Parse *pParse){
  pParse->disableLookaside++;
  sql_get()->lookaside.bDisable++;
}

#line 437 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  /**
   * For a compound SELECT statement, make sure
   * p->pPrior->pNext==p for all elements in the list. And make
   * sure list length does not exceed SQL_MAX_COMPOUND_SELECT.
   */
  static void parserDoubleLinkSelect(Parse *pParse, Select *p){
    if( p->pPrior ){
      Select *pNext = 0, *pLoop;
      int mxSelect, cnt = 0;
      for(pLoop=p; pLoop; pNext=pLoop, pLoop=pLoop->pPrior, cnt++){
        pLoop->pNext = pNext;
        pLoop->selFlags |= SF_Compound;
      }
      if((p->selFlags & SF_MultiValue) == 0 &&
         (mxSelect = sql_get()->aLimit[SQL_LIMIT_COMPOUND_SELECT]) > 0 &&
         cnt > mxSelect) {
          diag_set(ClientError, ER_SQL_PARSER_LIMIT, "The number of UNION or "\
                   "EXCEPT or INTERSECT operations", cnt,
                   sql_get()->aLimit[SQL_LIMIT_COMPOUND_SELECT]);
         pParse->is_aborted = true;
      }
    }
  }
#line 899 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  /* This is a utility routine used to set the ExprSpan.zStart and
  ** ExprSpan.zEnd values of pOut so that the span covers the complete
  ** range of text beginning with pStart and going to the end of pEnd.
  */
  static void spanSet(ExprSpan *pOut, Token *pStart, Token *pEnd){
    pOut->zStart = pStart->z;
    pOut->zEnd = &pEnd->z[pEnd->n];
  }

  /* Construct a new Expr object from a single identifier.  Use the
  ** new Expr to populate pOut.  Set the span of pOut to be the identifier
  ** that created the expression.
  */
  static void spanExpr(struct ExprSpan *pOut, int op, Token t){
    struct Expr *p = NULL;
    int name_sz = t.n + 1;
    p = sql_xmalloc(sizeof(Expr) + name_sz);
    memset(p, 0, sizeof(Expr));
    switch (op) {
    case TK_STRING:
      p->type = FIELD_TYPE_STRING;
      break;
    case TK_BLOB:
      p->type = FIELD_TYPE_VARBINARY;
      break;
    case TK_INTEGER:
      p->type = FIELD_TYPE_INTEGER;
      break;
    case TK_FLOAT:
      p->type = FIELD_TYPE_DOUBLE;
      break;
    case TK_DECIMAL:
      p->type = FIELD_TYPE_DECIMAL;
      break;
    case TK_TRUE:
    case TK_FALSE:
    case TK_UNKNOWN:
      p->type = FIELD_TYPE_BOOLEAN;
      break;
    default:
      p->type = FIELD_TYPE_SCALAR;
      break;
    }
    p->op = (u8)op;
    p->flags = EP_Leaf;
    p->iAgg = -1;
    p->u.zToken = (char*)&p[1];
    memcpy(p->u.zToken, t.z, t.n);
    p->u.zToken[t.n] = '\0';
    sqlDequote(p->u.zToken);
    if (op == TK_ID || op == TK_COLLATE || op == TK_FUNCTION)
      p->flags |= t.z[0] != '"' ? EP_Lookup2 : 0;
#if SQL_MAX_EXPR_DEPTH>0
    p->nHeight = 1;
#endif  
    pOut->pExpr = p;
    pOut->zStart = t.z;
    pOut->zEnd = &t.z[t.n];
    return;
  }
#line 1162 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  /* This routine constructs a binary expression node out of two ExprSpan
  ** objects and uses the result to populate a new ExprSpan object.
  */
  static void spanBinaryExpr(
    Parse *pParse,      /* The parsing context.  Errors accumulate here */
    int op,             /* The binary operation */
    ExprSpan *pLeft,    /* The left operand, and output */
    ExprSpan *pRight    /* The right operand */
  ){
    pLeft->pExpr = sqlPExpr(pParse, op, pLeft->pExpr, pRight->pExpr);
    pLeft->zEnd = pRight->zEnd;
  }

  /* If doNot is true, then add a TK_NOT Expr-node wrapper around the
  ** outside of *ppExpr.
  */
  static void exprNot(Parse *pParse, int doNot, ExprSpan *pSpan){
    if( doNot ){
      pSpan->pExpr = sqlPExpr(pParse, TK_NOT, pSpan->pExpr, 0);
    }
  }
#line 1234 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  /* Construct an expression node for a unary postfix operator
  */
  static void spanUnaryPostfix(
    Parse *pParse,         /* Parsing context to record errors */
    int op,                /* The operator */
    ExprSpan *pOperand,    /* The operand, and output */
    Token *pPostOp         /* The operand token for setting the span */
  ){
    pOperand->pExpr = sqlPExpr(pParse, op, pOperand->pExpr, 0);
    pOperand->zEnd = &pPostOp->z[pPostOp->n];
  }                           
#line 1255 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  /* Construct an expression node for a unary prefix operator
  */
  static void spanUnaryPrefix(
    ExprSpan *pOut,        /* Write the new expression node here */
    Parse *pParse,         /* Parsing context to record errors */
    int op,                /* The operator */
    ExprSpan *pOperand,    /* The operand */
    Token *pPreOp         /* The operand token for setting the span */
  ){
    pOut->zStart = pPreOp->z;
    pOut->pExpr = sqlPExpr(pParse, op, pOperand->pExpr, 0);
    pOut->zEnd = pOperand->zEnd;
  }
#line 1446 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  /* Add a single new term to an ExprList that is used to store a
  ** list of identifiers.  Report an error if the ID list contains
  ** a COLLATE clause or an ASC or DESC keyword, except ignore the
  ** error while parsing a legacy schema.
  */
  static ExprList *parserAddExprIdListTerm(
    Parse *pParse,
    ExprList *pPrior,
    Token *pIdToken
  ){
    ExprList *p = sql_expr_list_append(pPrior, NULL);
    sqlExprListSetName(pParse, p, pIdToken, 1);
    return p;
  }
#line 1665 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  struct alter_args {
    struct SrcList *table_name;
    /** Name of constraint OR new name of table in case of RENAME. */
    struct Token name;
  };
#line 242 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
/**************** End of %include directives **********************************/
/* These constants specify the various numeric values for terminal symbols
** in a format understandable to "makeheaders".  This section is blank unless
** "lemon" is run with the "-m" command-line option.
***************** Begin makeheaders token definitions *************************/
/**************** End makeheaders token definitions ***************************/

/* The next sections is a series of control #defines.
** various aspects of the generated parser.
**    YYCODETYPE         is the data type used to store the integer codes
**                       that represent terminal and non-terminal symbols.
**                       "unsigned char" is used if there are fewer than
**                       256 symbols.  Larger types otherwise.
**    YYNOCODE           is a number of type YYCODETYPE that is not used for
**                       any terminal or nonterminal symbol.
**    YYFALLBACK         If defined, this indicates that one or more tokens
**                       (also known as: "terminal symbols") have fall-back
**                       values which should be used if the original symbol
**                       would not parse.  This permits keywords to sometimes
**                       be used as identifiers, for example.
**    YYACTIONTYPE       is the data type used for "action codes" - numbers
**                       that indicate what to do in response to the next
**                       token.
**    sqlParserTOKENTYPE     is the data type used for minor type for terminal
**                       symbols.  Background: A "minor type" is a semantic
**                       value associated with a terminal or non-terminal
**                       symbols.  For example, for an "ID" terminal symbol,
**                       the minor type might be the name of the identifier.
**                       Each non-terminal can have a different minor type.
**                       Terminal symbols all have the same minor type, though.
**                       This macros defines the minor type for terminal 
**                       symbols.
**    YYMINORTYPE        is the data type used for all minor types.
**                       This is typically a union of many types, one of
**                       which is sqlParserTOKENTYPE.  The entry in the union
**                       for terminal symbols is called "yy0".
**    YYSTACKDEPTH       is the maximum depth of the parser's stack.  If
**                       zero the stack is dynamically sized using realloc()
**    sqlParserARG_SDECL     A static variable declaration for the %extra_argument
**    sqlParserARG_PDECL     A parameter declaration for the %extra_argument
**    sqlParserARG_STORE     Code to store %extra_argument into yypParser
**    sqlParserARG_FETCH     Code to extract %extra_argument from yypParser
**    YYERRORSYMBOL      is the code number of the error symbol.  If not
**                       defined, then do no error processing.
**    YYNSTATE           the combined number of states.
**    YYNRULE            the number of rules in the grammar
**    YY_MAX_SHIFT       Maximum value for shift actions
**    YY_MIN_SHIFTREDUCE Minimum value for shift-reduce actions
**    YY_MAX_SHIFTREDUCE Maximum value for shift-reduce actions
**    YY_MIN_REDUCE      Maximum value for reduce actions
**    YY_ERROR_ACTION    The yy_action[] code for syntax error
**    YY_ACCEPT_ACTION   The yy_action[] code for accept
**    YY_NO_ACTION       The yy_action[] code for no-op
*/
#ifndef INTERFACE
# define INTERFACE 1
#endif
/************* Begin control #defines *****************************************/
#define YYCODETYPE unsigned short int
#define YYNOCODE 275
#define YYACTIONTYPE unsigned short int
#define YYWILDCARD 78
#define sqlParserTOKENTYPE Token
typedef union {
  int yyinit;
  sqlParserTOKENTYPE yy0;
  ExprSpan yy34;
  int yy40;
  struct Token yy103;
  TriggerStep* yy115;
  struct TrigEvent yy122;
  struct SrcList * yy140;
  With* yy163;
  struct ExprList * yy211;
  ExprList * yy222;
  struct type_def yy226;
  struct Expr * yy231;
  IdList* yy240;
  Select* yy247;
  struct LimitVal yy248;
  Expr* yy262;
  enum trim_side_mask yy345;
  ExprList* yy414;
  struct alter_args yy446;
  SrcList* yy499;
} YYMINORTYPE;
#ifndef YYSTACKDEPTH
#define YYSTACKDEPTH 100
#endif
#define sqlParserARG_SDECL Parse *pParse;
#define sqlParserARG_PDECL ,Parse *pParse
#define sqlParserARG_FETCH Parse *pParse = yypParser->pParse
#define sqlParserARG_STORE yypParser->pParse = pParse
#define YYFALLBACK 1
#define YYNSTATE             449
#define YYNRULE              320
#define YY_MAX_SHIFT         448
#define YY_MIN_SHIFTREDUCE   666
#define YY_MAX_SHIFTREDUCE   985
#define YY_MIN_REDUCE        986
#define YY_MAX_REDUCE        1305
#define YY_ERROR_ACTION      1306
#define YY_ACCEPT_ACTION     1307
#define YY_NO_ACTION         1308
/************* End control #defines *******************************************/

/* Define the yytestcase() macro to be a no-op if is not already defined
** otherwise.
**
** Applications can choose to define yytestcase() in the %include section
** to a macro that can assist in verifying code coverage.  For production
** code the yytestcase() macro should be turned off.  But it is useful
** for testing.
*/
#ifndef yytestcase
# define yytestcase(X)
#endif


/* Next are the tables used to determine what action to take based on the
** current state and lookahead token.  These tables are used to implement
** functions that take a state number and lookahead value and return an
** action integer.  
**
** Suppose the action integer is N.  Then the action is determined as
** follows
**
**   0 <= N <= YY_MAX_SHIFT             Shift N.  That is, push the lookahead
**                                      token onto the stack and goto state N.
**
**   N between YY_MIN_SHIFTREDUCE       Shift to an arbitrary state then
**     and YY_MAX_SHIFTREDUCE           reduce by rule N-YY_MIN_SHIFTREDUCE.
**
**   N between YY_MIN_REDUCE            Reduce by rule N-YY_MIN_REDUCE
**     and YY_MAX_REDUCE
**
**   N == YY_ERROR_ACTION               A syntax error has occurred.
**
**   N == YY_ACCEPT_ACTION              The parser accepts its input.
**
**   N == YY_NO_ACTION                  No such action.  Denotes unused
**                                      slots in the yy_action[] table.
**
** The action table is constructed as a single large table named yy_action[].
** Given state S and lookahead X, the action is computed as either:
**
**    (A)   N = yy_action[ yy_shift_ofst[S] + X ]
**    (B)   N = yy_default[S]
**
** The (A) formula is preferred.  The B formula is used instead if:
**    (1)  The yy_shift_ofst[S]+X value is out of range, or
**    (2)  yy_lookahead[yy_shift_ofst[S]+X] is not equal to X, or
**    (3)  yy_shift_ofst[S] equal YY_SHIFT_USE_DFLT.
** (Implementation note: YY_SHIFT_USE_DFLT is chosen so that
** YY_SHIFT_USE_DFLT+X will be out of range for all possible lookaheads X.
** Hence only tests (1) and (2) need to be evaluated.)
**
** The formulas above are for computing the action when the lookahead is
** a terminal symbol.  If the lookahead is a non-terminal (as occurs after
** a reduce action) then the yy_reduce_ofst[] array is used in place of
** the yy_shift_ofst[] array and YY_REDUCE_USE_DFLT is used in place of
** YY_SHIFT_USE_DFLT.
**
** The following are the tables generated in this section:
**
**  yy_action[]        A single table containing all actions.
**  yy_lookahead[]     A table containing the lookahead for each entry in
**                     yy_action.  Used to detect hash collisions.
**  yy_shift_ofst[]    For each state, the offset into yy_action for
**                     shifting terminals.
**  yy_reduce_ofst[]   For each state, the offset into yy_action for
**                     shifting non-terminals after a reduce.
**  yy_default[]       Default action for each state.
**
*********** Begin parsing tables **********************************************/
#define YY_ACTTAB_COUNT (1871)
static const YYACTIONTYPE yy_action[] = {
 /*     0 */    54,   65,   65,   65,   65,   64,   64,   63,   63,   63,
 /*    10 */    62,  321,  321,   20,   20,  447,  447,   51,   52,  292,
 /*    20 */  1307,  448,    1,  148,   53,   10,  151,  446,  442,  670,
 /*    30 */   445,  247,  293,  872,  151,  676,  432,    3,  202,  202,
 /*    40 */   151,  794,  780,  322,  322,  135,  119,  119,  119,    6,
 /*    50 */   396,  714,  984,  973,   81,  973,  682,  967,  301,  305,
 /*    60 */   140,  411,  421,  913,  135,  347,  973,  323,  973,  323,
 /*    70 */   178,  340,  296, 1017,  789,  416,  412,  300,   20,  386,
 /*    80 */   442,  202,  202,  299, 1248, 1248,   64,   64,   63,   63,
 /*    90 */    63,   62,  321,  396,   20,  791,  235,  434,   96,   96,
 /*   100 */    96,  686,  316,  686,  306,  185,  139,  345,  295,  235,
 /*   110 */   344,  698,  793,  795,  796,  797,  798,  799,  800,  801,
 /*   120 */   303,  303,  440,   54,    9,  140,  435,  821,  822,  823,
 /*   130 */   979,   19,  794,  759,  962,  182,  282,  296,  213,   25,
 /*   140 */    51,   52,  300,  913,  414,  413,  338,   53,   10,  434,
 /*   150 */   397,  245,  142,  271,  899,   54,  442,  135,  698,  432,
 /*   160 */     3, 1181,  151,  442,  794,  789,  322,  322,   75,   75,
 /*   170 */    75,  681,   51,   52,   95,   95,   95,  666,  325,   53,
 /*   180 */    10,  119,  119,  119,  392,  421,  719,  719,   61,   59,
 /*   190 */   201,  432,    3,  327,  151,  442,  794,  789,  322,  322,
 /*   200 */   904,  736,  758,  793,  795,  796,  797,  798,  799,  800,
 /*   210 */   416,  394,  439,   76,   76,   76,  905,  421,  791,  434,
 /*   220 */   304,  151,  415,   23,  234,  434,  906,  707,  430,  789,
 /*   230 */    62,  321,  434,   20,  708,  793,  795,  796,  797,  798,
 /*   240 */   799,  800,  801,  303,  303,  440,  743,    9,  787,  435,
 /*   250 */   791,  681,  912,  979,   19,  346,  783,   61,   59,  201,
 /*   260 */   195,  194,  713,  901,  434,  135,  135,  793,  795,  796,
 /*   270 */   797,  798,  799,  800,  801,  303,  303,  440,   54,    9,
 /*   280 */    72,  435,  244,  393,  794,  979,   19,  874,  757,  874,
 /*   290 */   727,  360,  439,  386,  362,   51,   52,  728,  349,  189,
 /*   300 */   743,  251,   53,   10,  263,  441,  442,  353,  151,  442,
 /*   310 */    61,   59,  201,  369,  432,    3,  316,  789,  403,  794,
 /*   320 */   756,  322,  322,  881,   90,   90,   90,   76,   76,   76,
 /*   330 */   337, 1285,  783,  188,  312, 1285,   61,   59,  201,  390,
 /*   340 */   421,  910,  135,   63,   63,   63,   62,  321,  388,   20,
 /*   350 */   244,  393,  789,  391,  215,  793,  795,  796,  797,  798,
 /*   360 */   799,  800,  908,  442,  151,  245,  142,  271,   61,   59,
 /*   370 */   201,  362, 1285,  791,  359,  434,  193,  262,  434,  840,
 /*   380 */   263,  119,  119,  119,  326,  917,  328,  442,  783,  331,
 /*   390 */   793,  795,  796,  797,  798,  799,  800,  801,  303,  303,
 /*   400 */   440,   54,    9,  442,  435,  119,  119,  119,  979,   19,
 /*   410 */   416,  399,  383,  381,  380,  724,  282,  379,   51,   52,
 /*   420 */   723,  119,  119,  119,  901,   53,   10,  135,  419,  442,
 /*   430 */   359,  221,  434,  262,  416,  410,  263,  432,    3,  742,
 /*   440 */   442,   36,  794,  787,  322,  322,  716,  119,  119,  119,
 /*   450 */   416,  433,  360,  400,  974,  974,  434,  396,  119,  119,
 /*   460 */   119,  442,  267,  421,  964,  334,  923,  924,  682,   82,
 /*   470 */   335,  772,  434,  872,  358,  789,  416,  436,  168,   76,
 /*   480 */    76,   76,  208,  169,  683,  683,  314,  416,  389,  683,
 /*   490 */   683,  398,  967,  301,  135,   43,  791,  401,  434,  313,
 /*   500 */   683,  683,  309,  683,  683,  867,  975,  745,  426,  434,
 /*   510 */   186,  237,  266,  793,  795,  796,  797,  798,  799,  800,
 /*   520 */   801,  303,  303,  440,   50,    9,   48,  435,  368,  207,
 /*   530 */   434,  979,   19,   68,   69,  302,  320,  836,  836,  846,
 /*   540 */   849,   66,   66,   67,   67,   67,   67,   54,   65,   65,
 /*   550 */    65,   65,   64,   64,   63,   63,   63,   62,  321,  198,
 /*   560 */    20,  442,  357,  743,   51,   52,  368,  683,  683,  259,
 /*   570 */   368,   53,   10,  974,  974,  442,  973,  243,  973,   78,
 /*   580 */    78,   78,  442,  432,    3,  315,  442,   22,  794,   83,
 /*   590 */   322,  322,    2,  119,  119,  119,  245,  904,  683,  683,
 /*   600 */   119,  119,  119,  206,   76,   76,   76,  173,  884,  421,
 /*   610 */   883,  406,  220,  905,  706,  706,  189,  743,  352,  235,
 /*   620 */   191,  789,  317,  906,  420,  975,  744,  364,  729,  404,
 /*   630 */   434,  385,  693,  384,  230,  837,  837,  847,  850,   61,
 /*   640 */    59,  201,  791,  887,  434,  732,  354,  373,  370,  887,
 /*   650 */   261,  434,  683,  683,  223,  434,  798,  799,  800,  793,
 /*   660 */   795,  796,  797,  798,  799,  800,  801,  303,  303,  440,
 /*   670 */   689,    9,  963,  435,  704,  704,  682,  979,   19,   68,
 /*   680 */    69,  302,  320,  836,  836,  846,  849,   66,   66,   67,
 /*   690 */    67,   67,   67,  688,   65,   65,   65,   65,   64,   64,
 /*   700 */    63,   63,   63,   62,  321,  792,   20,   68,   69,  302,
 /*   710 */   320,  836,  836,  846,  849,   66,   66,   67,   67,   67,
 /*   720 */    67,  256,   65,   65,   65,   65,   64,   64,   63,   63,
 /*   730 */    63,   62,  321,  742,   20,  683,  683,   68,   69,  302,
 /*   740 */   320,  836,  836,  846,  849,   66,   66,   67,   67,   67,
 /*   750 */    67,   39,   65,   65,   65,   65,   64,   64,   63,   63,
 /*   760 */    63,   62,  321,  787,   20,  250,  787,  200,  237,  423,
 /*   770 */    70,  255,  135,  724,  427,  203,  431,  135,  723,  135,
 /*   780 */   366,  760,  761,  742,  742,  742,  280,   68,   69,  302,
 /*   790 */   320,  836,  836,  846,  849,   66,   66,   67,   67,   67,
 /*   800 */    67,  442,   65,   65,   65,   65,   64,   64,   63,   63,
 /*   810 */    63,   62,  321,  787,   20,  216,  217,  907,  698,  116,
 /*   820 */   116,  116,  310,  254,  367,  311,  972,  915,  253,  742,
 /*   830 */  1003,  742,  742,   68,   69,  302,  320,  836,  836,  846,
 /*   840 */   849,   66,   66,   67,   67,   67,   67,  442,   65,   65,
 /*   850 */    65,   65,   64,   64,   63,   63,   63,   62,  321,  927,
 /*   860 */    20,  873,  333,  710,  709,  121,  121,  121,  273,  281,
 /*   870 */   434,  275,  318,  701,  277,  742,  760,  761,  183,   68,
 /*   880 */    69,  302,  320,  836,  836,  846,  849,   66,   66,   67,
 /*   890 */    67,   67,   67,  442,   65,   65,   65,   65,   64,   64,
 /*   900 */    63,   63,   63,   62,  321,  147,   20,  258,  175,  876,
 /*   910 */   409,   92,   92,   92,   11,  249,  434,  919,  916,  694,
 /*   920 */   696,  696,   38,  332,  911,   68,   69,  302,  320,  836,
 /*   930 */   836,  846,  849,   66,   66,   67,   67,   67,   67,  442,
 /*   940 */    65,   65,   65,   65,   64,   64,   63,   63,   63,   62,
 /*   950 */   321,  785,   20,  804,  804,  233,  329,   85,   85,   85,
 /*   960 */   239,  775,  434,  211,  361,  687,  721,  909,  233,  307,
 /*   970 */    45,   68,   69,  302,  320,  836,  836,  846,  849,   66,
 /*   980 */    66,   67,   67,   67,   67,  442,   65,   65,   65,   65,
 /*   990 */    64,   64,   63,   63,   63,   62,  321,  363,   20,  298,
 /*  1000 */   879,  233,  866,   98,   98,   98,   38,  260,  434,  377,
 /*  1010 */   700,  827,  752,  239,   38,  341,  233,   68,   69,  302,
 /*  1020 */   320,  836,  836,  846,  849,   66,   66,   67,   67,   67,
 /*  1030 */    67,  308,   65,   65,   65,   65,   64,   64,   63,   63,
 /*  1040 */    63,   62,  321,  699,   20,  342,  343,   32,  870,  959,
 /*  1050 */   348,  252,  239,   74,  434,  350,  192,  788,  898,  896,
 /*  1060 */   365,  895,  371,   68,   69,  302,  320,  836,  836,  846,
 /*  1070 */   849,   66,   66,   67,   67,   67,   67,  372,   65,   65,
 /*  1080 */    65,   65,   64,   64,   63,   63,   63,   62,  321,  265,
 /*  1090 */    20,   68,   69,  302,  320,  836,  836,  846,  849,   66,
 /*  1100 */    66,   67,   67,   67,   67,  675,   65,   65,   65,   65,
 /*  1110 */    64,   64,   63,   63,   63,   62,  321,  268,   20,   68,
 /*  1120 */    69,  302,  320,  836,  836,  846,  849,   66,   66,   67,
 /*  1130 */    67,   67,   67,  269,   65,   65,   65,   65,   64,   64,
 /*  1140 */    63,   63,   63,   62,  321,  692,   20,   68,   69,  302,
 /*  1150 */   320,  836,  836,  846,  849,   66,   66,   67,   67,   67,
 /*  1160 */    67,  750,   65,   65,   65,   65,   64,   64,   63,   63,
 /*  1170 */    63,   62,  321,  786,   20,  802,  802,  279,  730,  187,
 /*  1180 */   236,  422,  289,  174,  871,   41,  283,   44,    8,  184,
 /*  1190 */   684,   68,   69,  302,  320,  836,  836,  846,  849,   66,
 /*  1200 */    66,   67,   67,   67,   67,  673,   65,   65,   65,   65,
 /*  1210 */    64,   64,   63,   63,   63,   62,  321,  672,   20,   68,
 /*  1220 */    69,  302,  320,  836,  836,  846,  849,   66,   66,   67,
 /*  1230 */    67,   67,   67,  674,   65,   65,   65,   65,   64,   64,
 /*  1240 */    63,   63,   63,   62,  321,  872,   20,  914,  166,  155,
 /*  1250 */   803,  145,   36,  157,   46,   67,   67,   67,   67,   60,
 /*  1260 */    65,   65,   65,   65,   64,   64,   63,   63,   63,   62,
 /*  1270 */   321,  319,   20,  771,  351,  159,   26,  218,  161,  162,
 /*  1280 */   882,  356,   49,  257,  376,  219,  170,   55,   68,   69,
 /*  1290 */   302,  320,  836,  836,  846,  849,   66,   66,   67,   67,
 /*  1300 */    67,   67,  900,   65,   65,   65,   65,   64,   64,   63,
 /*  1310 */    63,   63,   62,  321,  176,   20,   69,  302,  320,  836,
 /*  1320 */   836,  846,  849,   66,   66,   67,   67,   67,   67,  782,
 /*  1330 */    65,   65,   65,   65,   64,   64,   63,   63,   63,   62,
 /*  1340 */   321,   56,   20,  302,  320,  836,  836,  846,  849,   66,
 /*  1350 */    66,   67,   67,   67,   67,  226,   65,   65,   65,   65,
 /*  1360 */    64,   64,   63,   63,   63,   62,  321,  374,   20,  227,
 /*  1370 */   228,  702,  387,  695,  141,  691,  690,  402,   35,  740,
 /*  1380 */    68,   69,  302,  320,  836,  836,  846,  849,   66,   66,
 /*  1390 */    67,   67,   67,   67,  939,   65,   65,   65,   65,   64,
 /*  1400 */    64,   63,   63,   63,   62,  321,    7,   20,   68,   58,
 /*  1410 */   302,  320,  836,  836,  846,  849,   66,   66,   67,   67,
 /*  1420 */    67,   67,  272,   65,   65,   65,   65,   64,   64,   63,
 /*  1430 */    63,   63,   62,  321,  741,   20,  274,   47,  739,  738,
 /*  1440 */   276,  278,  951,   67,   67,   67,   67,  442,   65,   65,
 /*  1450 */    65,   65,   64,   64,   63,   63,   63,   62,  321,  854,
 /*  1460 */    20,  238,  425,  429,  438,   99,   99,   99,  946,   71,
 /*  1470 */   407,  720,  442,  933,  934,  935,  936,  937,  938,  940,
 /*  1480 */   941,  942,  943,  944,  284,  947,  948,  949,  949,  950,
 /*  1490 */   100,  100,  100,  442,  285,   37,  286,  405,  287,  442,
 /*  1500 */   288,  290,  442,  444,  960,  442,  297,   21,  442,  291,
 /*  1510 */   956,  101,  101,  101,  442,  246,  434,  102,  102,  102,
 /*  1520 */   103,  103,  103,   86,   86,   86,   79,   79,   79,  442,
 /*  1530 */   248,   80,   87,   87,   87,  144,  442,  958,  668,  442,
 /*  1540 */   324,  434,  918,   12,  204,  442,   24,   89,   89,   89,
 /*  1550 */   152,  330,  205,  926,  104,  104,  104,  105,  105,  105,
 /*  1560 */   442,  209,  434,  106,  106,  106,  442,  336,  434,  442,
 /*  1570 */   922,  434,  442,  153,  434,  442,  921,  434,   77,   77,
 /*  1580 */    77,  442,  210,  434,  107,  107,  107,  108,  108,  108,
 /*  1590 */   109,  109,  109,   91,   91,   91,  877,  442,  434,  110,
 /*  1600 */   110,  110,  154,  442,  339,  434,  442,  143,  434,  212,
 /*  1610 */   214,  781,  442,  156,  434,  111,  111,  111,  158,  160,
 /*  1620 */   149,  112,  112,  112,  113,  113,  113,  442,  885,  434,
 /*  1630 */    93,   93,   93,  442,  981,  434,  442,  163,  434,  711,
 /*  1640 */   442,  434,  355,  894,  434,  114,  114,  114,  233,  167,
 /*  1650 */   434,  115,  115,  115,   94,   94,   94,  164,  136,  136,
 /*  1660 */   136,  165,   27,   28,  442,   29,  434,  442,   30,  150,
 /*  1670 */   442,  897,  434,  442,  222,  434,  224,  893,   13,  442,
 /*  1680 */   225,  434,  137,  137,  137,  138,  138,  138,  124,  124,
 /*  1690 */   124,  117,  117,  117,  442,   73,  434,  118,  118,  118,
 /*  1700 */   442,  264,  434,  442,  177,  434,  442,  678,  375,  434,
 /*  1710 */   229,  171,   97,   97,   97,   31,   14,  378,  127,  127,
 /*  1720 */   127,  120,  120,  120,  128,  128,  128,  442,   15,  382,
 /*  1730 */    33,  718,  231,  434,  442,  826,  434,  442,  270,  434,
 /*  1740 */   442,  146,  434,  824,  852,  126,  126,  126,  434,   34,
 /*  1750 */    16,    5,  134,  134,  134,  133,  133,  133,  132,  132,
 /*  1760 */   132,  442,  722,  434,  442,  395,  197,  442,  196,  434,
 /*  1770 */   442,  199,  434,  442,  751,  434,  172,  232,  746,  131,
 /*  1780 */   131,  131,  129,  129,  129,  130,  130,  130,  123,  123,
 /*  1790 */   123,  125,  125,  125,  442,   45,  434,  442,   17,   38,
 /*  1800 */    18,  442,   40,  434,  408,  865,  434,  853,  851,  434,
 /*  1810 */    42,  810,  122,  122,  122,   84,   84,   84,  814,   88,
 /*  1820 */    88,   88,  903,  417,  902,  418,  190,  856,  931,  179,
 /*  1830 */   434,  424,  932,  434,  180,  240,  434,  855,  428,  434,
 /*  1840 */   181,  241,  434,  825,    4,  809,  945,  437,  805,   57,
 /*  1850 */   841,  242,  679,  443,  294,  669,  953, 1272,  988,  988,
 /*  1860 */   988,  988,  988,  434,  988,  988,  434,  988,  988,  988,
 /*  1870 */   434,
};
static const YYCODETYPE yy_lookahead[] = {
 /*     0 */     7,   20,   21,   22,   23,   24,   25,   26,   27,   28,
 /*    10 */    29,   30,   30,   32,   32,  176,  177,   24,   25,  180,
 /*    20 */   173,  174,  175,  184,   31,   32,   46,   33,  179,   35,
 /*    30 */    36,   37,   38,  179,   46,   41,   43,   44,  213,  214,
 /*    40 */    46,   48,  212,   50,   51,  215,  197,  198,  199,   44,
 /*    50 */   225,  212,  190,   48,  215,   50,  194,  195,  196,  205,
 /*    60 */   179,  212,   69,  193,  215,  240,   48,  186,   50,  188,
 /*    70 */   189,   77,  191,    1,   81,  226,  227,  196,   32,    7,
 /*    80 */   179,  213,  214,   89,   96,   97,   24,   25,   26,   27,
 /*    90 */    28,   29,   30,  225,   32,  102,   91,  248,  197,  198,
 /*   100 */   199,  197,   30,  199,  265,  266,  267,  113,  269,   91,
 /*   110 */   116,   86,  119,  120,  121,  122,  123,  124,  125,  126,
 /*   120 */   127,  128,  129,    7,  131,  179,  133,  134,  135,  136,
 /*   130 */   137,  138,   48,  194,  188,  189,  179,  191,  144,  145,
 /*   140 */    24,   25,  196,  273,  243,  244,  152,   31,   32,  248,
 /*   150 */   179,   79,   80,   81,  212,    7,  179,  215,   86,   43,
 /*   160 */    44,   45,   46,  179,   48,   81,   50,   51,  197,  198,
 /*   170 */   199,  193,   24,   25,  197,  198,  199,    1,    2,   31,
 /*   180 */    32,  197,  198,  199,   92,   69,   94,   95,  249,  250,
 /*   190 */   251,   43,   44,  270,   46,  179,   48,   81,   50,   51,
 /*   200 */    36,  233,  194,  119,  120,  121,  122,  123,  124,  125,
 /*   210 */   226,  227,  193,  197,  198,  199,   52,   69,  102,  248,
 /*   220 */   204,   46,  245,  246,  229,  248,   62,   63,  271,   81,
 /*   230 */    29,   30,  248,   32,   70,  119,  120,  121,  122,  123,
 /*   240 */   124,  125,  126,  127,  128,  129,   49,  131,  179,  133,
 /*   250 */   102,  273,  179,  137,  138,   80,   70,  249,  250,  251,
 /*   260 */    24,   25,  212,  212,  248,  215,  215,  119,  120,  121,
 /*   270 */   122,  123,  124,  125,  126,  127,  128,  129,    7,  131,
 /*   280 */    44,  133,   96,   97,   48,  137,  138,  197,  201,  199,
 /*   290 */   229,  240,  273,    7,  179,   24,   25,   26,  112,  102,
 /*   300 */   103,  115,   31,   32,  118,    7,  179,  238,   46,  179,
 /*   310 */   249,  250,  251,  262,   43,   44,   30,   81,  237,   48,
 /*   320 */   201,   50,   51,   61,  197,  198,  199,  197,  198,  199,
 /*   330 */   179,   45,   70,  237,  204,   49,  249,  250,  251,  212,
 /*   340 */    69,  268,  215,   26,   27,   28,   29,   30,   26,   32,
 /*   350 */    96,   97,   81,  226,  239,  119,  120,  121,  122,  123,
 /*   360 */   124,  125,  211,  179,   46,   79,   80,   81,  249,  250,
 /*   370 */   251,  179,   86,  102,  112,  248,   49,  115,  248,   81,
 /*   380 */   118,  197,  198,  199,   82,   83,   84,  179,   70,   87,
 /*   390 */   119,  120,  121,  122,  123,  124,  125,  126,  127,  128,
 /*   400 */   129,    7,  131,  179,  133,  197,  198,  199,  137,  138,
 /*   410 */   226,  227,   82,   83,   84,   93,  179,   87,   24,   25,
 /*   420 */    98,  197,  198,  199,  212,   31,   32,  215,  209,  179,
 /*   430 */   112,  239,  248,  115,  226,  227,  118,   43,   44,  179,
 /*   440 */   179,  114,   48,  179,   50,   51,  214,  197,  198,  199,
 /*   450 */   226,  227,  240,    7,   50,   51,  248,  225,  197,  198,
 /*   460 */   199,  179,   40,   69,  190,   82,   83,   84,  194,  257,
 /*   470 */    87,  211,  248,  179,  262,   81,  226,  227,   44,  197,
 /*   480 */   198,  199,   99,   44,   50,   51,  204,  226,  227,   50,
 /*   490 */    51,  212,  195,  196,  215,  101,  102,   51,  248,  205,
 /*   500 */    50,   51,  238,   50,   51,   83,  102,  103,  271,  248,
 /*   510 */   231,  232,   90,  119,  120,  121,  122,  123,  124,  125,
 /*   520 */   126,  127,  128,  129,  139,  131,  141,  133,  179,   79,
 /*   530 */   248,  137,  138,    5,    6,    7,    8,    9,   10,   11,
 /*   540 */    12,   13,   14,   15,   16,   17,   18,    7,   20,   21,
 /*   550 */    22,   23,   24,   25,   26,   27,   28,   29,   30,   45,
 /*   560 */    32,  179,  259,   49,   24,   25,  179,   50,   51,  147,
 /*   570 */   179,   31,   32,   50,   51,  179,   48,  218,   50,  197,
 /*   580 */   198,  199,  179,   43,   44,  203,  179,  228,   48,   44,
 /*   590 */    50,   51,   44,  197,  198,  199,   79,   36,   50,   51,
 /*   600 */   197,  198,  199,  153,  197,  198,  199,  154,   55,   69,
 /*   610 */    57,  204,  263,   52,  208,  209,  102,  103,   65,   91,
 /*   620 */   258,   81,  226,   62,   63,  102,  103,  179,   26,  226,
 /*   630 */   248,   82,   83,   84,   85,    9,   10,   11,   12,  249,
 /*   640 */   250,  251,  102,  112,  248,  100,  115,  255,  256,  118,
 /*   650 */   263,  248,   50,   51,  263,  248,  123,  124,  125,  119,
 /*   660 */   120,  121,  122,  123,  124,  125,  126,  127,  128,  129,
 /*   670 */   199,  131,  190,  133,  208,  209,  194,  137,  138,    5,
 /*   680 */     6,    7,    8,    9,   10,   11,   12,   13,   14,   15,
 /*   690 */    16,   17,   18,  199,   20,   21,   22,   23,   24,   25,
 /*   700 */    26,   27,   28,   29,   30,  179,   32,    5,    6,    7,
 /*   710 */     8,    9,   10,   11,   12,   13,   14,   15,   16,   17,
 /*   720 */    18,   40,   20,   21,   22,   23,   24,   25,   26,   27,
 /*   730 */    28,   29,   30,  179,   32,   50,   51,    5,    6,    7,
 /*   740 */     8,    9,   10,   11,   12,   13,   14,   15,   16,   17,
 /*   750 */    18,   49,   20,   21,   22,   23,   24,   25,   26,   27,
 /*   760 */    28,   29,   30,  179,   32,  211,  179,  231,  232,  212,
 /*   770 */    68,   90,  215,   93,  212,   44,  212,  215,   98,  215,
 /*   780 */     7,  107,  108,  179,  179,  179,  179,    5,    6,    7,
 /*   790 */     8,    9,   10,   11,   12,   13,   14,   15,   16,   17,
 /*   800 */    18,  179,   20,   21,   22,   23,   24,   25,   26,   27,
 /*   810 */    28,   29,   30,  179,   32,  211,  211,  211,   86,  197,
 /*   820 */   198,  199,  238,  142,   51,  238,   49,   45,  147,  179,
 /*   830 */    99,  179,  179,    5,    6,    7,    8,    9,   10,   11,
 /*   840 */    12,   13,   14,   15,   16,   17,   18,  179,   20,   21,
 /*   850 */    22,   23,   24,   25,   26,   27,   28,   29,   30,   84,
 /*   860 */    32,  211,   87,  211,  211,  197,  198,  199,  229,  247,
 /*   870 */   248,  229,  238,   45,  229,  179,  107,  108,  101,    5,
 /*   880 */     6,    7,    8,    9,   10,   11,   12,   13,   14,   15,
 /*   890 */    16,   17,   18,  179,   20,   21,   22,   23,   24,   25,
 /*   900 */    26,   27,   28,   29,   30,   54,   32,  211,  216,   45,
 /*   910 */   242,  197,  198,  199,  217,  179,  248,  179,   45,   45,
 /*   920 */    50,   51,   49,   72,  179,    5,    6,    7,    8,    9,
 /*   930 */    10,   11,   12,   13,   14,   15,   16,   17,   18,  179,
 /*   940 */    20,   21,   22,   23,   24,   25,   26,   27,   28,   29,
 /*   950 */    30,   45,   32,   50,   51,   49,   45,  197,  198,  199,
 /*   960 */    49,  220,  248,   99,   45,   45,   45,  179,   49,  179,
 /*   970 */    49,    5,    6,    7,    8,    9,   10,   11,   12,   13,
 /*   980 */    14,   15,   16,   17,   18,  179,   20,   21,   22,   23,
 /*   990 */    24,   25,   26,   27,   28,   29,   30,   45,   32,  179,
 /*  1000 */   179,   49,   45,  197,  198,  199,   49,  264,  248,   45,
 /*  1010 */    45,   45,   45,   49,   49,  179,   49,    5,    6,    7,
 /*  1020 */     8,    9,   10,   11,   12,   13,   14,   15,   16,   17,
 /*  1030 */    18,  179,   20,   21,   22,   23,   24,   25,   26,   27,
 /*  1040 */    28,   29,   30,   45,   32,  179,  179,   49,   45,   45,
 /*  1050 */   234,  234,   49,   49,  248,  179,  179,   45,  220,  179,
 /*  1060 */   264,  179,  179,    5,    6,    7,    8,    9,   10,   11,
 /*  1070 */    12,   13,   14,   15,   16,   17,   18,  179,   20,   21,
 /*  1080 */    22,   23,   24,   25,   26,   27,   28,   29,   30,  179,
 /*  1090 */    32,    5,    6,    7,    8,    9,   10,   11,   12,   13,
 /*  1100 */    14,   15,   16,   17,   18,  179,   20,   21,   22,   23,
 /*  1110 */    24,   25,   26,   27,   28,   29,   30,  179,   32,    5,
 /*  1120 */     6,    7,    8,    9,   10,   11,   12,   13,   14,   15,
 /*  1130 */    16,   17,   18,  179,   20,   21,   22,   23,   24,   25,
 /*  1140 */    26,   27,   28,   29,   30,  201,   32,    5,    6,    7,
 /*  1150 */     8,    9,   10,   11,   12,   13,   14,   15,   16,   17,
 /*  1160 */    18,  179,   20,   21,   22,   23,   24,   25,   26,   27,
 /*  1170 */    28,   29,   30,  179,   32,   50,   51,  234,  179,  179,
 /*  1180 */   230,  254,  219,   44,  179,  127,  179,  101,  217,  217,
 /*  1190 */   179,    5,    6,    7,    8,    9,   10,   11,   12,   13,
 /*  1200 */    14,   15,   16,   17,   18,  179,   20,   21,   22,   23,
 /*  1210 */    24,   25,   26,   27,   28,   29,   30,  179,   32,    5,
 /*  1220 */     6,    7,    8,    9,   10,   11,   12,   13,   14,   15,
 /*  1230 */    16,   17,   18,  179,   20,   21,   22,   23,   24,   25,
 /*  1240 */    26,   27,   28,   29,   30,  179,   32,  202,   44,  241,
 /*  1250 */   125,    5,  114,  207,  140,   15,   16,   17,   18,   19,
 /*  1260 */    20,   21,   22,   23,   24,   25,   26,   27,   28,   29,
 /*  1270 */    30,  205,   32,  220,   42,  210,  139,  210,  210,  210,
 /*  1280 */   261,  149,  140,  260,   42,  185,  207,  101,    5,    6,
 /*  1290 */     7,    8,    9,   10,   11,   12,   13,   14,   15,   16,
 /*  1300 */    17,   18,  220,   20,   21,   22,   23,   24,   25,   26,
 /*  1310 */    27,   28,   29,   30,  241,   32,    6,    7,    8,    9,
 /*  1320 */    10,   11,   12,   13,   14,   15,   16,   17,   18,  207,
 /*  1330 */    20,   21,   22,   23,   24,   25,   26,   27,   28,   29,
 /*  1340 */    30,  127,   32,    7,    8,    9,   10,   11,   12,   13,
 /*  1350 */    14,   15,   16,   17,   18,  185,   20,   21,   22,   23,
 /*  1360 */    24,   25,   26,   27,   28,   29,   30,  202,   32,  185,
 /*  1370 */   185,  202,   88,  202,   91,  200,  200,  105,   88,  236,
 /*  1380 */     5,    6,    7,    8,    9,   10,   11,   12,   13,   14,
 /*  1390 */    15,   16,   17,   18,   76,   20,   21,   22,   23,   24,
 /*  1400 */    25,   26,   27,   28,   29,   30,   44,   32,    5,    6,
 /*  1410 */     7,    8,    9,   10,   11,   12,   13,   14,   15,   16,
 /*  1420 */    17,   18,  235,   20,   21,   22,   23,   24,   25,   26,
 /*  1430 */    27,   28,   29,   30,  236,   32,  235,  139,  236,  236,
 /*  1440 */   235,  235,  124,   15,   16,   17,   18,  179,   20,   21,
 /*  1450 */    22,   23,   24,   25,   26,   27,   28,   29,   30,  252,
 /*  1460 */    32,  253,  202,  202,   44,  197,  198,  199,  272,  111,
 /*  1470 */   106,  224,  179,  155,  156,  157,  158,  159,  160,  161,
 /*  1480 */   162,  163,  164,  165,  166,  167,  168,  169,  170,  171,
 /*  1490 */   197,  198,  199,  179,  223,  110,  222,  109,  221,  179,
 /*  1500 */   220,   46,  179,   47,  187,  179,  182,   44,  179,  181,
 /*  1510 */    37,  197,  198,  199,  179,  178,  248,  197,  198,  199,
 /*  1520 */   197,  198,  199,  197,  198,  199,  197,  198,  199,  179,
 /*  1530 */   178,  192,  197,  198,  199,  192,  179,  183,    4,  179,
 /*  1540 */     3,  248,   67,   44,   85,  179,   44,  197,  198,  199,
 /*  1550 */    44,   67,   39,   67,  197,  198,  199,  197,  198,  199,
 /*  1560 */   179,   79,  248,  197,  198,  199,  179,   89,  248,  179,
 /*  1570 */    67,  248,  179,   40,  248,  179,   67,  248,  197,  198,
 /*  1580 */   199,  179,   40,  248,  197,  198,  199,  197,  198,  199,
 /*  1590 */   197,  198,  199,  197,  198,  199,   45,  179,  248,  197,
 /*  1600 */   198,  199,   40,  179,   41,  248,  179,   14,  248,   44,
 /*  1610 */   143,   97,  179,  117,  248,  197,  198,  199,  101,   88,
 /*  1620 */   116,  197,  198,  199,  197,  198,  199,  179,  148,  248,
 /*  1630 */   197,  198,  199,  179,  151,  248,  179,  148,  248,   43,
 /*  1640 */   179,  248,  150,    1,  248,  197,  198,  199,   49,  117,
 /*  1650 */   248,  197,  198,  199,  197,  198,  199,   88,  197,  198,
 /*  1660 */   199,  101,   14,   14,  179,   14,  248,  179,   14,  116,
 /*  1670 */   179,   51,  248,  179,  104,  248,   99,    1,   44,  179,
 /*  1680 */    88,  248,  197,  198,  199,  197,  198,  199,  197,  198,
 /*  1690 */   199,  197,  198,  199,  179,  146,  248,  197,  198,  199,
 /*  1700 */   179,  142,  248,  179,   91,  248,  179,   43,    7,  248,
 /*  1710 */    85,   44,  197,  198,  199,   44,   44,   67,  197,  198,
 /*  1720 */   199,  197,  198,  199,  197,  198,  199,  179,   44,   67,
 /*  1730 */    44,   93,   67,  248,  179,   45,  248,  179,   81,  248,
 /*  1740 */   179,   59,  248,   45,   45,  197,  198,  199,  248,   49,
 /*  1750 */    44,   44,  197,  198,  199,  197,  198,  199,  197,  198,
 /*  1760 */   199,  179,   45,  248,  179,   49,   45,  179,   99,  248,
 /*  1770 */   179,   45,  248,  179,   51,  248,   44,  104,  103,  197,
 /*  1780 */   198,  199,  197,  198,  199,  197,  198,  199,  197,  198,
 /*  1790 */   199,  197,  198,  199,  179,   49,  248,  179,  104,   49,
 /*  1800 */   104,  179,   32,  248,  130,   45,  248,   45,   45,  248,
 /*  1810 */    49,  132,  197,  198,  199,  197,  198,  199,   45,  197,
 /*  1820 */   198,  199,   45,   48,   45,   49,   44,   61,   45,   44,
 /*  1830 */   248,   91,   45,  248,   44,   49,  248,   45,   91,  248,
 /*  1840 */    44,   44,  248,   45,   44,  130,   45,  125,   45,   44,
 /*  1850 */    81,   99,   48,   14,   39,   34,    1,    0,  274,  274,
 /*  1860 */   274,  274,  274,  248,  274,  274,  248,  274,  274,  274,
 /*  1870 */   248,
};
#define YY_SHIFT_USE_DFLT (1871)
#define YY_SHIFT_COUNT    (448)
#define YY_SHIFT_MIN      (-20)
#define YY_SHIFT_MAX      (1857)
static const short yy_shift_ofst[] = {
 /*     0 */   176,   -6,  116,  148,   -7,  540,  540,  540,  540,  540,
 /*    10 */   540,  540,  540,  540,  540,  540,  540,  540,  540,  540,
 /*    20 */   540,  517,  271,  394,  540,  540,  540,  540,  540,  540,
 /*    30 */   540,  540,  540,  540,  540,  540,  540,  540,  540,  540,
 /*    40 */   540,  540,  540,  540,  540,  540,  540,  540,  540,  540,
 /*    50 */   540,  540,  540,  540,  540,  540,  540,  540,  540,  540,
 /*    60 */   540,  540,  540,  540,  540,  540,  540,  540,  540,  540,
 /*    70 */   540,  540,  540,  318,  517,  528,  674,  674,  732,  732,
 /*    80 */    72,  186,  262,  -12,  702,  782,  828,  874,  920,  966,
 /*    90 */  1012, 1058, 1086, 1114, 1142, 1186, 1214, 1283, 1375, 1375,
 /*   100 */  1375, 1375, 1375, 1375, 1375, 1375, 1375, 1375, 1375, 1375,
 /*   110 */  1375, 1375, 1375, 1375, 1375, 1375, 1375, 1375, 1375, 1375,
 /*   120 */  1403, 1375, 1375, 1310, 1336, 1336, 1240, 1428, 1428, 1428,
 /*   130 */   -19,   62,  317,  201,  -18,  254,   46,   46,   46, 1318,
 /*   140 */  1318, 1318,  236,   84,  286,  164,  164,  450,  422,  434,
 /*   150 */   434,  453,  685,  685,  685,  175,  685,  685,  685,  685,
 /*   160 */   685,  685,  685,  685,  685,  685,  685,  685,  685,  685,
 /*   170 */   685,  685,  685,  685,  685,  254,  -20,  -20,   25,  -20,
 /*   180 */   -20,  -20,   25, 1871, 1871,  302,  514,    5,  404,  523,
 /*   190 */   561,  531,  553,  439,  533,  533,  602,   18,   18,   18,
 /*   200 */   197,  548,   92,  322,  685,  685,  685,  685,  685,  685,
 /*   210 */   685,  685,  685,  685,  685,  327,  446,  446,  685,  685,
 /*   220 */   773,  327,  685,  773,  685,  685,  685,  685,  685,  685,
 /*   230 */   685,  769,  685,  685,  446,  685,  685,  545,  385,  685,
 /*   240 */   685,  680,  685,  777,  680,  685,  685,  685,  685, 1139,
 /*   250 */  1204, 1246, 1138, 1232, 1232, 1232, 1232, 1137, 1132, 1242,
 /*   260 */  1138, 1204, 1246, 1246, 1242, 1139, 1242, 1242, 1139, 1139,
 /*   270 */  1284, 1284, 1272, 1290, 1272, 1290, 1272, 1290, 1272, 1290,
 /*   280 */  1362, 1298, 1139, 1139, 1420, 1358, 1364, 1385, 1388, 1138,
 /*   290 */  1456, 1455, 1463, 1473, 1473, 1871, 1871, 1871,  383,  681,
 /*   300 */   330,  549,  626, 1125,  873,  911,  851,  775,  864,  906,
 /*   310 */   919,  952,  957,  964,  965,  998,  870,  921,  967, 1003,
 /*   320 */   298,  903,  731, 1004, 1534, 1537, 1475, 1499, 1502, 1459,
 /*   330 */  1506, 1484, 1513, 1486, 1503, 1509, 1482, 1478, 1533, 1542,
 /*   340 */  1563, 1551, 1565, 1593, 1467, 1562, 1514, 1496, 1504, 1517,
 /*   350 */  1531, 1596, 1480, 1599, 1489, 1483, 1492, 1569, 1642, 1560,
 /*   360 */  1532, 1648, 1649, 1651, 1654, 1553, 1620, 1570, 1577, 1676,
 /*   370 */  1549, 1634, 1592, 1559, 1613, 1664, 1701, 1625, 1667, 1650,
 /*   380 */  1671, 1672, 1684, 1662, 1686, 1665, 1657, 1682, 1690, 1698,
 /*   390 */  1699, 1700, 1638, 1706, 1717, 1707, 1716, 1669, 1721, 1726,
 /*   400 */  1723, 1673, 1732, 1675, 1746, 1694, 1750, 1696, 1770, 1674,
 /*   410 */  1760, 1762, 1763, 1761, 1679, 1773, 1746, 1777, 1775, 1776,
 /*   420 */  1779, 1782, 1766, 1783, 1785, 1740, 1786, 1787, 1790, 1747,
 /*   430 */  1786, 1792, 1796, 1798, 1797, 1800, 1715, 1801, 1722, 1803,
 /*   440 */  1805, 1769, 1752, 1804, 1839, 1815, 1821, 1855, 1857,
};
#define YY_REDUCE_USE_DFLT (-176)
#define YY_REDUCE_COUNT (297)
#define YY_REDUCE_MIN   (-175)
#define YY_REDUCE_MAX   (1622)
static const short yy_reduce_ofst[] = {
 /*     0 */  -153, -161, -151,  127,  -23,  -16,  184,  208,  224,  -99,
 /*    10 */   250,  261,   16,  130,  282,  382,  396,  403,  407,  622,
 /*    20 */   668, -119,  -29,  714,  760,  806, 1268, 1293, 1314, 1320,
 /*    30 */  1323, 1326, 1329, 1335, 1350, 1357, 1360, 1366, 1381, 1387,
 /*    40 */  1390, 1393, 1396, 1402, 1418, 1424, 1427, 1433, 1448, 1454,
 /*    50 */  1457, 1461, 1485, 1488, 1491, 1494, 1500, 1515, 1521, 1524,
 /*    60 */  1527, 1548, 1555, 1558, 1561, 1582, 1585, 1588, 1591, 1594,
 /*    70 */  1615, 1618, 1622,  212,  -54,   61,   87,  119,  -61,    8,
 /*    80 */  -138, -175,   51,  279,  390,  390,  390,  390,  390,  390,
 /*    90 */   390,  390,  390,  390,  390,  390,  390,  390,  390,  390,
 /*   100 */   390,  390,  390,  390,  390,  390,  390,  390,  390,  390,
 /*   110 */   390,  390,  390,  390,  390,  390,  390,  390,  390,  390,
 /*   120 */   390,  390,  390,  390,  390,  390,  390,  390,  390,  390,
 /*   130 */   390,  390,  390,  390,  390, -132,  390,  390,  390, -130,
 /*   140 */   -22,   19,  -96,   90,  297,  406,  466,   73,  392,  115,
 /*   150 */   192,  -43, -146,  151,  260, -170,  554,  604,  605,  606,
 /*   160 */   650,  652,  653,   69,  696,  349,  264,  387,  584,  587,
 /*   170 */   391,  294,  634,  237, 1066,  232,  -58,   50,  274,  557,
 /*   180 */   562,  564,  482,  536,  359,  -77,  -32,   -5,   81,   96,
 /*   190 */   219,  303,  362,  448,  471,  494,  526,  639,  642,  645,
 /*   200 */   -32,  607,  692,  697,  736,  738,  745,  788,  790,  820,
 /*   210 */   821,  836,  852,  866,  867,  741,  816,  817,  876,  877,
 /*   220 */   743,  838,  880,  796,  882,  883,  898,  910,  926,  938,
 /*   230 */   954,  944,  982,  994,  943,  999, 1000,  950,  927, 1005,
 /*   240 */  1007,  971,  526,  963,  972, 1011, 1026, 1038, 1054, 1045,
 /*   250 */  1008, 1046, 1053, 1065, 1067, 1068, 1069, 1019, 1023, 1100,
 /*   260 */  1082, 1073, 1079, 1122, 1170, 1165, 1184, 1185, 1169, 1171,
 /*   270 */  1175, 1176, 1143, 1187, 1198, 1201, 1202, 1205, 1203, 1206,
 /*   280 */  1207, 1208, 1260, 1261, 1196, 1247, 1271, 1274, 1277, 1280,
 /*   290 */  1317, 1324, 1328, 1337, 1352, 1339, 1343, 1354,
};
static const YYACTIONTYPE yy_default[] = {
 /*     0 */  1274, 1248, 1248, 1248, 1306, 1181, 1181, 1181, 1181, 1131,
 /*    10 */  1181, 1181, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1139,
 /*    20 */  1306, 1005, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*    30 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*    40 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*    50 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*    60 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*    70 */  1306, 1306, 1306, 1248, 1005, 1051, 1082, 1082, 1017, 1017,
 /*    80 */  1005, 1306, 1248, 1057, 1088, 1306, 1306, 1306, 1306, 1182,
 /*    90 */  1183, 1306, 1306, 1306, 1306, 1138, 1306, 1306, 1198, 1212,
 /*   100 */  1099, 1098, 1097, 1096, 1067, 1094, 1086, 1090, 1127, 1132,
 /*   110 */  1136, 1135, 1182, 1178, 1179, 1177, 1140, 1137, 1133, 1183,
 /*   120 */  1306, 1128, 1089, 1149, 1162, 1148, 1158, 1168, 1159, 1151,
 /*   130 */  1150, 1152, 1153, 1154, 1155, 1306, 1165, 1164, 1163, 1306,
 /*   140 */  1306, 1306, 1306, 1306, 1005, 1306, 1306, 1306, 1188, 1306,
 /*   150 */  1306, 1306, 1306, 1306, 1306, 1248, 1306, 1306, 1306, 1306,
 /*   160 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*   170 */  1306, 1306, 1306, 1306, 1306, 1306, 1248, 1248, 1017, 1248,
 /*   180 */  1248, 1248, 1017, 1057, 1046, 1306, 1306, 1051, 1306, 1306,
 /*   190 */  1306, 1306, 1206, 1306, 1306, 1306, 1306, 1051, 1051, 1051,
 /*   200 */  1055, 1306, 1035, 1045, 1306, 1306, 1306, 1306, 1306, 1306,
 /*   210 */  1200, 1306, 1306, 1306, 1306, 1093, 1069, 1069, 1306, 1306,
 /*   220 */  1303, 1093, 1306, 1303, 1306, 1306, 1306, 1306, 1306, 1306,
 /*   230 */  1306, 1082, 1306, 1306, 1069, 1306, 1306, 1053, 1180, 1306,
 /*   240 */  1306, 1045, 1306, 1054, 1045, 1306, 1306, 1306, 1306, 1189,
 /*   250 */  1104, 1025, 1093, 1032, 1032, 1032, 1032, 1211, 1210,  997,
 /*   260 */  1093, 1104, 1025, 1025,  997, 1189,  997,  997, 1189, 1189,
 /*   270 */  1023, 1023, 1073, 1068, 1073, 1068, 1073, 1068, 1073, 1068,
 /*   280 */  1184, 1306, 1189, 1189, 1306, 1087, 1074, 1085, 1083, 1093,
 /*   290 */  1306, 1281, 1306, 1277, 1277, 1288, 1288, 1000, 1240, 1306,
 /*   300 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1245, 1306, 1306,
 /*   310 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*   320 */  1306, 1306, 1110, 1306, 1306,  987, 1306, 1306, 1306, 1306,
 /*   330 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1062, 1306, 1306,
 /*   340 */  1306, 1306, 1195, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*   350 */  1306, 1306, 1306, 1209, 1208, 1306, 1306, 1306, 1306, 1306,
 /*   360 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1302, 1306,
 /*   370 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*   380 */  1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*   390 */  1306, 1306, 1037, 1306, 1306, 1306, 1291, 1306, 1306, 1306,
 /*   400 */  1306, 1306, 1306, 1306, 1084, 1306, 1075, 1306, 1126, 1306,
 /*   410 */  1306, 1306, 1306, 1298, 1306, 1306, 1300, 1306, 1306, 1306,
 /*   420 */  1306, 1306, 1306, 1306, 1306, 1306, 1250, 1306, 1306, 1306,
 /*   430 */  1249, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306, 1306,
 /*   440 */  1306, 1306, 1306, 1306, 1306,  991, 1306, 1306, 1306,
};
/********** End of lemon-generated parsing tables *****************************/

/* The next table maps tokens (terminal symbols) into fallback tokens.  
** If a construct like the following:
** 
**      %fallback ID X Y Z.
**
** appears in the grammar, then ID becomes a fallback token for X, Y,
** and Z.  Whenever one of the tokens X, Y, or Z is input to the parser
** but it does not parse, the type of the token is changed to ID and
** the parse is retried before an error is thrown.
**
** This feature can be used, for example, to cause some keywords in a language
** to revert to identifiers if they keyword does not apply in the context where
** it appears.
*/
#ifdef YYFALLBACK
static const YYCODETYPE yyFallback[] = {
    0,  /*          $ => nothing */
    0,  /*       SEMI => nothing */
    0,  /*    EXPLAIN => nothing */
   50,  /*      QUERY => ID */
   50,  /*       PLAN => ID */
    0,  /*         OR => nothing */
    0,  /*        AND => nothing */
    0,  /*        NOT => nothing */
    0,  /*         IS => nothing */
   50,  /*      MATCH => ID */
    0,  /*    LIKE_KW => nothing */
    0,  /*    BETWEEN => nothing */
    0,  /*         IN => nothing */
    0,  /*         NE => nothing */
    0,  /*         EQ => nothing */
    0,  /*         GT => nothing */
    0,  /*         LE => nothing */
    0,  /*         LT => nothing */
    0,  /*         GE => nothing */
    0,  /*     ESCAPE => nothing */
    0,  /*     BITAND => nothing */
    0,  /*      BITOR => nothing */
    0,  /*     LSHIFT => nothing */
    0,  /*     RSHIFT => nothing */
    0,  /*       PLUS => nothing */
    0,  /*      MINUS => nothing */
    0,  /*       STAR => nothing */
    0,  /*      SLASH => nothing */
    0,  /*        REM => nothing */
    0,  /*     CONCAT => nothing */
    0,  /*    COLLATE => nothing */
    0,  /*     BITNOT => nothing */
    0,  /*         LB => nothing */
    0,  /*      START => nothing */
    0,  /* TRANSACTION => nothing */
    0,  /*     COMMIT => nothing */
    0,  /*   ROLLBACK => nothing */
    0,  /*  SAVEPOINT => nothing */
   50,  /*    RELEASE => ID */
    0,  /*         TO => nothing */
    0,  /*      TABLE => nothing */
    0,  /*     CREATE => nothing */
   50,  /*         IF => ID */
    0,  /*     EXISTS => nothing */
    0,  /*         LP => nothing */
    0,  /*         RP => nothing */
    0,  /*       WITH => nothing */
   50,  /*     ENGINE => ID */
    0,  /*     STRING => nothing */
    0,  /*      COMMA => nothing */
    0,  /*         ID => nothing */
    0,  /*    INDEXED => nothing */
   50,  /*      ABORT => ID */
   50,  /*     ACTION => ID */
   50,  /*        ADD => ID */
   50,  /*      AFTER => ID */
   50,  /* AUTOINCREMENT => ID */
   50,  /*     BEFORE => ID */
   50,  /*    CASCADE => ID */
   50,  /*   CONFLICT => ID */
   50,  /*   DEFERRED => ID */
   50,  /*        END => ID */
   50,  /*       FAIL => ID */
   50,  /*     IGNORE => ID */
   50,  /*  INITIALLY => ID */
   50,  /*    INSTEAD => ID */
   50,  /*         NO => ID */
   50,  /*        KEY => ID */
   50,  /*     OFFSET => ID */
   50,  /*      RAISE => ID */
   50,  /*    REPLACE => ID */
   50,  /*   RESTRICT => ID */
   50,  /*     RENAME => ID */
   50,  /*   CTIME_KW => ID */
   50,  /*     ENABLE => ID */
   50,  /*    DISABLE => ID */
   50,  /*       UUID => ID */
   50,  /*       SHOW => ID */
};
#endif /* YYFALLBACK */

/* The following structure represents a single element of the
** parser's stack.  Information stored includes:
**
**   +  The state number for the parser at this level of the stack.
**
**   +  The value of the token stored at this level of the stack.
**      (In other words, the "major" token.)
**
**   +  The semantic value stored at this level of the stack.  This is
**      the information used by the action routines in the grammar.
**      It is sometimes called the "minor" token.
**
** After the "shift" half of a SHIFTREDUCE action, the stateno field
** actually contains the reduce action for the second half of the
** SHIFTREDUCE.
*/
struct yyStackEntry {
  YYACTIONTYPE stateno;  /* The state-number, or reduce action in SHIFTREDUCE */
  YYCODETYPE major;      /* The major token value.  This is the code
                         ** number for the token at this stack level */
  YYMINORTYPE minor;     /* The user-supplied minor token value.  This
                         ** is the value of the token  */
};
typedef struct yyStackEntry yyStackEntry;

/* The state of the parser is completely contained in an instance of
** the following structure */
struct yyParser {
  yyStackEntry *yytos;          /* Pointer to top element of the stack */
#ifdef YYTRACKMAXSTACKDEPTH
  int yyhwm;                    /* High-water mark of the stack */
#endif
#ifndef YYNOERRORRECOVERY
  int yyerrcnt;                 /* Shifts left before out of the error */
#endif
  bool is_fallback_failed;      /* Shows if fallback failed or not */
  sqlParserARG_SDECL                /* A place to hold %extra_argument */
#if YYSTACKDEPTH<=0
  int yystksz;                  /* Current side of the stack */
  yyStackEntry *yystack;        /* The parser's stack */
  yyStackEntry yystk0;          /* First stack entry */
#else
  yyStackEntry yystack[YYSTACKDEPTH];  /* The parser's stack */
#endif
};
typedef struct yyParser yyParser;

#ifndef NDEBUG
#include <stdio.h>
static FILE *yyTraceFILE = 0;
static char *yyTracePrompt = 0;
#endif /* NDEBUG */

#ifndef NDEBUG
/* 
** Turn parser tracing on by giving a stream to which to write the trace
** and a prompt to preface each trace message.  Tracing is turned off
** by making either argument NULL 
**
** Inputs:
** <ul>
** <li> A FILE* to which trace output should be written.
**      If NULL, then tracing is turned off.
** <li> A prefix string written at the beginning of every
**      line of trace output.  If NULL, then tracing is
**      turned off.
** </ul>
**
** Outputs:
** None.
*/
void sqlParserTrace(FILE *TraceFILE, char *zTracePrompt){
  yyTraceFILE = TraceFILE;
  yyTracePrompt = zTracePrompt;
  if( yyTraceFILE==0 ) yyTracePrompt = 0;
  else if( yyTracePrompt==0 ) yyTraceFILE = 0;
}
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing shifts, the names of all terminals and nonterminals
** are required.  The following table supplies these names */
static const char *const yyTokenName[] = { 
  "$",             "SEMI",          "EXPLAIN",       "QUERY",       
  "PLAN",          "OR",            "AND",           "NOT",         
  "IS",            "MATCH",         "LIKE_KW",       "BETWEEN",     
  "IN",            "NE",            "EQ",            "GT",          
  "LE",            "LT",            "GE",            "ESCAPE",      
  "BITAND",        "BITOR",         "LSHIFT",        "RSHIFT",      
  "PLUS",          "MINUS",         "STAR",          "SLASH",       
  "REM",           "CONCAT",        "COLLATE",       "BITNOT",      
  "LB",            "START",         "TRANSACTION",   "COMMIT",      
  "ROLLBACK",      "SAVEPOINT",     "RELEASE",       "TO",          
  "TABLE",         "CREATE",        "IF",            "EXISTS",      
  "LP",            "RP",            "WITH",          "ENGINE",      
  "STRING",        "COMMA",         "ID",            "INDEXED",     
  "ABORT",         "ACTION",        "ADD",           "AFTER",       
  "AUTOINCREMENT",  "BEFORE",        "CASCADE",       "CONFLICT",    
  "DEFERRED",      "END",           "FAIL",          "IGNORE",      
  "INITIALLY",     "INSTEAD",       "NO",            "KEY",         
  "OFFSET",        "RAISE",         "REPLACE",       "RESTRICT",    
  "RENAME",        "CTIME_KW",      "ENABLE",        "DISABLE",     
  "UUID",          "SHOW",          "WILDCARD",      "CONSTRAINT",  
  "DEFAULT",       "NULL",          "PRIMARY",       "UNIQUE",      
  "CHECK",         "REFERENCES",    "AUTOINCR",      "FOREIGN",     
  "ON",            "DROP",          "VIEW",          "AS",          
  "UNION",         "ALL",           "EXCEPT",        "INTERSECT",   
  "SELECT",        "VALUES",        "DISTINCT",      "DOT",         
  "SEQSCAN",       "FROM",          "JOIN_KW",       "JOIN",        
  "BY",            "USING",         "ORDER",         "ASC",         
  "DESC",          "GROUP",         "HAVING",        "LIMIT",       
  "DELETE",        "TRUNCATE",      "WHERE",         "UPDATE",      
  "SET",           "INTO",          "INSERT",        "BLOB",        
  "FALSE",         "TRUE",          "UNKNOWN",       "FLOAT",       
  "DECIMAL",       "INTEGER",       "VARNUM",        "COLON",       
  "VARIABLE",      "CAST",          "RB",            "LCB",         
  "RCB",           "TRIM",          "LEADING",       "TRAILING",    
  "BOTH",          "CHAR",          "CASE",          "WHEN",        
  "THEN",          "ELSE",          "INDEX",         "SESSION",     
  "PRAGMA",        "FUNCTION_KW",   "BEGIN",         "TRIGGER",     
  "OF",            "FOR",           "EACH",          "ROW",         
  "ALTER",         "COLUMN",        "RECURSIVE",     "TEXT",        
  "STRING_KW",     "SCALAR",        "BOOL",          "BOOLEAN",     
  "VARBINARY",     "ANY",           "ARRAY",         "MAP",         
  "DATETIME",      "INTERVAL",      "VARCHAR",       "NUMBER",      
  "DOUBLE",        "INT",           "INTEGER_KW",    "UNSIGNED",    
  "error",         "input",         "ecmd",          "explain",     
  "cmdx",          "cmd",           "savepoint_opt",  "nm",          
  "create_table",  "create_table_args",  "with_opts",     "create_table_end",
  "createkw",      "ifnotexists",   "columnlist",    "engine_opts", 
  "tcons",         "column_def",    "create_column_end",  "column_name_and_type",
  "carglist",      "typedef",       "autoinc",       "ccons",       
  "cconsname",     "term",          "expr",          "number",      
  "onconf",        "sortorder",     "eidlist_opt",   "col_list_with_autoinc",
  "sortlist",      "eidlist",       "index_onconf",  "orconf",      
  "resolvetype",   "raisetype",     "ifexists",      "fullname",    
  "select",        "selectnowith",  "oneselect",     "with",        
  "multiselect_op",  "distinct",      "selcollist",    "from",        
  "where_opt",     "groupby_opt",   "having_opt",    "orderby_opt", 
  "limit_opt",     "values",        "nexprlist",     "exprlist",    
  "sclp",          "as",            "seqscan",       "seltablist",  
  "stl_prefix",    "joinop",        "indexed_opt",   "on_opt",      
  "using_opt",     "join_nm",       "idlist",        "setlist",     
  "insert_cmd",    "idlist_opt",    "getlist",       "maplist",     
  "nmaplist",      "trim_operands",  "trim_specification",  "expr_optional",
  "type_func",     "likeop",        "between_op",    "in_op",       
  "paren_exprlist",  "case_exprlist",  "case_else",     "uniqueflag",  
  "trigger_decl",  "trigger_cmd_list",  "trigger_time",  "trigger_event",
  "foreach_clause",  "when_clause",   "trigger_cmd",   "trnm",        
  "tridxby",       "alter_table_start",  "alter_add_constraint",  "alter_add_column",
  "column_name",   "alter_column_def",  "unique_spec",   "wqlist",      
  "char_len",      "number_typedef",
};
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing reduce actions, the names of all rules are required.
*/
static const char *const yyRuleName[] = {
 /*   0 */ "ecmd ::= SEMI",
 /*   1 */ "explain ::= EXPLAIN",
 /*   2 */ "explain ::= EXPLAIN QUERY PLAN",
 /*   3 */ "cmd ::= START TRANSACTION",
 /*   4 */ "cmd ::= COMMIT",
 /*   5 */ "cmd ::= ROLLBACK",
 /*   6 */ "cmd ::= SAVEPOINT nm",
 /*   7 */ "cmd ::= RELEASE savepoint_opt nm",
 /*   8 */ "cmd ::= ROLLBACK TO savepoint_opt nm",
 /*   9 */ "create_table ::= createkw TABLE ifnotexists nm",
 /*  10 */ "createkw ::= CREATE",
 /*  11 */ "ifnotexists ::=",
 /*  12 */ "ifnotexists ::= IF NOT EXISTS",
 /*  13 */ "engine_opts ::= ENGINE EQ STRING",
 /*  14 */ "create_table_end ::=",
 /*  15 */ "column_name_and_type ::= nm typedef",
 /*  16 */ "create_column_end ::= autoinc",
 /*  17 */ "nm ::= ID|INDEXED",
 /*  18 */ "cconsname ::= CONSTRAINT nm",
 /*  19 */ "cconsname ::=",
 /*  20 */ "ccons ::= DEFAULT term",
 /*  21 */ "ccons ::= DEFAULT LP expr RP",
 /*  22 */ "ccons ::= DEFAULT PLUS number",
 /*  23 */ "ccons ::= DEFAULT MINUS number",
 /*  24 */ "ccons ::= NULL onconf",
 /*  25 */ "ccons ::= NOT NULL onconf",
 /*  26 */ "ccons ::= cconsname PRIMARY KEY sortorder",
 /*  27 */ "ccons ::= cconsname UNIQUE",
 /*  28 */ "ccons ::= cconsname CHECK LP expr RP",
 /*  29 */ "ccons ::= cconsname REFERENCES nm eidlist_opt",
 /*  30 */ "ccons ::= COLLATE ID|INDEXED",
 /*  31 */ "autoinc ::=",
 /*  32 */ "autoinc ::= AUTOINCR",
 /*  33 */ "tcons ::= cconsname PRIMARY KEY LP col_list_with_autoinc RP",
 /*  34 */ "tcons ::= cconsname UNIQUE LP sortlist RP",
 /*  35 */ "tcons ::= cconsname CHECK LP expr RP",
 /*  36 */ "tcons ::= cconsname FOREIGN KEY LP eidlist RP REFERENCES nm eidlist_opt",
 /*  37 */ "onconf ::=",
 /*  38 */ "onconf ::= ON CONFLICT resolvetype",
 /*  39 */ "orconf ::=",
 /*  40 */ "orconf ::= OR resolvetype",
 /*  41 */ "resolvetype ::= IGNORE",
 /*  42 */ "resolvetype ::= REPLACE",
 /*  43 */ "cmd ::= DROP TABLE ifexists fullname",
 /*  44 */ "cmd ::= DROP VIEW ifexists fullname",
 /*  45 */ "ifexists ::= IF EXISTS",
 /*  46 */ "ifexists ::=",
 /*  47 */ "cmd ::= createkw VIEW ifnotexists nm eidlist_opt AS select",
 /*  48 */ "cmd ::= select",
 /*  49 */ "select ::= with selectnowith",
 /*  50 */ "selectnowith ::= selectnowith multiselect_op oneselect",
 /*  51 */ "multiselect_op ::= UNION",
 /*  52 */ "multiselect_op ::= UNION ALL",
 /*  53 */ "multiselect_op ::= EXCEPT|INTERSECT",
 /*  54 */ "oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt orderby_opt limit_opt",
 /*  55 */ "values ::= VALUES LP nexprlist RP",
 /*  56 */ "values ::= values COMMA LP exprlist RP",
 /*  57 */ "distinct ::= DISTINCT",
 /*  58 */ "distinct ::= ALL",
 /*  59 */ "distinct ::=",
 /*  60 */ "sclp ::=",
 /*  61 */ "selcollist ::= sclp expr as",
 /*  62 */ "selcollist ::= sclp STAR",
 /*  63 */ "selcollist ::= sclp nm DOT STAR",
 /*  64 */ "as ::= AS nm",
 /*  65 */ "as ::=",
 /*  66 */ "seqscan ::= SEQSCAN",
 /*  67 */ "seqscan ::=",
 /*  68 */ "from ::=",
 /*  69 */ "from ::= FROM seltablist",
 /*  70 */ "stl_prefix ::= seltablist joinop",
 /*  71 */ "stl_prefix ::=",
 /*  72 */ "seltablist ::= stl_prefix seqscan nm as indexed_opt on_opt using_opt",
 /*  73 */ "seltablist ::= stl_prefix seqscan nm LP exprlist RP as on_opt using_opt",
 /*  74 */ "seltablist ::= stl_prefix LP select RP as on_opt using_opt",
 /*  75 */ "seltablist ::= stl_prefix LP seltablist RP as on_opt using_opt",
 /*  76 */ "fullname ::= nm",
 /*  77 */ "joinop ::= COMMA|JOIN",
 /*  78 */ "joinop ::= JOIN_KW JOIN",
 /*  79 */ "joinop ::= JOIN_KW join_nm JOIN",
 /*  80 */ "joinop ::= JOIN_KW join_nm join_nm JOIN",
 /*  81 */ "on_opt ::= ON expr",
 /*  82 */ "on_opt ::=",
 /*  83 */ "indexed_opt ::=",
 /*  84 */ "indexed_opt ::= INDEXED BY nm",
 /*  85 */ "indexed_opt ::= NOT INDEXED",
 /*  86 */ "using_opt ::= USING LP idlist RP",
 /*  87 */ "using_opt ::=",
 /*  88 */ "orderby_opt ::=",
 /*  89 */ "orderby_opt ::= ORDER BY sortlist",
 /*  90 */ "sortlist ::= sortlist COMMA expr sortorder",
 /*  91 */ "sortlist ::= expr sortorder",
 /*  92 */ "col_list_with_autoinc ::= col_list_with_autoinc COMMA expr autoinc",
 /*  93 */ "col_list_with_autoinc ::= expr autoinc",
 /*  94 */ "sortorder ::= ASC",
 /*  95 */ "sortorder ::= DESC",
 /*  96 */ "sortorder ::=",
 /*  97 */ "groupby_opt ::=",
 /*  98 */ "groupby_opt ::= GROUP BY nexprlist",
 /*  99 */ "having_opt ::=",
 /* 100 */ "having_opt ::= HAVING expr",
 /* 101 */ "limit_opt ::=",
 /* 102 */ "limit_opt ::= LIMIT expr",
 /* 103 */ "limit_opt ::= LIMIT expr OFFSET expr",
 /* 104 */ "limit_opt ::= LIMIT expr COMMA expr",
 /* 105 */ "cmd ::= with DELETE FROM fullname indexed_opt where_opt",
 /* 106 */ "cmd ::= TRUNCATE TABLE fullname",
 /* 107 */ "where_opt ::=",
 /* 108 */ "where_opt ::= WHERE expr",
 /* 109 */ "cmd ::= with UPDATE orconf fullname indexed_opt SET setlist where_opt",
 /* 110 */ "setlist ::= setlist COMMA nm EQ expr",
 /* 111 */ "setlist ::= setlist COMMA LP idlist RP EQ expr",
 /* 112 */ "setlist ::= nm EQ expr",
 /* 113 */ "setlist ::= LP idlist RP EQ expr",
 /* 114 */ "cmd ::= with insert_cmd INTO fullname idlist_opt select",
 /* 115 */ "cmd ::= with insert_cmd INTO fullname idlist_opt DEFAULT VALUES",
 /* 116 */ "insert_cmd ::= INSERT orconf",
 /* 117 */ "insert_cmd ::= REPLACE",
 /* 118 */ "idlist_opt ::=",
 /* 119 */ "idlist_opt ::= LP idlist RP",
 /* 120 */ "idlist ::= idlist COMMA nm",
 /* 121 */ "idlist ::= nm",
 /* 122 */ "expr ::= LP expr RP",
 /* 123 */ "term ::= NULL",
 /* 124 */ "expr ::= ID|INDEXED",
 /* 125 */ "expr ::= JOIN_KW",
 /* 126 */ "expr ::= nm DOT nm",
 /* 127 */ "term ::= BLOB",
 /* 128 */ "term ::= STRING",
 /* 129 */ "term ::= FALSE",
 /* 130 */ "term ::= TRUE",
 /* 131 */ "term ::= UNKNOWN",
 /* 132 */ "number ::= FLOAT",
 /* 133 */ "number ::= DECIMAL",
 /* 134 */ "number ::= INTEGER",
 /* 135 */ "expr ::= VARNUM",
 /* 136 */ "expr ::= COLON|VARIABLE ID|INDEXED",
 /* 137 */ "expr ::= COLON|VARIABLE INTEGER",
 /* 138 */ "expr ::= expr COLLATE ID|INDEXED",
 /* 139 */ "expr ::= CAST LP expr AS typedef RP",
 /* 140 */ "expr ::= expr LB getlist RB",
 /* 141 */ "getlist ::= getlist RB LB expr",
 /* 142 */ "getlist ::= expr",
 /* 143 */ "expr ::= LB exprlist RB",
 /* 144 */ "expr ::= LCB maplist RCB",
 /* 145 */ "maplist ::=",
 /* 146 */ "nmaplist ::= nmaplist COMMA expr COLON expr",
 /* 147 */ "nmaplist ::= expr COLON expr",
 /* 148 */ "expr ::= TRIM LP trim_operands RP",
 /* 149 */ "trim_operands ::= trim_specification expr FROM expr",
 /* 150 */ "trim_operands ::= trim_specification FROM expr",
 /* 151 */ "trim_operands ::= expr FROM expr",
 /* 152 */ "trim_operands ::= expr",
 /* 153 */ "expr_optional ::=",
 /* 154 */ "expr_optional ::= expr",
 /* 155 */ "trim_specification ::= LEADING",
 /* 156 */ "trim_specification ::= TRAILING",
 /* 157 */ "trim_specification ::= BOTH",
 /* 158 */ "expr ::= ID|INDEXED LP distinct exprlist RP",
 /* 159 */ "expr ::= type_func LP distinct exprlist RP",
 /* 160 */ "expr ::= ID|INDEXED LP STAR RP",
 /* 161 */ "expr ::= LP nexprlist COMMA expr RP",
 /* 162 */ "expr ::= expr AND expr",
 /* 163 */ "expr ::= expr OR expr",
 /* 164 */ "expr ::= expr LT|GT|GE|LE expr",
 /* 165 */ "expr ::= expr EQ|NE expr",
 /* 166 */ "expr ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr",
 /* 167 */ "expr ::= expr PLUS|MINUS expr",
 /* 168 */ "expr ::= expr STAR|SLASH|REM expr",
 /* 169 */ "expr ::= expr CONCAT expr",
 /* 170 */ "likeop ::= LIKE_KW|MATCH",
 /* 171 */ "likeop ::= NOT LIKE_KW|MATCH",
 /* 172 */ "expr ::= expr likeop expr",
 /* 173 */ "expr ::= expr likeop expr ESCAPE expr",
 /* 174 */ "expr ::= expr IS NULL",
 /* 175 */ "expr ::= expr IS NOT NULL",
 /* 176 */ "expr ::= NOT expr",
 /* 177 */ "expr ::= BITNOT expr",
 /* 178 */ "expr ::= MINUS expr",
 /* 179 */ "expr ::= PLUS expr",
 /* 180 */ "between_op ::= BETWEEN",
 /* 181 */ "between_op ::= NOT BETWEEN",
 /* 182 */ "expr ::= expr between_op expr AND expr",
 /* 183 */ "in_op ::= IN",
 /* 184 */ "in_op ::= NOT IN",
 /* 185 */ "expr ::= expr in_op LP exprlist RP",
 /* 186 */ "expr ::= LP select RP",
 /* 187 */ "expr ::= expr in_op LP select RP",
 /* 188 */ "expr ::= expr in_op nm paren_exprlist",
 /* 189 */ "expr ::= EXISTS LP select RP",
 /* 190 */ "expr ::= CASE expr_optional case_exprlist case_else END",
 /* 191 */ "case_exprlist ::= case_exprlist WHEN expr THEN expr",
 /* 192 */ "case_exprlist ::= WHEN expr THEN expr",
 /* 193 */ "case_else ::= ELSE expr",
 /* 194 */ "case_else ::=",
 /* 195 */ "exprlist ::=",
 /* 196 */ "nexprlist ::= nexprlist COMMA expr",
 /* 197 */ "nexprlist ::= expr",
 /* 198 */ "paren_exprlist ::=",
 /* 199 */ "paren_exprlist ::= LP exprlist RP",
 /* 200 */ "cmd ::= createkw uniqueflag INDEX ifnotexists nm ON nm LP sortlist RP",
 /* 201 */ "uniqueflag ::= UNIQUE",
 /* 202 */ "uniqueflag ::=",
 /* 203 */ "eidlist_opt ::=",
 /* 204 */ "eidlist_opt ::= LP eidlist RP",
 /* 205 */ "eidlist ::= eidlist COMMA nm",
 /* 206 */ "eidlist ::= nm",
 /* 207 */ "cmd ::= DROP INDEX ifexists nm ON fullname",
 /* 208 */ "cmd ::= SET SESSION nm EQ term",
 /* 209 */ "cmd ::= PRAGMA nm",
 /* 210 */ "cmd ::= PRAGMA nm LP nm RP",
 /* 211 */ "cmd ::= PRAGMA nm LP nm DOT nm RP",
 /* 212 */ "cmd ::= FUNCTION_KW expr",
 /* 213 */ "cmd ::= SHOW CREATE TABLE nm",
 /* 214 */ "cmd ::= SHOW CREATE TABLE",
 /* 215 */ "cmd ::= createkw trigger_decl BEGIN trigger_cmd_list END",
 /* 216 */ "trigger_decl ::= TRIGGER ifnotexists nm trigger_time trigger_event ON fullname foreach_clause when_clause",
 /* 217 */ "trigger_time ::= BEFORE",
 /* 218 */ "trigger_time ::= AFTER",
 /* 219 */ "trigger_time ::= INSTEAD OF",
 /* 220 */ "trigger_time ::=",
 /* 221 */ "trigger_event ::= DELETE|INSERT",
 /* 222 */ "trigger_event ::= UPDATE",
 /* 223 */ "trigger_event ::= UPDATE OF idlist",
 /* 224 */ "foreach_clause ::=",
 /* 225 */ "when_clause ::=",
 /* 226 */ "when_clause ::= WHEN expr",
 /* 227 */ "trigger_cmd_list ::= trigger_cmd_list trigger_cmd SEMI",
 /* 228 */ "trigger_cmd_list ::= trigger_cmd SEMI",
 /* 229 */ "trnm ::= nm DOT nm",
 /* 230 */ "tridxby ::= INDEXED BY nm",
 /* 231 */ "tridxby ::= NOT INDEXED",
 /* 232 */ "trigger_cmd ::= UPDATE orconf trnm tridxby SET setlist where_opt",
 /* 233 */ "trigger_cmd ::= insert_cmd INTO trnm idlist_opt select",
 /* 234 */ "trigger_cmd ::= DELETE FROM trnm tridxby where_opt",
 /* 235 */ "trigger_cmd ::= select",
 /* 236 */ "expr ::= RAISE LP IGNORE RP",
 /* 237 */ "expr ::= RAISE LP raisetype COMMA STRING RP",
 /* 238 */ "raisetype ::= ROLLBACK",
 /* 239 */ "raisetype ::= ABORT",
 /* 240 */ "raisetype ::= FAIL",
 /* 241 */ "cmd ::= DROP TRIGGER ifexists fullname",
 /* 242 */ "alter_table_start ::= ALTER TABLE fullname",
 /* 243 */ "alter_add_constraint ::= alter_table_start ADD CONSTRAINT nm",
 /* 244 */ "alter_add_column ::= alter_table_start ADD column_name",
 /* 245 */ "column_name ::= COLUMN nm",
 /* 246 */ "column_name ::= nm",
 /* 247 */ "alter_column_def ::= alter_add_column typedef",
 /* 248 */ "cmd ::= alter_add_constraint FOREIGN KEY LP eidlist RP REFERENCES nm eidlist_opt",
 /* 249 */ "cmd ::= alter_add_constraint CHECK LP expr RP",
 /* 250 */ "cmd ::= alter_add_constraint unique_spec LP sortlist RP",
 /* 251 */ "unique_spec ::= UNIQUE",
 /* 252 */ "unique_spec ::= PRIMARY KEY",
 /* 253 */ "cmd ::= alter_table_start RENAME TO nm",
 /* 254 */ "cmd ::= ALTER TABLE nm DROP CONSTRAINT nm",
 /* 255 */ "cmd ::= ALTER TABLE nm DROP CONSTRAINT nm FOREIGN KEY",
 /* 256 */ "cmd ::= ALTER TABLE nm DROP CONSTRAINT nm PRIMARY KEY",
 /* 257 */ "cmd ::= ALTER TABLE nm DROP CONSTRAINT nm UNIQUE",
 /* 258 */ "cmd ::= ALTER TABLE nm DROP CONSTRAINT nm CHECK",
 /* 259 */ "cmd ::= ALTER TABLE nm DROP CONSTRAINT nm DOT nm",
 /* 260 */ "cmd ::= ALTER TABLE nm DROP CONSTRAINT nm DOT nm FOREIGN KEY",
 /* 261 */ "cmd ::= ALTER TABLE nm DROP CONSTRAINT nm DOT nm CHECK",
 /* 262 */ "with ::=",
 /* 263 */ "with ::= WITH wqlist",
 /* 264 */ "with ::= WITH RECURSIVE wqlist",
 /* 265 */ "wqlist ::= nm eidlist_opt AS LP select RP",
 /* 266 */ "wqlist ::= wqlist COMMA nm eidlist_opt AS LP select RP",
 /* 267 */ "typedef ::= TEXT",
 /* 268 */ "typedef ::= STRING_KW",
 /* 269 */ "typedef ::= SCALAR",
 /* 270 */ "typedef ::= BOOL",
 /* 271 */ "typedef ::= BOOLEAN",
 /* 272 */ "typedef ::= VARBINARY",
 /* 273 */ "typedef ::= UUID",
 /* 274 */ "typedef ::= ANY",
 /* 275 */ "typedef ::= ARRAY",
 /* 276 */ "typedef ::= MAP",
 /* 277 */ "typedef ::= DATETIME",
 /* 278 */ "typedef ::= INTERVAL",
 /* 279 */ "char_len ::= LP INTEGER RP",
 /* 280 */ "typedef ::= VARCHAR char_len",
 /* 281 */ "number_typedef ::= NUMBER",
 /* 282 */ "number_typedef ::= DOUBLE",
 /* 283 */ "number_typedef ::= INT|INTEGER_KW",
 /* 284 */ "number_typedef ::= UNSIGNED",
 /* 285 */ "number_typedef ::= DECIMAL",
 /* 286 */ "input ::= ecmd",
 /* 287 */ "ecmd ::= explain cmdx SEMI",
 /* 288 */ "explain ::=",
 /* 289 */ "cmdx ::= cmd",
 /* 290 */ "savepoint_opt ::= SAVEPOINT",
 /* 291 */ "savepoint_opt ::=",
 /* 292 */ "cmd ::= create_table create_table_args with_opts create_table_end",
 /* 293 */ "create_table_args ::= LP columnlist RP",
 /* 294 */ "with_opts ::= WITH engine_opts",
 /* 295 */ "with_opts ::=",
 /* 296 */ "columnlist ::= columnlist COMMA tcons",
 /* 297 */ "columnlist ::= columnlist COMMA column_def create_column_end",
 /* 298 */ "columnlist ::= column_def create_column_end",
 /* 299 */ "column_def ::= column_name_and_type carglist",
 /* 300 */ "columnlist ::= tcons",
 /* 301 */ "carglist ::= carglist ccons",
 /* 302 */ "carglist ::=",
 /* 303 */ "resolvetype ::= raisetype",
 /* 304 */ "selectnowith ::= oneselect",
 /* 305 */ "oneselect ::= values",
 /* 306 */ "sclp ::= selcollist COMMA",
 /* 307 */ "as ::= ID|STRING",
 /* 308 */ "join_nm ::= ID|INDEXED",
 /* 309 */ "join_nm ::= JOIN_KW",
 /* 310 */ "expr ::= term",
 /* 311 */ "term ::= number",
 /* 312 */ "maplist ::= nmaplist",
 /* 313 */ "type_func ::= CHAR",
 /* 314 */ "exprlist ::= nexprlist",
 /* 315 */ "foreach_clause ::= FOR EACH ROW",
 /* 316 */ "trnm ::= nm",
 /* 317 */ "tridxby ::=",
 /* 318 */ "cmd ::= alter_column_def carglist create_column_end",
 /* 319 */ "typedef ::= number_typedef",
};
#endif /* NDEBUG */


#if YYSTACKDEPTH<=0
/*
** Try to increase the size of the parser stack.  Return the number
** of errors.  Return 0 on success.
*/
static int yyGrowStack(yyParser *p){
  int newSize;
  int idx;
  yyStackEntry *pNew;

  newSize = p->yystksz*2 + 100;
  idx = p->yytos ? (int)(p->yytos - p->yystack) : 0;
  if( p->yystack==&p->yystk0 ){
    pNew = malloc(newSize*sizeof(pNew[0]));
    if( pNew ) pNew[0] = p->yystk0;
  }else{
    pNew = realloc(p->yystack, newSize*sizeof(pNew[0]));
  }
  if( pNew ){
    p->yystack = pNew;
    p->yytos = &p->yystack[idx];
#ifndef NDEBUG
    if( yyTraceFILE ){
      fprintf(yyTraceFILE,"%sStack grows from %d to %d entries.\n",
              yyTracePrompt, p->yystksz, newSize);
    }
#endif
    p->yystksz = newSize;
  }
  return pNew==0; 
}
#endif

/* Datatype of the argument to the memory allocated passed as the
** second argument to sqlParserAlloc() below.  This can be changed by
** putting an appropriate #define in the %include section of the input
** grammar.
*/
#ifndef YYMALLOCARGTYPE
# define YYMALLOCARGTYPE size_t
#endif

/* 
** This function allocates a new parser.
** The only argument is a pointer to a function which works like
** malloc.
**
** Inputs:
** A pointer to the function used to allocate memory.
**
** Outputs:
** A pointer to a parser.  This pointer is used in subsequent calls
** to sqlParser and sqlParserFree.
*/
void *sqlParserAlloc(void *(*mallocProc)(YYMALLOCARGTYPE)){
  yyParser *pParser;
  pParser = (yyParser*)(*mallocProc)( (YYMALLOCARGTYPE)sizeof(yyParser) );
  if( pParser ){
#ifdef YYTRACKMAXSTACKDEPTH
    pParser->yyhwm = 0;
#endif
    pParser->is_fallback_failed = false;
#if YYSTACKDEPTH<=0
    pParser->yytos = NULL;
    pParser->yystack = NULL;
    pParser->yystksz = 0;
    if( yyGrowStack(pParser) ){
      pParser->yystack = &pParser->yystk0;
      pParser->yystksz = 1;
    }
#endif
#ifndef YYNOERRORRECOVERY
    pParser->yyerrcnt = -1;
#endif
    pParser->yytos = pParser->yystack;
    pParser->yystack[0].stateno = 0;
    pParser->yystack[0].major = 0;
  }
  return pParser;
}

/* The following function deletes the "minor type" or semantic value
** associated with a symbol.  The symbol can be either a terminal
** or nonterminal. "yymajor" is the symbol code, and "yypminor" is
** a pointer to the value to be deleted.  The code used to do the 
** deletions is derived from the %destructor and/or %token_destructor
** directives of the input grammar.
*/
static void yy_destructor(
  YYCODETYPE yymajor,     /* Type code for object to destroy */
  YYMINORTYPE *yypminor   /* The object to be destroyed */
){
  switch( yymajor ){
    /* Here is inserted the actions which take place when a
    ** terminal or non-terminal is destroyed.  This can happen
    ** when the symbol is popped from the stack during a
    ** reduce or during error processing or when a parser is 
    ** being destroyed before it is finished parsing.
    **
    ** Note: during a reduce, the only symbols destroyed are those
    ** which appear on the RHS of the rule, but which are *not* used
    ** inside the C code.
    */
/********* Begin destructor definitions ***************************************/
    case 197: /* term */
    case 198: /* expr */
    case 199: /* number */
{
#line 895 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sql_expr_delete((yypminor->yy34).pExpr);
#line 1625 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 202: /* eidlist_opt */
    case 203: /* col_list_with_autoinc */
    case 204: /* sortlist */
    case 205: /* eidlist */
    case 218: /* selcollist */
    case 221: /* groupby_opt */
    case 223: /* orderby_opt */
    case 226: /* nexprlist */
    case 227: /* exprlist */
    case 228: /* sclp */
    case 239: /* setlist */
    case 252: /* paren_exprlist */
    case 253: /* case_exprlist */
{
#line 1444 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sql_expr_list_delete((yypminor->yy414));
#line 1644 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 211: /* fullname */
    case 219: /* from */
    case 231: /* seltablist */
    case 232: /* stl_prefix */
{
#line 660 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sqlSrcListDelete((yypminor->yy499));
#line 1654 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 212: /* select */
    case 213: /* selectnowith */
    case 214: /* oneselect */
    case 225: /* values */
{
#line 431 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sql_select_delete((yypminor->yy247));
#line 1664 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 215: /* with */
    case 271: /* wqlist */
{
#line 1772 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sqlWithDelete((yypminor->yy163));
#line 1672 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 220: /* where_opt */
    case 222: /* having_opt */
    case 235: /* on_opt */
    case 254: /* case_else */
    case 261: /* when_clause */
{
#line 812 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sql_expr_delete((yypminor->yy262));
#line 1683 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 236: /* using_opt */
    case 238: /* idlist */
    case 241: /* idlist_opt */
{
#line 699 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sqlIdListDelete((yypminor->yy240));
#line 1692 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 242: /* getlist */
    case 243: /* maplist */
    case 244: /* nmaplist */
{
#line 1032 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sql_expr_list_delete((yypminor->yy222));
#line 1701 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 245: /* trim_operands */
{
#line 1076 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sql_expr_list_delete((yypminor->yy211));
#line 1708 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 247: /* expr_optional */
{
#line 1105 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sql_expr_delete((yypminor->yy231));
#line 1715 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 257: /* trigger_cmd_list */
    case 262: /* trigger_cmd */
{
#line 1562 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sqlDeleteTriggerStep((yypminor->yy115));
#line 1723 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
    case 259: /* trigger_event */
{
#line 1543 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
sqlIdListDelete((yypminor->yy122).b);
#line 1730 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
}
      break;
/********* End destructor definitions *****************************************/
    default:  break;   /* If no destructor action specified: do nothing */
  }
}

/*
** Pop the parser's stack once.
**
** If there is a destructor routine associated with the token which
** is popped from the stack, then call it.
*/
static void yy_pop_parser_stack(yyParser *pParser){
  yyStackEntry *yytos;
  assert( pParser->yytos!=0 );
  assert( pParser->yytos > pParser->yystack );
  yytos = pParser->yytos--;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sPopping %s\n",
      yyTracePrompt,
      yyTokenName[yytos->major]);
  }
#endif
  yy_destructor(yytos->major, &yytos->minor);
}

/* 
** Deallocate and destroy a parser.  Destructors are called for
** all stack elements before shutting the parser down.
**
** If the YYPARSEFREENEVERNULL macro exists (for example because it
** is defined in a %include section of the input grammar) then it is
** assumed that the input pointer is never NULL.
*/
void sqlParserFree(
  void *p,                    /* The parser to be deleted */
  void (*freeProc)(void*)     /* Function used to reclaim memory */
){
  yyParser *pParser = (yyParser*)p;
#ifndef YYPARSEFREENEVERNULL
  if( pParser==0 ) return;
#endif
  while( pParser->yytos>pParser->yystack ) yy_pop_parser_stack(pParser);
#if YYSTACKDEPTH<=0
  if( pParser->yystack!=&pParser->yystk0 ) free(pParser->yystack);
#endif
  (*freeProc)((void*)pParser);
}

/*
** Return the peak depth of the stack for a parser.
*/
#ifdef YYTRACKMAXSTACKDEPTH
int sqlParserStackPeak(void *p){
  yyParser *pParser = (yyParser*)p;
  return pParser->yyhwm;
}
#endif

/*
** Find the appropriate action for a parser given the terminal
** look-ahead token iLookAhead.
*/
static unsigned int yy_find_shift_action(
  yyParser *pParser,        /* The parser */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
  int stateno = pParser->yytos->stateno;
 
  if( stateno>=YY_MIN_REDUCE ) return stateno;
  assert( stateno <= YY_SHIFT_COUNT );
  do{
    i = yy_shift_ofst[stateno];
    assert( iLookAhead!=YYNOCODE );
    i += iLookAhead;
    if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
#ifdef YYFALLBACK
      YYCODETYPE iFallback = -1;            /* Fallback token */
      if( iLookAhead<sizeof(yyFallback)/sizeof(yyFallback[0])
             && (iFallback = yyFallback[iLookAhead])!=0 ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE, "%sFALLBACK %s => %s\n",
             yyTracePrompt, yyTokenName[iLookAhead], yyTokenName[iFallback]);
        }
#endif
        assert( yyFallback[iFallback]==0 ); /* Fallback loop must terminate */
        iLookAhead = iFallback;
        continue;
      } else if ( iFallback==0 ) {
        pParser->is_fallback_failed = true;
      }
#endif
#ifdef YYWILDCARD
      {
        int j = i - iLookAhead + YYWILDCARD;
        if( 
#if YY_SHIFT_MIN+YYWILDCARD<0
          j>=0 &&
#endif
#if YY_SHIFT_MAX+YYWILDCARD>=YY_ACTTAB_COUNT
          j<YY_ACTTAB_COUNT &&
#endif
          yy_lookahead[j]==YYWILDCARD && iLookAhead>0
        ){
#ifndef NDEBUG
          if( yyTraceFILE ){
            fprintf(yyTraceFILE, "%sWILDCARD %s => %s\n",
               yyTracePrompt, yyTokenName[iLookAhead],
               yyTokenName[YYWILDCARD]);
          }
#endif /* NDEBUG */
          return yy_action[j];
        }
      }
#endif /* YYWILDCARD */
      return yy_default[stateno];
    }else{
      return yy_action[i];
    }
  }while(1);
}

/*
** Find the appropriate action for a parser given the non-terminal
** look-ahead token iLookAhead.
*/
static int yy_find_reduce_action(
  int stateno,              /* Current state number */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
#ifdef YYERRORSYMBOL
  if( stateno>YY_REDUCE_COUNT ){
    return yy_default[stateno];
  }
#else
  assert( stateno<=YY_REDUCE_COUNT );
#endif
  i = yy_reduce_ofst[stateno];
  assert( i!=YY_REDUCE_USE_DFLT );
  assert( iLookAhead!=YYNOCODE );
  i += iLookAhead;
#ifdef YYERRORSYMBOL
  if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
    return yy_default[stateno];
  }
#else
  assert( i>=0 && i<YY_ACTTAB_COUNT );
  assert( yy_lookahead[i]==iLookAhead );
#endif
  return yy_action[i];
}

/*
** The following routine is called if the stack overflows.
*/
static void yyStackOverflow(yyParser *yypParser){
   sqlParserARG_FETCH;
#ifndef NDEBUG
   if( yyTraceFILE ){
     fprintf(yyTraceFILE,"%sStack Overflow!\n",yyTracePrompt);
   }
#endif
   while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
   /* Here code is inserted which will execute if the parser
   ** stack every overflows */
/******** Begin %stack_overflow code ******************************************/
#line 44 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  diag_set(ClientError, ER_SQL_STACK_OVERFLOW);
  pParse->is_aborted = true;
#line 1906 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
/******** End %stack_overflow code ********************************************/
   sqlParserARG_STORE; /* Suppress warning about unused %extra_argument var */
}

/*
** Print tracing information for a SHIFT action
*/
#ifndef NDEBUG
static void yyTraceShift(yyParser *yypParser, int yyNewState){
  if( yyTraceFILE ){
    if( yyNewState<YYNSTATE ){
      fprintf(yyTraceFILE,"%sShift '%s', go to state %d\n",
         yyTracePrompt,yyTokenName[yypParser->yytos->major],
         yyNewState);
    }else{
      fprintf(yyTraceFILE,"%sShift '%s'\n",
         yyTracePrompt,yyTokenName[yypParser->yytos->major]);
    }
  }
}
#else
# define yyTraceShift(X,Y)
#endif

/*
** Perform a shift action.
*/
static void yy_shift(
  yyParser *yypParser,          /* The parser to be shifted */
  int yyNewState,               /* The new state to shift in */
  int yyMajor,                  /* The major token to shift in */
  sqlParserTOKENTYPE yyMinor        /* The minor token to shift in */
){
  yyStackEntry *yytos;
  yypParser->yytos++;
#ifdef YYTRACKMAXSTACKDEPTH
  if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
    yypParser->yyhwm++;
    assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack) );
  }
#endif
#if YYSTACKDEPTH>0 
  if( yypParser->yytos>=&yypParser->yystack[YYSTACKDEPTH] ){
    yypParser->yytos--;
    yyStackOverflow(yypParser);
    return;
  }
#else
  if( yypParser->yytos>=&yypParser->yystack[yypParser->yystksz] ){
    if( yyGrowStack(yypParser) ){
      yypParser->yytos--;
      yyStackOverflow(yypParser);
      return;
    }
  }
#endif
  if( yyNewState > YY_MAX_SHIFT ){
    yyNewState += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
  }
  yytos = yypParser->yytos;
  yytos->stateno = (YYACTIONTYPE)yyNewState;
  yytos->major = (YYCODETYPE)yyMajor;
  yytos->minor.yy0 = yyMinor;
  yyTraceShift(yypParser, yyNewState);
}

/* The following table contains information about every rule that
** is used during the reduce.
*/
static const struct {
  YYCODETYPE lhs;         /* Symbol on the left-hand side of the rule */
  unsigned char nrhs;     /* Number of right-hand side symbols in the rule */
} yyRuleInfo[] = {
  { 174, 1 },
  { 175, 1 },
  { 175, 3 },
  { 177, 2 },
  { 177, 1 },
  { 177, 1 },
  { 177, 2 },
  { 177, 3 },
  { 177, 4 },
  { 180, 4 },
  { 184, 1 },
  { 185, 0 },
  { 185, 3 },
  { 187, 3 },
  { 183, 0 },
  { 191, 2 },
  { 190, 1 },
  { 179, 1 },
  { 196, 2 },
  { 196, 0 },
  { 195, 2 },
  { 195, 4 },
  { 195, 3 },
  { 195, 3 },
  { 195, 2 },
  { 195, 3 },
  { 195, 4 },
  { 195, 2 },
  { 195, 5 },
  { 195, 4 },
  { 195, 2 },
  { 194, 0 },
  { 194, 1 },
  { 188, 6 },
  { 188, 5 },
  { 188, 5 },
  { 188, 9 },
  { 200, 0 },
  { 200, 3 },
  { 207, 0 },
  { 207, 2 },
  { 208, 1 },
  { 208, 1 },
  { 177, 4 },
  { 177, 4 },
  { 210, 2 },
  { 210, 0 },
  { 177, 7 },
  { 177, 1 },
  { 212, 2 },
  { 213, 3 },
  { 216, 1 },
  { 216, 2 },
  { 216, 1 },
  { 214, 9 },
  { 225, 4 },
  { 225, 5 },
  { 217, 1 },
  { 217, 1 },
  { 217, 0 },
  { 228, 0 },
  { 218, 3 },
  { 218, 2 },
  { 218, 4 },
  { 229, 2 },
  { 229, 0 },
  { 230, 1 },
  { 230, 0 },
  { 219, 0 },
  { 219, 2 },
  { 232, 2 },
  { 232, 0 },
  { 231, 7 },
  { 231, 9 },
  { 231, 7 },
  { 231, 7 },
  { 211, 1 },
  { 233, 1 },
  { 233, 2 },
  { 233, 3 },
  { 233, 4 },
  { 235, 2 },
  { 235, 0 },
  { 234, 0 },
  { 234, 3 },
  { 234, 2 },
  { 236, 4 },
  { 236, 0 },
  { 223, 0 },
  { 223, 3 },
  { 204, 4 },
  { 204, 2 },
  { 203, 4 },
  { 203, 2 },
  { 201, 1 },
  { 201, 1 },
  { 201, 0 },
  { 221, 0 },
  { 221, 3 },
  { 222, 0 },
  { 222, 2 },
  { 224, 0 },
  { 224, 2 },
  { 224, 4 },
  { 224, 4 },
  { 177, 6 },
  { 177, 3 },
  { 220, 0 },
  { 220, 2 },
  { 177, 8 },
  { 239, 5 },
  { 239, 7 },
  { 239, 3 },
  { 239, 5 },
  { 177, 6 },
  { 177, 7 },
  { 240, 2 },
  { 240, 1 },
  { 241, 0 },
  { 241, 3 },
  { 238, 3 },
  { 238, 1 },
  { 198, 3 },
  { 197, 1 },
  { 198, 1 },
  { 198, 1 },
  { 198, 3 },
  { 197, 1 },
  { 197, 1 },
  { 197, 1 },
  { 197, 1 },
  { 197, 1 },
  { 199, 1 },
  { 199, 1 },
  { 199, 1 },
  { 198, 1 },
  { 198, 2 },
  { 198, 2 },
  { 198, 3 },
  { 198, 6 },
  { 198, 4 },
  { 242, 4 },
  { 242, 1 },
  { 198, 3 },
  { 198, 3 },
  { 243, 0 },
  { 244, 5 },
  { 244, 3 },
  { 198, 4 },
  { 245, 4 },
  { 245, 3 },
  { 245, 3 },
  { 245, 1 },
  { 247, 0 },
  { 247, 1 },
  { 246, 1 },
  { 246, 1 },
  { 246, 1 },
  { 198, 5 },
  { 198, 5 },
  { 198, 4 },
  { 198, 5 },
  { 198, 3 },
  { 198, 3 },
  { 198, 3 },
  { 198, 3 },
  { 198, 3 },
  { 198, 3 },
  { 198, 3 },
  { 198, 3 },
  { 249, 1 },
  { 249, 2 },
  { 198, 3 },
  { 198, 5 },
  { 198, 3 },
  { 198, 4 },
  { 198, 2 },
  { 198, 2 },
  { 198, 2 },
  { 198, 2 },
  { 250, 1 },
  { 250, 2 },
  { 198, 5 },
  { 251, 1 },
  { 251, 2 },
  { 198, 5 },
  { 198, 3 },
  { 198, 5 },
  { 198, 4 },
  { 198, 4 },
  { 198, 5 },
  { 253, 5 },
  { 253, 4 },
  { 254, 2 },
  { 254, 0 },
  { 227, 0 },
  { 226, 3 },
  { 226, 1 },
  { 252, 0 },
  { 252, 3 },
  { 177, 10 },
  { 255, 1 },
  { 255, 0 },
  { 202, 0 },
  { 202, 3 },
  { 205, 3 },
  { 205, 1 },
  { 177, 6 },
  { 177, 5 },
  { 177, 2 },
  { 177, 5 },
  { 177, 7 },
  { 177, 2 },
  { 177, 4 },
  { 177, 3 },
  { 177, 5 },
  { 256, 9 },
  { 258, 1 },
  { 258, 1 },
  { 258, 2 },
  { 258, 0 },
  { 259, 1 },
  { 259, 1 },
  { 259, 3 },
  { 260, 0 },
  { 261, 0 },
  { 261, 2 },
  { 257, 3 },
  { 257, 2 },
  { 263, 3 },
  { 264, 3 },
  { 264, 2 },
  { 262, 7 },
  { 262, 5 },
  { 262, 5 },
  { 262, 1 },
  { 198, 4 },
  { 198, 6 },
  { 209, 1 },
  { 209, 1 },
  { 209, 1 },
  { 177, 4 },
  { 265, 3 },
  { 266, 4 },
  { 267, 3 },
  { 268, 2 },
  { 268, 1 },
  { 269, 2 },
  { 177, 9 },
  { 177, 5 },
  { 177, 5 },
  { 270, 1 },
  { 270, 2 },
  { 177, 4 },
  { 177, 6 },
  { 177, 8 },
  { 177, 8 },
  { 177, 7 },
  { 177, 7 },
  { 177, 8 },
  { 177, 10 },
  { 177, 9 },
  { 215, 0 },
  { 215, 2 },
  { 215, 3 },
  { 271, 6 },
  { 271, 8 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 193, 1 },
  { 272, 3 },
  { 193, 2 },
  { 273, 1 },
  { 273, 1 },
  { 273, 1 },
  { 273, 1 },
  { 273, 1 },
  { 173, 1 },
  { 174, 3 },
  { 175, 0 },
  { 176, 1 },
  { 178, 1 },
  { 178, 0 },
  { 177, 4 },
  { 181, 3 },
  { 182, 2 },
  { 182, 0 },
  { 186, 3 },
  { 186, 4 },
  { 186, 2 },
  { 189, 2 },
  { 186, 1 },
  { 192, 2 },
  { 192, 0 },
  { 208, 1 },
  { 213, 1 },
  { 214, 1 },
  { 228, 2 },
  { 229, 1 },
  { 237, 1 },
  { 237, 1 },
  { 198, 1 },
  { 197, 1 },
  { 243, 1 },
  { 248, 1 },
  { 227, 1 },
  { 260, 3 },
  { 263, 1 },
  { 264, 0 },
  { 177, 3 },
  { 193, 1 },
};

static void yy_accept(yyParser*);  /* Forward Declaration */

/*
** Perform a reduce action and the shift that must immediately
** follow the reduce.
*/
static void yy_reduce(
  yyParser *yypParser,         /* The parser */
  unsigned int yyruleno        /* Number of the rule by which to reduce */
){
  int yygoto;                     /* The next state */
  int yyact;                      /* The next action */
  yyStackEntry *yymsp;            /* The top of the parser's stack */
  int yysize;                     /* Amount to pop the stack */
  sqlParserARG_FETCH;
  yymsp = yypParser->yytos;
#ifndef NDEBUG
  if( yyTraceFILE && yyruleno<(int)(sizeof(yyRuleName)/sizeof(yyRuleName[0])) ){
    yysize = yyRuleInfo[yyruleno].nrhs;
    fprintf(yyTraceFILE, "%sReduce [%s], go to state %d.\n", yyTracePrompt,
      yyRuleName[yyruleno], yymsp[-yysize].stateno);
  }
#endif /* NDEBUG */

  /* Check that the stack is large enough to grow by a single entry
  ** if the RHS of the rule is empty.  This ensures that there is room
  ** enough on the stack to push the LHS value */
  if( yyRuleInfo[yyruleno].nrhs==0 ){
#ifdef YYTRACKMAXSTACKDEPTH
    if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
      yypParser->yyhwm++;
      assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack));
    }
#endif
#if YYSTACKDEPTH>0 
    if( yypParser->yytos>=&yypParser->yystack[YYSTACKDEPTH-1] ){
      yyStackOverflow(yypParser);
      return;
    }
#else
    if( yypParser->yytos>=&yypParser->yystack[yypParser->yystksz-1] ){
      if( yyGrowStack(yypParser) ){
        yyStackOverflow(yypParser);
        return;
      }
      yymsp = yypParser->yytos;
    }
#endif
  }

  switch( yyruleno ){
  /* Beginning here are the reduction cases.  A typical example
  ** follows:
  **   case 0:
  **  #line <lineno> <grammarfile>
  **     { ... }           // User supplied code
  **  #line <lineno> <thisfile>
  **     break;
  */
/********** Begin reduce actions **********************************************/
        YYMINORTYPE yylhsminor;
      case 0: /* ecmd ::= SEMI */
#line 112 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  diag_set(ClientError, ER_SQL_STATEMENT_EMPTY);
  pParse->is_aborted = true;
}
#line 2369 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 1: /* explain ::= EXPLAIN */
#line 117 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ pParse->explain = 1; }
#line 2374 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 2: /* explain ::= EXPLAIN QUERY PLAN */
#line 118 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ pParse->explain = 2; }
#line 2379 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 3: /* cmd ::= START TRANSACTION */
#line 149 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sql_ast_init_start_transaction(pParse);
}
#line 2386 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 4: /* cmd ::= COMMIT */
#line 152 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sql_ast_init_commit(pParse);
}
#line 2393 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 5: /* cmd ::= ROLLBACK */
#line 155 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sql_ast_init_rollback(pParse);
}
#line 2400 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 6: /* cmd ::= SAVEPOINT nm */
#line 161 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sql_ast_init_savepoint(pParse, &yymsp[0].minor.yy0);
}
#line 2407 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 7: /* cmd ::= RELEASE savepoint_opt nm */
#line 164 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sql_ast_init_release_savepoint(pParse, &yymsp[0].minor.yy0);
}
#line 2414 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 8: /* cmd ::= ROLLBACK TO savepoint_opt nm */
#line 167 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sql_ast_init_rollback_to_savepoint(pParse, &yymsp[0].minor.yy0);
}
#line 2421 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 9: /* create_table ::= createkw TABLE ifnotexists nm */
#line 174 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_table_def_init(&pParse->create_table_def, &yymsp[0].minor.yy0, yymsp[-1].minor.yy40);
  create_ck_constraint_parse_def_init(&pParse->create_ck_constraint_parse_def);
  create_fk_constraint_parse_def_init(&pParse->create_fk_constraint_parse_def);
  pParse->create_table_def.new_space = sqlStartTable(pParse, &yymsp[0].minor.yy0);
  pParse->initiateTTrans = true;
}
#line 2432 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 10: /* createkw ::= CREATE */
#line 181 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{disableLookaside(pParse);}
#line 2437 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 11: /* ifnotexists ::= */
      case 31: /* autoinc ::= */ yytestcase(yyruleno==31);
      case 46: /* ifexists ::= */ yytestcase(yyruleno==46);
      case 59: /* distinct ::= */ yytestcase(yyruleno==59);
#line 184 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy40 = 0;}
#line 2445 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 12: /* ifnotexists ::= IF NOT EXISTS */
#line 185 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy40 = 1;}
#line 2450 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 13: /* engine_opts ::= ENGINE EQ STRING */
#line 192 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  /* Note that specifying engine clause overwrites default engine. */
  if (yymsp[0].minor.yy0.n > ENGINE_NAME_MAX) {
    diag_set(ClientError, ER_CREATE_SPACE,
             pParse->create_table_def.new_space->def->name,
             "space engine name is too long");
    pParse->is_aborted = true;
    return;
  }
  /* Need to dequote name. */
  char *normalized_name = sql_name_from_token(&yymsp[0].minor.yy0);
  memcpy(pParse->create_table_def.new_space->def->engine_name, normalized_name,
         strlen(normalized_name) + 1);
  sql_xfree(normalized_name);
}
#line 2469 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 14: /* create_table_end ::= */
#line 208 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ sqlEndTable(pParse); }
#line 2474 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 15: /* column_name_and_type ::= nm typedef */
#line 226 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_column_def_init(&pParse->create_column_def, NULL, &yymsp[-1].minor.yy0, &yymsp[0].minor.yy226);
  sql_create_column_start(pParse);
}
#line 2482 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 16: /* create_column_end ::= autoinc */
#line 231 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  uint32_t fieldno = pParse->create_column_def.space->def->field_count - 1;
  if (yymsp[0].minor.yy40 == 1 && sql_add_autoincrement(pParse, fieldno) != 0)
    return;
  if (pParse->create_table_def.new_space == NULL)
    sql_create_column_end(pParse);
}
#line 2493 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 17: /* nm ::= ID|INDEXED */
#line 273 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  if(yymsp[0].minor.yy0.isReserved) {
    diag_set(ClientError, ER_SQL_KEYWORD_IS_RESERVED, pParse->line_count,
             pParse->line_pos, yymsp[0].minor.yy0.n, yymsp[0].minor.yy0.z, yymsp[0].minor.yy0.n, yymsp[0].minor.yy0.z);
    pParse->is_aborted = true;
  }
}
#line 2504 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 18: /* cconsname ::= CONSTRAINT nm */
#line 289 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[-1].minor.yy103 = yymsp[0].minor.yy0; }
#line 2509 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 19: /* cconsname ::= */
#line 290 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[1].minor.yy103 = Token_nil; }
#line 2514 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 20: /* ccons ::= DEFAULT term */
      case 22: /* ccons ::= DEFAULT PLUS number */ yytestcase(yyruleno==22);
#line 291 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{sql_add_term_default(pParse, &yymsp[0].minor.yy34);}
#line 2520 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 21: /* ccons ::= DEFAULT LP expr RP */
#line 292 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  if (sql_expr_is_term(yymsp[-1].minor.yy34.pExpr))
    sql_add_term_default(pParse, &yymsp[-1].minor.yy34);
  else
    sql_add_func_default(pParse, &yymsp[-1].minor.yy34);
}
#line 2530 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 23: /* ccons ::= DEFAULT MINUS number */
#line 299 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  ExprSpan v;
  v.pExpr = sqlPExpr(pParse, TK_UMINUS, yymsp[0].minor.yy34.pExpr, 0);
  v.zStart = yymsp[-1].minor.yy0.z;
  v.zEnd = yymsp[0].minor.yy34.zEnd;
  sql_add_term_default(pParse, &v);
}
#line 2541 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 24: /* ccons ::= NULL onconf */
#line 310 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
    sql_column_add_nullable_action(pParse, ON_CONFLICT_ACTION_NONE);
    /* Trigger nullability mismatch error if required. */
    if (yymsp[0].minor.yy40 != ON_CONFLICT_ACTION_ABORT)
        sql_column_add_nullable_action(pParse, yymsp[0].minor.yy40);
}
#line 2551 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 25: /* ccons ::= NOT NULL onconf */
#line 316 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{sql_column_add_nullable_action(pParse, yymsp[0].minor.yy40);}
#line 2556 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 26: /* ccons ::= cconsname PRIMARY KEY sortorder */
#line 317 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_index_def_init(&pParse->create_index_def, NULL, &yymsp[-3].minor.yy103, NULL,
                        SQL_INDEX_TYPE_CONSTRAINT_PK, yymsp[0].minor.yy40, false);
  sqlAddPrimaryKey(pParse);
}
#line 2565 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 27: /* ccons ::= cconsname UNIQUE */
#line 322 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_index_def_init(&pParse->create_index_def, NULL, &yymsp[-1].minor.yy103, NULL,
                        SQL_INDEX_TYPE_CONSTRAINT_UNIQUE, SORT_ORDER_ASC,
                        false);
  sql_create_index(pParse);
}
#line 2575 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 28: /* ccons ::= cconsname CHECK LP expr RP */
#line 329 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_ck_def_init(&pParse->create_ck_def, NULL, &yymsp[-4].minor.yy103, &yymsp[-1].minor.yy34);
  sql_create_check_contraint(pParse, true);
}
#line 2583 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 29: /* ccons ::= cconsname REFERENCES nm eidlist_opt */
#line 334 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_fk_def_init(&pParse->create_fk_def, NULL, &yymsp[-3].minor.yy103, NULL, &yymsp[-1].minor.yy0, yymsp[0].minor.yy414);
  sql_create_foreign_key(pParse);
}
#line 2591 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 30: /* ccons ::= COLLATE ID|INDEXED */
#line 338 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{sqlAddCollateType(pParse, &yymsp[0].minor.yy0);}
#line 2596 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 32: /* autoinc ::= AUTOINCR */
#line 343 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = 1;}
#line 2601 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 33: /* tcons ::= cconsname PRIMARY KEY LP col_list_with_autoinc RP */
#line 346 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_index_def_init(&pParse->create_index_def, NULL, &yymsp[-5].minor.yy103, yymsp[-1].minor.yy414,
                        SQL_INDEX_TYPE_CONSTRAINT_PK, SORT_ORDER_ASC, false);
  sqlAddPrimaryKey(pParse);
}
#line 2610 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 34: /* tcons ::= cconsname UNIQUE LP sortlist RP */
#line 351 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_index_def_init(&pParse->create_index_def, NULL, &yymsp[-4].minor.yy103, yymsp[-1].minor.yy414,
                        SQL_INDEX_TYPE_CONSTRAINT_UNIQUE, SORT_ORDER_ASC,
                        false);
  sql_create_index(pParse);
}
#line 2620 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 35: /* tcons ::= cconsname CHECK LP expr RP */
#line 357 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_ck_def_init(&pParse->create_ck_def, NULL, &yymsp[-4].minor.yy103, &yymsp[-1].minor.yy34);
  sql_create_check_contraint(pParse, false);
}
#line 2628 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 36: /* tcons ::= cconsname FOREIGN KEY LP eidlist RP REFERENCES nm eidlist_opt */
#line 362 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_fk_def_init(&pParse->create_fk_def, NULL, &yymsp[-8].minor.yy103, yymsp[-4].minor.yy414, &yymsp[-1].minor.yy0, yymsp[0].minor.yy414);
  sql_create_foreign_key(pParse);
}
#line 2636 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 37: /* onconf ::= */
#line 374 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy40 = ON_CONFLICT_ACTION_ABORT;}
#line 2641 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 38: /* onconf ::= ON CONFLICT resolvetype */
#line 375 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy40 = yymsp[0].minor.yy40;}
#line 2646 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 39: /* orconf ::= */
#line 376 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy40 = ON_CONFLICT_ACTION_DEFAULT;}
#line 2651 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 40: /* orconf ::= OR resolvetype */
      case 116: /* insert_cmd ::= INSERT orconf */ yytestcase(yyruleno==116);
#line 377 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy40 = yymsp[0].minor.yy40;}
#line 2657 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 41: /* resolvetype ::= IGNORE */
#line 379 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = ON_CONFLICT_ACTION_IGNORE;}
#line 2662 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 42: /* resolvetype ::= REPLACE */
      case 117: /* insert_cmd ::= REPLACE */ yytestcase(yyruleno==117);
#line 380 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = ON_CONFLICT_ACTION_REPLACE;}
#line 2668 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 43: /* cmd ::= DROP TABLE ifexists fullname */
#line 385 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct Token t = Token_nil;
  drop_table_def_init(&pParse->drop_table_def, yymsp[0].minor.yy499, &t, yymsp[-1].minor.yy40);
  pParse->initiateTTrans = true;
  sql_drop_table(pParse);
}
#line 2678 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 44: /* cmd ::= DROP VIEW ifexists fullname */
#line 392 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct Token t = Token_nil;
  drop_view_def_init(&pParse->drop_view_def, yymsp[0].minor.yy499, &t, yymsp[-1].minor.yy40);
  pParse->initiateTTrans = true;
  sql_drop_table(pParse);
}
#line 2688 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 45: /* ifexists ::= IF EXISTS */
      case 181: /* between_op ::= NOT BETWEEN */ yytestcase(yyruleno==181);
      case 184: /* in_op ::= NOT IN */ yytestcase(yyruleno==184);
#line 400 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy40 = 1;}
#line 2695 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 47: /* cmd ::= createkw VIEW ifnotexists nm eidlist_opt AS select */
#line 406 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  if (!pParse->parse_only) {
    create_view_def_init(&pParse->create_view_def, &yymsp[-3].minor.yy0, &yymsp[-6].minor.yy0, yymsp[-2].minor.yy414, yymsp[0].minor.yy247, yymsp[-4].minor.yy40);
    pParse->initiateTTrans = true;
    sql_create_view(pParse);
  } else {
    sql_store_select(pParse, yymsp[0].minor.yy247);
  }
}
#line 2708 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 48: /* cmd ::= select */
#line 418 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  SelectDest dest = {SRT_Output, 0, 0, 0, 0, 0, 0};
  if(pParse->parse_only) {
    diag_set(ClientError, ER_SQL_PARSER_GENERIC,
             "Failed to parse SQL expression");
    pParse->is_aborted = true;
    return;
  }
  sqlSelect(pParse, yymsp[0].minor.yy247, &dest);
  sql_select_delete(yymsp[0].minor.yy247);
}
#line 2723 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 49: /* select ::= with selectnowith */
#line 463 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  Select *p = yymsp[0].minor.yy247;
  if( p ){
    p->pWith = yymsp[-1].minor.yy163;
    parserDoubleLinkSelect(pParse, p);
  }else{
    sqlWithDelete(yymsp[-1].minor.yy163);
  }
  yymsp[-1].minor.yy247 = p; /*A-overwrites-W*/
}
#line 2737 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 50: /* selectnowith ::= selectnowith multiselect_op oneselect */
#line 476 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  Select *pRhs = yymsp[0].minor.yy247;
  Select *pLhs = yymsp[-2].minor.yy247;
  if( pRhs && pRhs->pPrior ){
    SrcList *pFrom;
    Token x;
    x.n = 0;
    parserDoubleLinkSelect(pParse, pRhs);
    pFrom = sqlSrcListAppendFromTerm(pParse,0,0,&x,pRhs,0,0,false);
    pRhs = sqlSelectNew(pParse,0,pFrom,0,0,0,0,0,0,0);
  }
  if( pRhs ){
    pRhs->op = (u8)yymsp[-1].minor.yy40;
    pRhs->pPrior = pLhs;
    if( ALWAYS(pLhs) ) pLhs->selFlags &= ~SF_MultiValue;
    pRhs->selFlags &= ~SF_MultiValue;
    if( yymsp[-1].minor.yy40!=TK_ALL ) pParse->hasCompound = 1;
  }else{
    sql_select_delete(pLhs);
  }
  yymsp[-2].minor.yy247 = pRhs;
}
#line 2763 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 51: /* multiselect_op ::= UNION */
      case 53: /* multiselect_op ::= EXCEPT|INTERSECT */ yytestcase(yyruleno==53);
#line 499 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = yymsp[0].major; /*A-overwrites-OP*/}
#line 2769 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 52: /* multiselect_op ::= UNION ALL */
#line 500 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy40 = TK_ALL;}
#line 2774 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 54: /* oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt orderby_opt limit_opt */
#line 504 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
#ifdef SQL_DEBUG
  Token s = yymsp[-8].minor.yy0; /*A-overwrites-S*/
#endif
  yymsp[-8].minor.yy247 = sqlSelectNew(pParse,yymsp[-6].minor.yy414,yymsp[-5].minor.yy499,yymsp[-4].minor.yy262,yymsp[-3].minor.yy414,yymsp[-2].minor.yy262,yymsp[-1].minor.yy414,yymsp[-7].minor.yy40,yymsp[0].minor.yy248.pLimit,yymsp[0].minor.yy248.pOffset);
#ifdef SQL_DEBUG
  /* Populate the Select.zSelName[] string that is used to help with
  ** query planner debugging, to differentiate between multiple Select
  ** objects in a complex query.
  **
  ** If the SELECT keyword is immediately followed by a C-style comment
  ** then extract the first few alphanumeric characters from within that
  ** comment to be the zSelName value.  Otherwise, the label is #N where
  ** is an integer that is incremented with each SELECT statement seen.
  */
  if( yymsp[-8].minor.yy247!=0 ){
    const char *z = s.z+6;
    int i;
    sql_snprintf(sizeof(yymsp[-8].minor.yy247->zSelName), yymsp[-8].minor.yy247->zSelName, "#%d",
                     ++pParse->nSelect);
    while( z[0]==' ' ) z++;
    if( z[0]=='/' && z[1]=='*' ){
      z += 2;
      while( z[0]==' ' ) z++;
      for(i=0; sqlIsalnum(z[i]); i++){}
      sql_snprintf(sizeof(yymsp[-8].minor.yy247->zSelName), yymsp[-8].minor.yy247->zSelName, "%.*s", i, z);
    }
  }
#endif /* SQL_DEBUG */
}
#line 2808 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 55: /* values ::= VALUES LP nexprlist RP */
#line 538 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-3].minor.yy247 = sqlSelectNew(pParse,yymsp[-1].minor.yy414,0,0,0,0,0,SF_Values,0,0);
}
#line 2815 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 56: /* values ::= values COMMA LP exprlist RP */
#line 541 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  Select *pRight, *pLeft = yymsp[-4].minor.yy247;
  pRight = sqlSelectNew(pParse,yymsp[-1].minor.yy414,0,0,0,0,0,SF_Values|SF_MultiValue,0,0);
  if( ALWAYS(pLeft) ) pLeft->selFlags &= ~SF_MultiValue;
  if( pRight ){
    pRight->op = TK_ALL;
    pRight->pPrior = pLeft;
    yymsp[-4].minor.yy247 = pRight;
  }else{
    yymsp[-4].minor.yy247 = pLeft;
  }
}
#line 2831 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 57: /* distinct ::= DISTINCT */
#line 558 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = SF_Distinct;}
#line 2836 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 58: /* distinct ::= ALL */
#line 559 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = SF_All;}
#line 2841 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 60: /* sclp ::= */
      case 88: /* orderby_opt ::= */ yytestcase(yyruleno==88);
      case 97: /* groupby_opt ::= */ yytestcase(yyruleno==97);
      case 195: /* exprlist ::= */ yytestcase(yyruleno==195);
      case 198: /* paren_exprlist ::= */ yytestcase(yyruleno==198);
      case 203: /* eidlist_opt ::= */ yytestcase(yyruleno==203);
#line 572 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy414 = 0;}
#line 2851 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 61: /* selcollist ::= sclp expr as */
#line 573 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
   yymsp[-2].minor.yy414 = sql_expr_list_append(yymsp[-2].minor.yy414, yymsp[-1].minor.yy34.pExpr);
   if( yymsp[0].minor.yy0.n>0 ) sqlExprListSetName(pParse, yymsp[-2].minor.yy414, &yymsp[0].minor.yy0, 1);
   sqlExprListSetSpan(yymsp[-2].minor.yy414, &yymsp[-1].minor.yy34);
}
#line 2860 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 62: /* selcollist ::= sclp STAR */
#line 578 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-1].minor.yy414 = sql_expr_list_append(yymsp[-1].minor.yy414, sql_expr_new_anon(TK_ASTERISK));
}
#line 2867 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 63: /* selcollist ::= sclp nm DOT STAR */
#line 581 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct Expr *pLeft = sql_expr_new_dequoted(TK_ID, &yymsp[-2].minor.yy0);
  Expr *pRight = sqlPExpr(pParse, TK_ASTERISK, 0, 0);
  Expr *pDot = sqlPExpr(pParse, TK_DOT, pLeft, pRight);
  yymsp[-3].minor.yy414 = sql_expr_list_append(yymsp[-3].minor.yy414, pDot);
}
#line 2877 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 64: /* as ::= AS nm */
#line 592 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy0 = yymsp[0].minor.yy0;}
#line 2882 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 65: /* as ::= */
#line 594 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy0.n = 0; yymsp[1].minor.yy0.z = 0;}
#line 2887 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 66: /* seqscan ::= SEQSCAN */
      case 180: /* between_op ::= BETWEEN */ yytestcase(yyruleno==180);
      case 183: /* in_op ::= IN */ yytestcase(yyruleno==183);
#line 597 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = 0;}
#line 2894 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 67: /* seqscan ::= */
#line 598 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy40 = 1;}
#line 2899 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 68: /* from ::= */
#line 609 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy499 = sql_xmalloc0(sizeof(*yymsp[1].minor.yy499));}
#line 2904 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 69: /* from ::= FROM seltablist */
#line 610 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-1].minor.yy499 = yymsp[0].minor.yy499;
  sqlSrcListShiftJoinType(yymsp[-1].minor.yy499);
}
#line 2912 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 70: /* stl_prefix ::= seltablist joinop */
#line 618 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
   if( ALWAYS(yymsp[-1].minor.yy499 && yymsp[-1].minor.yy499->nSrc>0) ) yymsp[-1].minor.yy499->a[yymsp[-1].minor.yy499->nSrc-1].fg.jointype = (u8)yymsp[0].minor.yy40;
}
#line 2919 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 71: /* stl_prefix ::= */
#line 621 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy499 = 0;}
#line 2924 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 72: /* seltablist ::= stl_prefix seqscan nm as indexed_opt on_opt using_opt */
#line 623 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-6].minor.yy499 = sqlSrcListAppendFromTerm(pParse,yymsp[-6].minor.yy499,&yymsp[-4].minor.yy0,&yymsp[-3].minor.yy0,0,yymsp[-1].minor.yy262,yymsp[0].minor.yy240,yymsp[-5].minor.yy40);
  sqlSrcListIndexedBy(yymsp[-6].minor.yy499, &yymsp[-2].minor.yy0);
}
#line 2932 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 73: /* seltablist ::= stl_prefix seqscan nm LP exprlist RP as on_opt using_opt */
#line 628 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-8].minor.yy499 = sqlSrcListAppendFromTerm(pParse,yymsp[-8].minor.yy499,&yymsp[-6].minor.yy0,&yymsp[-2].minor.yy0,0,yymsp[-1].minor.yy262,yymsp[0].minor.yy240,yymsp[-7].minor.yy40);
  sqlSrcListFuncArgs(yymsp[-8].minor.yy499, yymsp[-4].minor.yy414);
}
#line 2940 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 74: /* seltablist ::= stl_prefix LP select RP as on_opt using_opt */
#line 633 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-6].minor.yy499 = sqlSrcListAppendFromTerm(pParse,yymsp[-6].minor.yy499,0,&yymsp[-2].minor.yy0,yymsp[-4].minor.yy247,yymsp[-1].minor.yy262,yymsp[0].minor.yy240,false);
}
#line 2947 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 75: /* seltablist ::= stl_prefix LP seltablist RP as on_opt using_opt */
#line 637 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  if( yymsp[-6].minor.yy499==0 && yymsp[-2].minor.yy0.n==0 && yymsp[-1].minor.yy262==0 && yymsp[0].minor.yy240==0 ){
    yymsp[-6].minor.yy499 = yymsp[-4].minor.yy499;
  }else if( yymsp[-4].minor.yy499->nSrc==1 ){
    yymsp[-6].minor.yy499 = sqlSrcListAppendFromTerm(pParse,yymsp[-6].minor.yy499,0,&yymsp[-2].minor.yy0,0,yymsp[-1].minor.yy262,yymsp[0].minor.yy240,false);
    if( yymsp[-6].minor.yy499 ){
      struct SrcList_item *pNew = &yymsp[-6].minor.yy499->a[yymsp[-6].minor.yy499->nSrc-1];
      struct SrcList_item *pOld = yymsp[-4].minor.yy499->a;
      pNew->zName = pOld->zName;
      pNew->pSelect = pOld->pSelect;
      pOld->zName =  0;
      pOld->pSelect = 0;
    }
    sqlSrcListDelete(yymsp[-4].minor.yy499);
  }else{
    Select *pSubquery;
    sqlSrcListShiftJoinType(yymsp[-4].minor.yy499);
    pSubquery = sqlSelectNew(pParse,0,yymsp[-4].minor.yy499,0,0,0,0,SF_NestedFrom,0,0);
    yymsp[-6].minor.yy499 = sqlSrcListAppendFromTerm(pParse,yymsp[-6].minor.yy499,0,&yymsp[-2].minor.yy0,pSubquery,yymsp[-1].minor.yy262,yymsp[0].minor.yy240,false);
  }
}
#line 2972 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 76: /* fullname ::= nm */
#line 661 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  /* yylhsminor.yy499-overwrites-yymsp[0].minor.yy0. */
  yylhsminor.yy499 = sql_src_list_append(NULL ,&yymsp[0].minor.yy0);
}
#line 2980 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy499 = yylhsminor.yy499;
        break;
      case 77: /* joinop ::= COMMA|JOIN */
#line 670 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy40 = JT_INNER; }
#line 2986 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 78: /* joinop ::= JOIN_KW JOIN */
#line 672 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy40 = sqlJoinType(pParse,&yymsp[-1].minor.yy0,0,0);  /*X-overwrites-A*/}
#line 2991 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 79: /* joinop ::= JOIN_KW join_nm JOIN */
#line 674 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy40 = sqlJoinType(pParse,&yymsp[-2].minor.yy0,&yymsp[-1].minor.yy0,0); /*X-overwrites-A*/}
#line 2996 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 80: /* joinop ::= JOIN_KW join_nm join_nm JOIN */
#line 676 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-3].minor.yy40 = sqlJoinType(pParse,&yymsp[-3].minor.yy0,&yymsp[-2].minor.yy0,&yymsp[-1].minor.yy0);/*X-overwrites-A*/}
#line 3001 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 81: /* on_opt ::= ON expr */
      case 100: /* having_opt ::= HAVING expr */ yytestcase(yyruleno==100);
      case 108: /* where_opt ::= WHERE expr */ yytestcase(yyruleno==108);
      case 193: /* case_else ::= ELSE expr */ yytestcase(yyruleno==193);
#line 680 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy262 = yymsp[0].minor.yy34.pExpr;}
#line 3009 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 82: /* on_opt ::= */
      case 99: /* having_opt ::= */ yytestcase(yyruleno==99);
      case 107: /* where_opt ::= */ yytestcase(yyruleno==107);
      case 194: /* case_else ::= */ yytestcase(yyruleno==194);
#line 681 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy262 = 0;}
#line 3017 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 83: /* indexed_opt ::= */
#line 694 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy0.z=0; yymsp[1].minor.yy0.n=0;}
#line 3022 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 84: /* indexed_opt ::= INDEXED BY nm */
#line 695 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy0 = yymsp[0].minor.yy0;}
#line 3027 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 85: /* indexed_opt ::= NOT INDEXED */
#line 696 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy0.z=0; yymsp[-1].minor.yy0.n=1;}
#line 3032 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 86: /* using_opt ::= USING LP idlist RP */
#line 700 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-3].minor.yy240 = yymsp[-1].minor.yy240;}
#line 3037 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 87: /* using_opt ::= */
      case 118: /* idlist_opt ::= */ yytestcase(yyruleno==118);
#line 701 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy240 = 0;}
#line 3043 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 89: /* orderby_opt ::= ORDER BY sortlist */
      case 98: /* groupby_opt ::= GROUP BY nexprlist */ yytestcase(yyruleno==98);
#line 715 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy414 = yymsp[0].minor.yy414;}
#line 3049 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 90: /* sortlist ::= sortlist COMMA expr sortorder */
#line 716 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-3].minor.yy414 = sql_expr_list_append(yymsp[-3].minor.yy414, yymsp[-1].minor.yy34.pExpr);
  sqlExprListSetSortOrder(yymsp[-3].minor.yy414,yymsp[0].minor.yy40);
}
#line 3057 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 91: /* sortlist ::= expr sortorder */
#line 720 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  /* yylhsminor.yy414-overwrites-yymsp[-1].minor.yy34. */
  yylhsminor.yy414 = sql_expr_list_append(NULL, yymsp[-1].minor.yy34.pExpr);
  sqlExprListSetSortOrder(yylhsminor.yy414,yymsp[0].minor.yy40);
}
#line 3066 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-1].minor.yy414 = yylhsminor.yy414;
        break;
      case 92: /* col_list_with_autoinc ::= col_list_with_autoinc COMMA expr autoinc */
#line 734 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  uint32_t fieldno;
  if (yymsp[0].minor.yy40 == 1) {
    if (sql_fieldno_by_name(pParse, yymsp[-1].minor.yy34.pExpr, &fieldno) != 0)
      return;
    if (sql_add_autoincrement(pParse, fieldno) != 0)
      return;
  }
  yymsp[-3].minor.yy414 = sql_expr_list_append(yymsp[-3].minor.yy414, yymsp[-1].minor.yy34.pExpr);
}
#line 3081 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 93: /* col_list_with_autoinc ::= expr autoinc */
#line 745 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  if (yymsp[0].minor.yy40 == 1) {
    uint32_t fieldno = 0;
    if (sql_fieldno_by_name(pParse, yymsp[-1].minor.yy34.pExpr, &fieldno) != 0)
      return;
    if (sql_add_autoincrement(pParse, fieldno) != 0)
      return;
  }
  /* yylhsminor.yy414-overwrites-yymsp[-1].minor.yy34. */
  yylhsminor.yy414 = sql_expr_list_append(NULL, yymsp[-1].minor.yy34.pExpr);
}
#line 3096 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-1].minor.yy414 = yylhsminor.yy414;
        break;
      case 94: /* sortorder ::= ASC */
#line 759 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = SORT_ORDER_ASC;}
#line 3102 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 95: /* sortorder ::= DESC */
#line 760 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = SORT_ORDER_DESC;}
#line 3107 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 96: /* sortorder ::= */
#line 761 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy40 = SORT_ORDER_UNDEF;}
#line 3112 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 101: /* limit_opt ::= */
#line 786 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy248.pLimit = 0; yymsp[1].minor.yy248.pOffset = 0;}
#line 3117 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 102: /* limit_opt ::= LIMIT expr */
#line 787 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy248.pLimit = yymsp[0].minor.yy34.pExpr; yymsp[-1].minor.yy248.pOffset = 0;}
#line 3122 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 103: /* limit_opt ::= LIMIT expr OFFSET expr */
#line 789 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-3].minor.yy248.pLimit = yymsp[-2].minor.yy34.pExpr; yymsp[-3].minor.yy248.pOffset = yymsp[0].minor.yy34.pExpr;}
#line 3127 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 104: /* limit_opt ::= LIMIT expr COMMA expr */
#line 791 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-3].minor.yy248.pOffset = yymsp[-2].minor.yy34.pExpr; yymsp[-3].minor.yy248.pLimit = yymsp[0].minor.yy34.pExpr;}
#line 3132 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 105: /* cmd ::= with DELETE FROM fullname indexed_opt where_opt */
#line 795 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sqlWithPush(pParse, yymsp[-5].minor.yy163, 1);
  sqlSrcListIndexedBy(yymsp[-2].minor.yy499, &yymsp[-1].minor.yy0);
  sqlSubProgramsRemaining = SQL_MAX_COMPILING_TRIGGERS;
  /* Instruct SQL to initate Tarantool's transaction.  */
  pParse->initiateTTrans = true;
  sql_table_delete_from(pParse,yymsp[-2].minor.yy499,yymsp[0].minor.yy262);
}
#line 3144 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 106: /* cmd ::= TRUNCATE TABLE fullname */
#line 806 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_table_truncate(pParse, yymsp[0].minor.yy499);
}
#line 3152 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 109: /* cmd ::= with UPDATE orconf fullname indexed_opt SET setlist where_opt */
#line 820 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sqlWithPush(pParse, yymsp[-7].minor.yy163, 1);
  sqlSrcListIndexedBy(yymsp[-4].minor.yy499, &yymsp[-3].minor.yy0);
  if (yymsp[-1].minor.yy414 != NULL && yymsp[-1].minor.yy414->nExpr > SQL_MAX_COLUMN) {
    diag_set(ClientError, ER_SQL_PARSER_LIMIT, "The number of columns in set "\
             "list", yymsp[-1].minor.yy414->nExpr, SQL_MAX_COLUMN);
    pParse->is_aborted = true;
  }
  sqlSubProgramsRemaining = SQL_MAX_COMPILING_TRIGGERS;
  /* Instruct SQL to initate Tarantool's transaction.  */
  pParse->initiateTTrans = true;
  sqlUpdate(pParse,yymsp[-4].minor.yy499,yymsp[-1].minor.yy414,yymsp[0].minor.yy262,yymsp[-5].minor.yy40);
}
#line 3169 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 110: /* setlist ::= setlist COMMA nm EQ expr */
#line 837 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-4].minor.yy414 = sql_expr_list_append(yymsp[-4].minor.yy414, yymsp[0].minor.yy34.pExpr);
  sqlExprListSetName(pParse, yymsp[-4].minor.yy414, &yymsp[-2].minor.yy0, 1);
}
#line 3177 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 111: /* setlist ::= setlist COMMA LP idlist RP EQ expr */
#line 841 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-6].minor.yy414 = sqlExprListAppendVector(pParse, yymsp[-6].minor.yy414, yymsp[-3].minor.yy240, yymsp[0].minor.yy34.pExpr);
}
#line 3184 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 112: /* setlist ::= nm EQ expr */
#line 844 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy414 = sql_expr_list_append(NULL, yymsp[0].minor.yy34.pExpr);
  sqlExprListSetName(pParse, yylhsminor.yy414, &yymsp[-2].minor.yy0, 1);
}
#line 3192 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-2].minor.yy414 = yylhsminor.yy414;
        break;
      case 113: /* setlist ::= LP idlist RP EQ expr */
#line 848 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-4].minor.yy414 = sqlExprListAppendVector(pParse, 0, yymsp[-3].minor.yy240, yymsp[0].minor.yy34.pExpr);
}
#line 3200 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 114: /* cmd ::= with insert_cmd INTO fullname idlist_opt select */
#line 854 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sqlWithPush(pParse, yymsp[-5].minor.yy163, 1);
  sqlSubProgramsRemaining = SQL_MAX_COMPILING_TRIGGERS;
  /* Instruct SQL to initate Tarantool's transaction.  */
  pParse->initiateTTrans = true;
  sqlInsert(pParse, yymsp[-2].minor.yy499, yymsp[0].minor.yy247, yymsp[-1].minor.yy240, yymsp[-4].minor.yy40);
}
#line 3211 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 115: /* cmd ::= with insert_cmd INTO fullname idlist_opt DEFAULT VALUES */
#line 862 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sqlWithPush(pParse, yymsp[-6].minor.yy163, 1);
  sqlSubProgramsRemaining = SQL_MAX_COMPILING_TRIGGERS;
  /* Instruct SQL to initate Tarantool's transaction.  */
  pParse->initiateTTrans = true;
  sqlInsert(pParse, yymsp[-3].minor.yy499, 0, yymsp[-2].minor.yy240, yymsp[-5].minor.yy40);
}
#line 3222 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 119: /* idlist_opt ::= LP idlist RP */
#line 880 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy240 = yymsp[-1].minor.yy240;}
#line 3227 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 120: /* idlist ::= idlist COMMA nm */
#line 881 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-2].minor.yy240 = sql_id_list_append(yymsp[-2].minor.yy240, &yymsp[0].minor.yy0);
}
#line 3234 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 121: /* idlist ::= nm */
#line 884 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  /* yylhsminor.yy240-overwrites-yymsp[0].minor.yy0. */
  yylhsminor.yy240 = sql_id_list_append(NULL, &yymsp[0].minor.yy0);
}
#line 3242 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy240 = yylhsminor.yy240;
        break;
      case 122: /* expr ::= LP expr RP */
#line 964 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanSet(&yymsp[-2].minor.yy34,&yymsp[-2].minor.yy0,&yymsp[0].minor.yy0); /*A-overwrites-B*/  yymsp[-2].minor.yy34.pExpr = yymsp[-1].minor.yy34.pExpr;}
#line 3248 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 123: /* term ::= NULL */
      case 127: /* term ::= BLOB */ yytestcase(yyruleno==127);
      case 128: /* term ::= STRING */ yytestcase(yyruleno==128);
      case 129: /* term ::= FALSE */ yytestcase(yyruleno==129);
      case 130: /* term ::= TRUE */ yytestcase(yyruleno==130);
      case 131: /* term ::= UNKNOWN */ yytestcase(yyruleno==131);
      case 132: /* number ::= FLOAT */ yytestcase(yyruleno==132);
      case 133: /* number ::= DECIMAL */ yytestcase(yyruleno==133);
#line 965 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanExpr(&yymsp[0].minor.yy34, yymsp[0].major, yymsp[0].minor.yy0);/*A-overwrites-X*/}
#line 3260 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 124: /* expr ::= ID|INDEXED */
      case 125: /* expr ::= JOIN_KW */ yytestcase(yyruleno==125);
#line 966 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanExpr(&yymsp[0].minor.yy34, TK_ID, yymsp[0].minor.yy0); /*A-overwrites-X*/}
#line 3266 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 126: /* expr ::= nm DOT nm */
#line 968 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct Expr *temp1 = sql_expr_new_dequoted(TK_ID, &yymsp[-2].minor.yy0);
  struct Expr *temp2 = sql_expr_new_dequoted(TK_ID, &yymsp[0].minor.yy0);
  spanSet(&yymsp[-2].minor.yy34,&yymsp[-2].minor.yy0,&yymsp[0].minor.yy0); /*A-overwrites-X*/
  yymsp[-2].minor.yy34.pExpr = sqlPExpr(pParse, TK_DOT, temp1, temp2);
}
#line 3276 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 134: /* number ::= INTEGER */
#line 982 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy34.pExpr = sql_expr_new_dequoted(TK_INTEGER, &yymsp[0].minor.yy0);
  yylhsminor.yy34.pExpr->type = FIELD_TYPE_INTEGER;
  yylhsminor.yy34.zStart = yymsp[0].minor.yy0.z;
  yylhsminor.yy34.zEnd = yymsp[0].minor.yy0.z + yymsp[0].minor.yy0.n;
  yylhsminor.yy34.pExpr->flags |= EP_Leaf;
}
#line 3287 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy34 = yylhsminor.yy34;
        break;
      case 135: /* expr ::= VARNUM */
#line 989 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy34.pExpr = expr_new_variable(pParse, &yymsp[0].minor.yy0, NULL);
  spanSet(&yylhsminor.yy34, &yymsp[0].minor.yy0, &yymsp[0].minor.yy0);
}
#line 3296 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy34 = yylhsminor.yy34;
        break;
      case 136: /* expr ::= COLON|VARIABLE ID|INDEXED */
      case 137: /* expr ::= COLON|VARIABLE INTEGER */ yytestcase(yyruleno==137);
#line 993 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy34.pExpr = expr_new_variable(pParse, &yymsp[-1].minor.yy0, &yymsp[0].minor.yy0);
  spanSet(&yylhsminor.yy34, &yymsp[-1].minor.yy0, &yymsp[0].minor.yy0);
}
#line 3306 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-1].minor.yy34 = yylhsminor.yy34;
        break;
      case 138: /* expr ::= expr COLLATE ID|INDEXED */
#line 1001 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-2].minor.yy34.pExpr = sqlExprAddCollateToken(yymsp[-2].minor.yy34.pExpr, &yymsp[0].minor.yy0, 1);
  yymsp[-2].minor.yy34.zEnd = &yymsp[0].minor.yy0.z[yymsp[0].minor.yy0.n];
}
#line 3315 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 139: /* expr ::= CAST LP expr AS typedef RP */
#line 1006 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  spanSet(&yymsp[-5].minor.yy34,&yymsp[-5].minor.yy0,&yymsp[0].minor.yy0); /*A-overwrites-X*/
  yymsp[-5].minor.yy34.pExpr = sql_expr_new_dequoted(TK_CAST, NULL);
  yymsp[-5].minor.yy34.pExpr->type = yymsp[-1].minor.yy226.type;
  sqlExprAttachSubtrees(yymsp[-5].minor.yy34.pExpr, yymsp[-3].minor.yy34.pExpr, 0);
}
#line 3325 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 140: /* expr ::= expr LB getlist RB */
#line 1013 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct Expr *expr = sql_expr_new_anon(TK_GETITEM);
  yymsp[-1].minor.yy222 = sql_expr_list_append(yymsp[-1].minor.yy222, yymsp[-3].minor.yy34.pExpr);
  expr->x.pList = yymsp[-1].minor.yy222;
  expr->type = FIELD_TYPE_ANY;
  sqlExprSetHeightAndFlags(pParse, expr);
  yylhsminor.yy34.pExpr = expr;
  yylhsminor.yy34.zStart = yymsp[-3].minor.yy34.zStart;
  yylhsminor.yy34.zEnd = &yymsp[0].minor.yy0.z[yymsp[0].minor.yy0.n];
}
#line 3339 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-3].minor.yy34 = yylhsminor.yy34;
        break;
      case 141: /* getlist ::= getlist RB LB expr */
#line 1024 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-3].minor.yy222 = sql_expr_list_append(yymsp[-3].minor.yy222, yymsp[0].minor.yy34.pExpr);
}
#line 3347 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 142: /* getlist ::= expr */
#line 1027 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy222 = sql_expr_list_append(NULL, yymsp[0].minor.yy34.pExpr);
}
#line 3354 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy222 = yylhsminor.yy222;
        break;
      case 143: /* expr ::= LB exprlist RB */
#line 1034 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct Expr *expr = sql_expr_new_anon(TK_ARRAY);
  expr->x.pList = yymsp[-1].minor.yy414;
  expr->type = FIELD_TYPE_ARRAY;
  sqlExprSetHeightAndFlags(pParse, expr);
  yylhsminor.yy34.pExpr = expr;
  spanSet(&yylhsminor.yy34, &yymsp[-2].minor.yy0, &yymsp[0].minor.yy0);
}
#line 3367 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-2].minor.yy34 = yylhsminor.yy34;
        break;
      case 144: /* expr ::= LCB maplist RCB */
#line 1043 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct Expr *expr = sql_expr_new_anon(TK_MAP);
  expr->x.pList = yymsp[-1].minor.yy222;
  expr->type = FIELD_TYPE_MAP;
  sqlExprSetHeightAndFlags(pParse, expr);
  yylhsminor.yy34.pExpr = expr;
  spanSet(&yylhsminor.yy34, &yymsp[-2].minor.yy0, &yymsp[0].minor.yy0);
}
#line 3380 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-2].minor.yy34 = yylhsminor.yy34;
        break;
      case 145: /* maplist ::= */
#line 1053 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[1].minor.yy222 = NULL;
}
#line 3388 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 146: /* nmaplist ::= nmaplist COMMA expr COLON expr */
#line 1056 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-4].minor.yy222 = sql_expr_list_append(yymsp[-4].minor.yy222, yymsp[-2].minor.yy34.pExpr);
  yymsp[-4].minor.yy222 = sql_expr_list_append(yymsp[-4].minor.yy222, yymsp[0].minor.yy34.pExpr);
}
#line 3396 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 147: /* nmaplist ::= expr COLON expr */
#line 1060 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy222 = sql_expr_list_append(NULL, yymsp[-2].minor.yy34.pExpr);
  yylhsminor.yy222 = sql_expr_list_append(yylhsminor.yy222, yymsp[0].minor.yy34.pExpr);
}
#line 3404 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-2].minor.yy222 = yylhsminor.yy222;
        break;
      case 148: /* expr ::= TRIM LP trim_operands RP */
#line 1070 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy34.pExpr = sqlExprFunction(pParse, yymsp[-1].minor.yy211, &yymsp[-3].minor.yy0);
  spanSet(&yylhsminor.yy34, &yymsp[-3].minor.yy0, &yymsp[0].minor.yy0);
}
#line 3413 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-3].minor.yy34 = yylhsminor.yy34;
        break;
      case 149: /* trim_operands ::= trim_specification expr FROM expr */
#line 1078 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy211 = sql_expr_list_append(NULL, yymsp[0].minor.yy34.pExpr);
  struct Expr *p = sql_expr_new_dequoted(TK_INTEGER, &sqlIntTokens[yymsp[-3].minor.yy345]);
  yylhsminor.yy211 = sql_expr_list_append(yylhsminor.yy211, p);
  yylhsminor.yy211 = sql_expr_list_append(yylhsminor.yy211, yymsp[-2].minor.yy34.pExpr);
}
#line 3424 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-3].minor.yy211 = yylhsminor.yy211;
        break;
      case 150: /* trim_operands ::= trim_specification FROM expr */
#line 1085 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy211 = sql_expr_list_append(NULL, yymsp[0].minor.yy34.pExpr);
  struct Expr *p = sql_expr_new_dequoted(TK_INTEGER, &sqlIntTokens[yymsp[-2].minor.yy345]);
  yylhsminor.yy211 = sql_expr_list_append(yylhsminor.yy211, p);
}
#line 3434 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-2].minor.yy211 = yylhsminor.yy211;
        break;
      case 151: /* trim_operands ::= expr FROM expr */
#line 1091 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy211 = sql_expr_list_append(NULL, yymsp[0].minor.yy34.pExpr);
  struct Expr *p = sql_expr_new_dequoted(TK_INTEGER, &sqlIntTokens[TRIM_BOTH]);
  yylhsminor.yy211 = sql_expr_list_append(yylhsminor.yy211, p);
  yylhsminor.yy211 = sql_expr_list_append(yylhsminor.yy211, yymsp[-2].minor.yy34.pExpr);
}
#line 3445 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-2].minor.yy211 = yylhsminor.yy211;
        break;
      case 152: /* trim_operands ::= expr */
#line 1098 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy211 = sql_expr_list_append(NULL, yymsp[0].minor.yy34.pExpr);
  struct Expr *p = sql_expr_new_dequoted(TK_INTEGER, &sqlIntTokens[TRIM_BOTH]);
  yylhsminor.yy211 = sql_expr_list_append(yylhsminor.yy211, p);
}
#line 3455 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy211 = yylhsminor.yy211;
        break;
      case 153: /* expr_optional ::= */
#line 1107 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[1].minor.yy231 = NULL; }
#line 3461 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 154: /* expr_optional ::= expr */
#line 1108 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yylhsminor.yy231 = yymsp[0].minor.yy34.pExpr; }
#line 3466 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy231 = yylhsminor.yy231;
        break;
      case 155: /* trim_specification ::= LEADING */
#line 1112 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy345 = TRIM_LEADING; }
#line 3472 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 156: /* trim_specification ::= TRAILING */
#line 1113 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy345 = TRIM_TRAILING; }
#line 3477 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 157: /* trim_specification ::= BOTH */
#line 1114 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy345 = TRIM_BOTH; }
#line 3482 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 158: /* expr ::= ID|INDEXED LP distinct exprlist RP */
      case 159: /* expr ::= type_func LP distinct exprlist RP */ yytestcase(yyruleno==159);
#line 1116 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  if (yymsp[-1].minor.yy414 != NULL && yymsp[-1].minor.yy414->nExpr > SQL_MAX_FUNCTION_ARG){
    const char *err =
      tt_sprintf("Number of arguments to function %.*s", yymsp[-4].minor.yy0.n, yymsp[-4].minor.yy0.z);
    diag_set(ClientError, ER_SQL_PARSER_LIMIT, err, yymsp[-1].minor.yy414->nExpr,
             SQL_MAX_FUNCTION_ARG);
    pParse->is_aborted = true;
  }
  yylhsminor.yy34.pExpr = sqlExprFunction(pParse, yymsp[-1].minor.yy414, &yymsp[-4].minor.yy0);
  spanSet(&yylhsminor.yy34,&yymsp[-4].minor.yy0,&yymsp[0].minor.yy0);
  if( yymsp[-2].minor.yy40==SF_Distinct && yylhsminor.yy34.pExpr ){
    yylhsminor.yy34.pExpr->flags |= EP_Distinct;
  }
}
#line 3501 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-4].minor.yy34 = yylhsminor.yy34;
        break;
      case 160: /* expr ::= ID|INDEXED LP STAR RP */
#line 1151 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy34.pExpr = sqlExprFunction(pParse, 0, &yymsp[-3].minor.yy0);
  spanSet(&yylhsminor.yy34,&yymsp[-3].minor.yy0,&yymsp[0].minor.yy0);
}
#line 3510 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-3].minor.yy34 = yylhsminor.yy34;
        break;
      case 161: /* expr ::= LP nexprlist COMMA expr RP */
#line 1186 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  ExprList *pList = sql_expr_list_append(yymsp[-3].minor.yy414, yymsp[-1].minor.yy34.pExpr);
  yylhsminor.yy34.pExpr = sqlPExpr(pParse, TK_VECTOR, 0, 0);
  if( yylhsminor.yy34.pExpr ){
    yylhsminor.yy34.pExpr->x.pList = pList;
    spanSet(&yylhsminor.yy34, &yymsp[-4].minor.yy0, &yymsp[0].minor.yy0);
  }else{
    sql_expr_list_delete(pList);
  }
}
#line 3525 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-4].minor.yy34 = yylhsminor.yy34;
        break;
      case 162: /* expr ::= expr AND expr */
      case 163: /* expr ::= expr OR expr */ yytestcase(yyruleno==163);
      case 164: /* expr ::= expr LT|GT|GE|LE expr */ yytestcase(yyruleno==164);
      case 165: /* expr ::= expr EQ|NE expr */ yytestcase(yyruleno==165);
      case 166: /* expr ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr */ yytestcase(yyruleno==166);
      case 167: /* expr ::= expr PLUS|MINUS expr */ yytestcase(yyruleno==167);
      case 168: /* expr ::= expr STAR|SLASH|REM expr */ yytestcase(yyruleno==168);
      case 169: /* expr ::= expr CONCAT expr */ yytestcase(yyruleno==169);
#line 1197 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanBinaryExpr(pParse,yymsp[-1].major,&yymsp[-2].minor.yy34,&yymsp[0].minor.yy34);}
#line 3538 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 170: /* likeop ::= LIKE_KW|MATCH */
#line 1210 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy0=yymsp[0].minor.yy0;/*A-overwrites-X*/}
#line 3543 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 171: /* likeop ::= NOT LIKE_KW|MATCH */
#line 1211 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-1].minor.yy0=yymsp[0].minor.yy0; yymsp[-1].minor.yy0.n|=0x80000000; /*yymsp[-1].minor.yy0-overwrite-yymsp[0].minor.yy0*/}
#line 3548 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 172: /* expr ::= expr likeop expr */
#line 1212 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  ExprList *pList;
  int bNot = yymsp[-1].minor.yy0.n & 0x80000000;
  yymsp[-1].minor.yy0.n &= 0x7fffffff;
  pList = sql_expr_list_append(NULL, yymsp[0].minor.yy34.pExpr);
  pList = sql_expr_list_append(pList, yymsp[-2].minor.yy34.pExpr);
  yymsp[-2].minor.yy34.pExpr = sqlExprFunction(pParse, pList, &yymsp[-1].minor.yy0);
  exprNot(pParse, bNot, &yymsp[-2].minor.yy34);
  yymsp[-2].minor.yy34.zEnd = yymsp[0].minor.yy34.zEnd;
}
#line 3562 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 173: /* expr ::= expr likeop expr ESCAPE expr */
#line 1222 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  ExprList *pList;
  int bNot = yymsp[-3].minor.yy0.n & 0x80000000;
  yymsp[-3].minor.yy0.n &= 0x7fffffff;
  pList = sql_expr_list_append(NULL, yymsp[-2].minor.yy34.pExpr);
  pList = sql_expr_list_append(pList, yymsp[-4].minor.yy34.pExpr);
  pList = sql_expr_list_append(pList, yymsp[0].minor.yy34.pExpr);
  yymsp[-4].minor.yy34.pExpr = sqlExprFunction(pParse, pList, &yymsp[-3].minor.yy0);
  exprNot(pParse, bNot, &yymsp[-4].minor.yy34);
  yymsp[-4].minor.yy34.zEnd = yymsp[0].minor.yy34.zEnd;
}
#line 3577 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 174: /* expr ::= expr IS NULL */
#line 1251 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanUnaryPostfix(pParse,TK_ISNULL,&yymsp[-2].minor.yy34,&yymsp[0].minor.yy0);}
#line 3582 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 175: /* expr ::= expr IS NOT NULL */
#line 1252 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanUnaryPostfix(pParse,TK_NOTNULL,&yymsp[-3].minor.yy34,&yymsp[0].minor.yy0);}
#line 3587 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 176: /* expr ::= NOT expr */
      case 177: /* expr ::= BITNOT expr */ yytestcase(yyruleno==177);
#line 1274 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanUnaryPrefix(&yymsp[-1].minor.yy34,pParse,yymsp[-1].major,&yymsp[0].minor.yy34,&yymsp[-1].minor.yy0);/*A-overwrites-B*/}
#line 3593 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 178: /* expr ::= MINUS expr */
#line 1278 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanUnaryPrefix(&yymsp[-1].minor.yy34,pParse,TK_UMINUS,&yymsp[0].minor.yy34,&yymsp[-1].minor.yy0);/*A-overwrites-B*/}
#line 3598 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 179: /* expr ::= PLUS expr */
#line 1280 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{spanUnaryPrefix(&yymsp[-1].minor.yy34,pParse,TK_UPLUS,&yymsp[0].minor.yy34,&yymsp[-1].minor.yy0);/*A-overwrites-B*/}
#line 3603 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 182: /* expr ::= expr between_op expr AND expr */
#line 1285 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  ExprList *pList = sql_expr_list_append(NULL, yymsp[-2].minor.yy34.pExpr);
  pList = sql_expr_list_append(pList, yymsp[0].minor.yy34.pExpr);
  yymsp[-4].minor.yy34.pExpr = sqlPExpr(pParse, TK_BETWEEN, yymsp[-4].minor.yy34.pExpr, 0);
  if( yymsp[-4].minor.yy34.pExpr ){
    yymsp[-4].minor.yy34.pExpr->x.pList = pList;
  }else{
    sql_expr_list_delete(pList);
  } 
  exprNot(pParse, yymsp[-3].minor.yy40, &yymsp[-4].minor.yy34);
  yymsp[-4].minor.yy34.zEnd = yymsp[0].minor.yy34.zEnd;
}
#line 3619 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 185: /* expr ::= expr in_op LP exprlist RP */
#line 1300 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  if( yymsp[-1].minor.yy414==0 ){
    /* Expressions of the form
    **
    **      expr1 IN ()
    **      expr1 NOT IN ()
    **
    ** simplify to constants 0 (false) and 1 (true), respectively,
    ** regardless of the value of expr1.
    */
    sql_expr_delete(yymsp[-4].minor.yy34.pExpr);
    int tk = yymsp[-3].minor.yy40 == 0 ? TK_FALSE : TK_TRUE;
    yymsp[-4].minor.yy34.pExpr = sql_expr_new_anon(tk);
    yymsp[-4].minor.yy34.pExpr->type = FIELD_TYPE_BOOLEAN;
  }else if( yymsp[-1].minor.yy414->nExpr==1 ){
    /* Expressions of the form:
    **
    **      expr1 IN (?1)
    **      expr1 NOT IN (?2)
    **
    ** with exactly one value on the RHS can be simplified to something
    ** like this:
    **
    **      expr1 == ?1
    **      expr1 <> ?2
    */
    Expr *pRHS = yymsp[-1].minor.yy414->a[0].pExpr;
    yymsp[-1].minor.yy414->a[0].pExpr = 0;
    sql_expr_list_delete(yymsp[-1].minor.yy414);
    yymsp[-4].minor.yy34.pExpr = sqlPExpr(pParse, yymsp[-3].minor.yy40 ? TK_NE : TK_EQ, yymsp[-4].minor.yy34.pExpr, pRHS);
  }else{
    yymsp[-4].minor.yy34.pExpr = sqlPExpr(pParse, TK_IN, yymsp[-4].minor.yy34.pExpr, 0);
    if( yymsp[-4].minor.yy34.pExpr ){
      yymsp[-4].minor.yy34.pExpr->x.pList = yymsp[-1].minor.yy414;
      sqlExprSetHeightAndFlags(pParse, yymsp[-4].minor.yy34.pExpr);
    }else{
      sql_expr_list_delete(yymsp[-1].minor.yy414);
    }
    exprNot(pParse, yymsp[-3].minor.yy40, &yymsp[-4].minor.yy34);
  }
  yymsp[-4].minor.yy34.zEnd = &yymsp[0].minor.yy0.z[yymsp[0].minor.yy0.n];
}
#line 3665 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 186: /* expr ::= LP select RP */
#line 1342 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  spanSet(&yymsp[-2].minor.yy34,&yymsp[-2].minor.yy0,&yymsp[0].minor.yy0); /*A-overwrites-B*/
  yymsp[-2].minor.yy34.pExpr = sqlPExpr(pParse, TK_SELECT, 0, 0);
  sqlPExprAddSelect(pParse, yymsp[-2].minor.yy34.pExpr, yymsp[-1].minor.yy247);
}
#line 3674 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 187: /* expr ::= expr in_op LP select RP */
#line 1347 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-4].minor.yy34.pExpr = sqlPExpr(pParse, TK_IN, yymsp[-4].minor.yy34.pExpr, 0);
  sqlPExprAddSelect(pParse, yymsp[-4].minor.yy34.pExpr, yymsp[-1].minor.yy247);
  exprNot(pParse, yymsp[-3].minor.yy40, &yymsp[-4].minor.yy34);
  yymsp[-4].minor.yy34.zEnd = &yymsp[0].minor.yy0.z[yymsp[0].minor.yy0.n];
}
#line 3684 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 188: /* expr ::= expr in_op nm paren_exprlist */
#line 1353 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct SrcList *pSrc = sql_src_list_append(NULL, &yymsp[-1].minor.yy0);
  Select *pSelect = sqlSelectNew(pParse, 0,pSrc,0,0,0,0,0,0,0);
  if(yymsp[0].minor.yy414 != NULL)
    sqlSrcListFuncArgs(pSelect != NULL ? pSrc : NULL, yymsp[0].minor.yy414);
  yymsp[-3].minor.yy34.pExpr = sqlPExpr(pParse, TK_IN, yymsp[-3].minor.yy34.pExpr, 0);
  sqlPExprAddSelect(pParse, yymsp[-3].minor.yy34.pExpr, pSelect);
  exprNot(pParse, yymsp[-2].minor.yy40, &yymsp[-3].minor.yy34);
  yymsp[-3].minor.yy34.zEnd = &yymsp[-1].minor.yy0.z[yymsp[-1].minor.yy0.n];
}
#line 3698 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 189: /* expr ::= EXISTS LP select RP */
#line 1363 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  Expr *p;
  spanSet(&yymsp[-3].minor.yy34,&yymsp[-3].minor.yy0,&yymsp[0].minor.yy0); /*A-overwrites-B*/
  p = yymsp[-3].minor.yy34.pExpr = sqlPExpr(pParse, TK_EXISTS, 0, 0);
  sqlPExprAddSelect(pParse, p, yymsp[-1].minor.yy247);
}
#line 3708 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 190: /* expr ::= CASE expr_optional case_exprlist case_else END */
#line 1371 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  spanSet(&yymsp[-4].minor.yy34,&yymsp[-4].minor.yy0,&yymsp[0].minor.yy0);  /*A-overwrites-C*/
  yymsp[-4].minor.yy34.pExpr = sqlPExpr(pParse, TK_CASE, yymsp[-3].minor.yy231, 0);
  if( yymsp[-4].minor.yy34.pExpr ){
    yymsp[-4].minor.yy34.pExpr->x.pList = yymsp[-1].minor.yy262 ? sql_expr_list_append(yymsp[-2].minor.yy414, yymsp[-1].minor.yy262) : yymsp[-2].minor.yy414;
    sqlExprSetHeightAndFlags(pParse, yymsp[-4].minor.yy34.pExpr);
  }else{
    sql_expr_list_delete(yymsp[-2].minor.yy414);
    sql_expr_delete(yymsp[-1].minor.yy262);
  }
}
#line 3723 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 191: /* case_exprlist ::= case_exprlist WHEN expr THEN expr */
#line 1384 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-4].minor.yy414 = sql_expr_list_append(yymsp[-4].minor.yy414, yymsp[-2].minor.yy34.pExpr);
  yymsp[-4].minor.yy414 = sql_expr_list_append(yymsp[-4].minor.yy414, yymsp[0].minor.yy34.pExpr);
}
#line 3731 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 192: /* case_exprlist ::= WHEN expr THEN expr */
#line 1388 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-3].minor.yy414 = sql_expr_list_append(NULL, yymsp[-2].minor.yy34.pExpr);
  yymsp[-3].minor.yy414 = sql_expr_list_append(yymsp[-3].minor.yy414, yymsp[0].minor.yy34.pExpr);
}
#line 3739 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 196: /* nexprlist ::= nexprlist COMMA expr */
#line 1405 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy414 = sql_expr_list_append(yymsp[-2].minor.yy414, yymsp[0].minor.yy34.pExpr);}
#line 3744 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 197: /* nexprlist ::= expr */
#line 1407 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy414 = sql_expr_list_append(NULL, yymsp[0].minor.yy34.pExpr); /*A-overwrites-Y*/}
#line 3749 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 199: /* paren_exprlist ::= LP exprlist RP */
      case 204: /* eidlist_opt ::= LP eidlist RP */ yytestcase(yyruleno==204);
#line 1414 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy414 = yymsp[-1].minor.yy414;}
#line 3755 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 200: /* cmd ::= createkw uniqueflag INDEX ifnotexists nm ON nm LP sortlist RP */
#line 1420 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct SrcList *src_list = sql_src_list_append(NULL ,&yymsp[-3].minor.yy0);
  create_index_def_init(&pParse->create_index_def, src_list, &yymsp[-5].minor.yy0, yymsp[-1].minor.yy414, yymsp[-8].minor.yy40,
                        SORT_ORDER_ASC, yymsp[-6].minor.yy40);
  pParse->initiateTTrans = true;
  sql_create_index(pParse);
}
#line 3766 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 201: /* uniqueflag ::= UNIQUE */
#line 1429 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = SQL_INDEX_TYPE_UNIQUE;}
#line 3771 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 202: /* uniqueflag ::= */
#line 1430 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy40 = SQL_INDEX_TYPE_NON_UNIQUE;}
#line 3776 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 205: /* eidlist ::= eidlist COMMA nm */
#line 1465 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-2].minor.yy414 = parserAddExprIdListTerm(pParse, yymsp[-2].minor.yy414, &yymsp[0].minor.yy0);
}
#line 3783 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 206: /* eidlist ::= nm */
#line 1468 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[0].minor.yy414 = parserAddExprIdListTerm(pParse, 0, &yymsp[0].minor.yy0); /*A-overwrites-Y*/
}
#line 3790 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 207: /* cmd ::= DROP INDEX ifexists nm ON fullname */
#line 1475 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  drop_index_def_init(&pParse->drop_index_def, yymsp[0].minor.yy499, &yymsp[-2].minor.yy0, yymsp[-3].minor.yy40);
  pParse->initiateTTrans = true;
  sql_drop_index(pParse);
}
#line 3799 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 208: /* cmd ::= SET SESSION nm EQ term */
#line 1483 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
    sql_setting_set(pParse,&yymsp[-2].minor.yy0,yymsp[0].minor.yy34.pExpr);
}
#line 3806 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 209: /* cmd ::= PRAGMA nm */
#line 1489 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
    sqlPragma(pParse,&yymsp[0].minor.yy0,0,0);
}
#line 3813 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 210: /* cmd ::= PRAGMA nm LP nm RP */
#line 1492 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
    sqlPragma(pParse,&yymsp[-3].minor.yy0,&yymsp[-1].minor.yy0,0);
}
#line 3820 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 211: /* cmd ::= PRAGMA nm LP nm DOT nm RP */
#line 1495 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
    sqlPragma(pParse,&yymsp[-5].minor.yy0,&yymsp[-3].minor.yy0,&yymsp[-1].minor.yy0);
}
#line 3827 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 212: /* cmd ::= FUNCTION_KW expr */
#line 1498 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  if (!pParse->is_expr) {
    diag_set(ClientError, ER_SQL_SYNTAX_NEAR_TOKEN, pParse->line_count, yymsp[-1].minor.yy0.n,
             yymsp[-1].minor.yy0.z);
    pParse->is_aborted = true;
    return;
  }
  pParse->parsed_ast_type = AST_TYPE_EXPR;
  pParse->parsed_ast.expr = yymsp[0].minor.yy34.pExpr;
}
#line 3841 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 213: /* cmd ::= SHOW CREATE TABLE nm */
#line 1510 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sql_emit_show_create_table_one(pParse, &yymsp[0].minor.yy0);
}
#line 3848 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 214: /* cmd ::= SHOW CREATE TABLE */
#line 1513 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  sql_emit_show_create_table_all(pParse);
}
#line 3855 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 215: /* cmd ::= createkw trigger_decl BEGIN trigger_cmd_list END */
#line 1519 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  Token all;
  all.z = yymsp[-3].minor.yy0.z;
  all.n = (int)(yymsp[0].minor.yy0.z - yymsp[-3].minor.yy0.z) + yymsp[0].minor.yy0.n;
  pParse->initiateTTrans = true;
  sql_trigger_finish(pParse, yymsp[-1].minor.yy115, &all);
}
#line 3866 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 216: /* trigger_decl ::= TRIGGER ifnotexists nm trigger_time trigger_event ON fullname foreach_clause when_clause */
#line 1529 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_trigger_def_init(&pParse->create_trigger_def, yymsp[-2].minor.yy499, &yymsp[-6].minor.yy0, yymsp[-5].minor.yy40, yymsp[-4].minor.yy122.a, yymsp[-4].minor.yy122.b, yymsp[0].minor.yy262,
                          yymsp[-7].minor.yy40);
  sql_trigger_begin(pParse);
  yymsp[-8].minor.yy0 = yymsp[-6].minor.yy0; /*yymsp[-8].minor.yy0-overwrites-T*/
}
#line 3876 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 217: /* trigger_time ::= BEFORE */
#line 1537 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy40 = TK_BEFORE; }
#line 3881 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 218: /* trigger_time ::= AFTER */
#line 1538 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy40 = TK_AFTER;  }
#line 3886 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 219: /* trigger_time ::= INSTEAD OF */
#line 1539 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[-1].minor.yy40 = TK_INSTEAD;}
#line 3891 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 220: /* trigger_time ::= */
#line 1540 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[1].minor.yy40 = TK_BEFORE; }
#line 3896 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 221: /* trigger_event ::= DELETE|INSERT */
      case 222: /* trigger_event ::= UPDATE */ yytestcase(yyruleno==222);
#line 1544 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy122.a = yymsp[0].major; /*A-overwrites-X*/ yymsp[0].minor.yy122.b = 0;}
#line 3902 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 223: /* trigger_event ::= UPDATE OF idlist */
#line 1546 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[-2].minor.yy122.a = TK_UPDATE; yymsp[-2].minor.yy122.b = yymsp[0].minor.yy240;}
#line 3907 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 224: /* foreach_clause ::= */
#line 1548 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  diag_set(ClientError, ER_SQL_PARSER_GENERIC_WITH_POS, pParse->line_count,
           pParse->line_pos, "FOR EACH STATEMENT triggers are not implemented, "
           "please supply FOR EACH ROW clause");
  pParse->is_aborted = true;
}
#line 3917 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 225: /* when_clause ::= */
#line 1558 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[1].minor.yy262 = 0; }
#line 3922 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 226: /* when_clause ::= WHEN expr */
#line 1559 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[-1].minor.yy262 = yymsp[0].minor.yy34.pExpr; }
#line 3927 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 227: /* trigger_cmd_list ::= trigger_cmd_list trigger_cmd SEMI */
#line 1563 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  assert( yymsp[-2].minor.yy115!=0 );
  yymsp[-2].minor.yy115->pLast->pNext = yymsp[-1].minor.yy115;
  yymsp[-2].minor.yy115->pLast = yymsp[-1].minor.yy115;
}
#line 3936 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 228: /* trigger_cmd_list ::= trigger_cmd SEMI */
#line 1568 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ 
  assert( yymsp[-1].minor.yy115!=0 );
  yymsp[-1].minor.yy115->pLast = yymsp[-1].minor.yy115;
}
#line 3944 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 229: /* trnm ::= nm DOT nm */
#line 1579 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-2].minor.yy0 = yymsp[0].minor.yy0;
  diag_set(ClientError, ER_SQL_PARSER_GENERIC_WITH_POS, pParse->line_count,
           pParse->line_pos, "qualified table names are not allowed on INSERT, "
           "UPDATE, and DELETE statements within triggers");
  pParse->is_aborted = true;
}
#line 3955 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 230: /* tridxby ::= INDEXED BY nm */
#line 1592 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  diag_set(ClientError, ER_SQL_SYNTAX_WITH_POS, pParse->line_count,
           pParse->line_pos, "the INDEXED BY clause is not allowed on UPDATE "\
           "or DELETE statements within triggers");
  pParse->is_aborted = true;
}
#line 3965 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 231: /* tridxby ::= NOT INDEXED */
#line 1598 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  diag_set(ClientError, ER_SQL_SYNTAX_WITH_POS, pParse->line_count,
           pParse->line_pos, "the NOT INDEXED BY clause is not allowed on "\
           "UPDATE or DELETE statements within triggers");
  pParse->is_aborted = true;
}
#line 3975 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 232: /* trigger_cmd ::= UPDATE orconf trnm tridxby SET setlist where_opt */
#line 1611 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
     yymsp[-6].minor.yy115 = sql_trigger_update_step(&yymsp[-4].minor.yy0, yymsp[-1].minor.yy414, yymsp[0].minor.yy262, yymsp[-5].minor.yy40);
     if (yymsp[-6].minor.yy115 == NULL) {
        pParse->is_aborted = true;
        return;
     }
   }
#line 3986 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 233: /* trigger_cmd ::= insert_cmd INTO trnm idlist_opt select */
#line 1620 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  /*yylhsminor.yy115-overwrites-yymsp[-4].minor.yy40. */
  yylhsminor.yy115 = sql_trigger_insert_step(&yymsp[-2].minor.yy0, yymsp[-1].minor.yy240, yymsp[0].minor.yy247, yymsp[-4].minor.yy40);
}
#line 3994 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-4].minor.yy115 = yylhsminor.yy115;
        break;
      case 234: /* trigger_cmd ::= DELETE FROM trnm tridxby where_opt */
#line 1626 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-4].minor.yy115 = sql_trigger_delete_step(&yymsp[-2].minor.yy0, yymsp[0].minor.yy262);
}
#line 4002 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 235: /* trigger_cmd ::= select */
#line 1631 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  /* yylhsminor.yy115-overwrites-yymsp[0].minor.yy247. */
  yylhsminor.yy115 = sql_trigger_select_step(yymsp[0].minor.yy247);
}
#line 4010 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy115 = yylhsminor.yy115;
        break;
      case 236: /* expr ::= RAISE LP IGNORE RP */
#line 1637 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  spanSet(&yymsp[-3].minor.yy34,&yymsp[-3].minor.yy0,&yymsp[0].minor.yy0);  /*A-overwrites-X*/
  yymsp[-3].minor.yy34.pExpr = sqlPExpr(pParse, TK_RAISE, 0, 0);
  if( yymsp[-3].minor.yy34.pExpr ){
    yymsp[-3].minor.yy34.pExpr->on_conflict_action = ON_CONFLICT_ACTION_IGNORE;
  }
}
#line 4022 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 237: /* expr ::= RAISE LP raisetype COMMA STRING RP */
#line 1644 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  spanSet(&yymsp[-5].minor.yy34,&yymsp[-5].minor.yy0,&yymsp[0].minor.yy0);  /*A-overwrites-X*/
  yymsp[-5].minor.yy34.pExpr = sql_expr_new_dequoted(TK_RAISE, &yymsp[-1].minor.yy0);
  yymsp[-5].minor.yy34.pExpr->on_conflict_action = (enum on_conflict_action) yymsp[-3].minor.yy40;
}
#line 4031 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 238: /* raisetype ::= ROLLBACK */
#line 1651 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = ON_CONFLICT_ACTION_ROLLBACK;}
#line 4036 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 239: /* raisetype ::= ABORT */
#line 1652 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = ON_CONFLICT_ACTION_ABORT;}
#line 4041 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 240: /* raisetype ::= FAIL */
#line 1653 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[0].minor.yy40 = ON_CONFLICT_ACTION_FAIL;}
#line 4046 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 241: /* cmd ::= DROP TRIGGER ifexists fullname */
#line 1657 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  struct Token t = Token_nil;
  drop_trigger_def_init(&pParse->drop_trigger_def, yymsp[0].minor.yy499, &t, yymsp[-1].minor.yy40);
  pParse->initiateTTrans = true;
  sql_drop_trigger(pParse);
}
#line 4056 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 242: /* alter_table_start ::= ALTER TABLE fullname */
#line 1674 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[-2].minor.yy140 = yymsp[0].minor.yy499; }
#line 4061 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 243: /* alter_add_constraint ::= alter_table_start ADD CONSTRAINT nm */
#line 1677 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
   yylhsminor.yy446.table_name = yymsp[-3].minor.yy140;
   yylhsminor.yy446.name = yymsp[0].minor.yy0;
   pParse->initiateTTrans = true;
 }
#line 4070 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-3].minor.yy446 = yylhsminor.yy446;
        break;
      case 244: /* alter_add_column ::= alter_table_start ADD column_name */
#line 1684 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yylhsminor.yy446.table_name = yymsp[-2].minor.yy140;
  yylhsminor.yy446.name = yymsp[0].minor.yy0;
  pParse->initiateTTrans = true;
}
#line 4080 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[-2].minor.yy446 = yylhsminor.yy446;
        break;
      case 245: /* column_name ::= COLUMN nm */
#line 1690 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[-1].minor.yy0 = yymsp[0].minor.yy0; }
#line 4086 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 246: /* column_name ::= nm */
#line 1691 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yylhsminor.yy0 = yymsp[0].minor.yy0; }
#line 4091 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 247: /* alter_column_def ::= alter_add_column typedef */
#line 1695 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_column_def_init(&pParse->create_column_def, yymsp[-1].minor.yy446.table_name, &yymsp[-1].minor.yy446.name, &yymsp[0].minor.yy226);
  create_ck_constraint_parse_def_init(&pParse->create_ck_constraint_parse_def);
  create_fk_constraint_parse_def_init(&pParse->create_fk_constraint_parse_def);
  sql_create_column_start(pParse);
}
#line 4102 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 248: /* cmd ::= alter_add_constraint FOREIGN KEY LP eidlist RP REFERENCES nm eidlist_opt */
#line 1703 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_fk_def_init(&pParse->create_fk_def, yymsp[-8].minor.yy446.table_name, &yymsp[-8].minor.yy446.name, yymsp[-4].minor.yy414, &yymsp[-1].minor.yy0, yymsp[0].minor.yy414);
  sql_create_foreign_key(pParse);
}
#line 4110 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 249: /* cmd ::= alter_add_constraint CHECK LP expr RP */
#line 1708 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
    create_ck_def_init(&pParse->create_ck_def, yymsp[-4].minor.yy446.table_name, &yymsp[-4].minor.yy446.name, &yymsp[-1].minor.yy34);
    sql_create_check_contraint(pParse, false);
}
#line 4118 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 250: /* cmd ::= alter_add_constraint unique_spec LP sortlist RP */
#line 1713 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  create_index_def_init(&pParse->create_index_def, yymsp[-4].minor.yy446.table_name, &yymsp[-4].minor.yy446.name, yymsp[-1].minor.yy414, yymsp[-3].minor.yy40,
                        SORT_ORDER_ASC, false);
  sql_create_index(pParse);
}
#line 4127 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 251: /* unique_spec ::= UNIQUE */
#line 1720 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy40 = SQL_INDEX_TYPE_CONSTRAINT_UNIQUE; }
#line 4132 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 252: /* unique_spec ::= PRIMARY KEY */
#line 1721 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[-1].minor.yy40 = SQL_INDEX_TYPE_CONSTRAINT_PK; }
#line 4137 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 253: /* cmd ::= alter_table_start RENAME TO nm */
#line 1723 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
    rename_entity_def_init(&pParse->rename_entity_def, yymsp[-3].minor.yy140, &yymsp[0].minor.yy0);
    pParse->initiateTTrans = true;
    sql_alter_table_rename(pParse);
}
#line 4146 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 254: /* cmd ::= ALTER TABLE nm DROP CONSTRAINT nm */
#line 1729 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_drop_table_constraint(pParse, &yymsp[-3].minor.yy0, &yymsp[0].minor.yy0);
}
#line 4154 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 255: /* cmd ::= ALTER TABLE nm DROP CONSTRAINT nm FOREIGN KEY */
#line 1734 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_drop_tuple_foreign_key(pParse, &yymsp[-5].minor.yy0, &yymsp[-2].minor.yy0);
}
#line 4162 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 256: /* cmd ::= ALTER TABLE nm DROP CONSTRAINT nm PRIMARY KEY */
#line 1739 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_drop_primary_key(pParse, &yymsp[-5].minor.yy0, &yymsp[-2].minor.yy0);
}
#line 4170 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 257: /* cmd ::= ALTER TABLE nm DROP CONSTRAINT nm UNIQUE */
#line 1744 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_drop_unique(pParse, &yymsp[-4].minor.yy0, &yymsp[-1].minor.yy0);
}
#line 4178 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 258: /* cmd ::= ALTER TABLE nm DROP CONSTRAINT nm CHECK */
#line 1749 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_drop_tuple_check(pParse, &yymsp[-4].minor.yy0, &yymsp[-1].minor.yy0);
}
#line 4186 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 259: /* cmd ::= ALTER TABLE nm DROP CONSTRAINT nm DOT nm */
#line 1754 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_drop_field_constraint(pParse, &yymsp[-5].minor.yy0, &yymsp[-2].minor.yy0, &yymsp[0].minor.yy0);
}
#line 4194 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 260: /* cmd ::= ALTER TABLE nm DROP CONSTRAINT nm DOT nm FOREIGN KEY */
#line 1759 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_drop_field_foreign_key(pParse, &yymsp[-7].minor.yy0, &yymsp[-4].minor.yy0, &yymsp[-2].minor.yy0);
}
#line 4202 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 261: /* cmd ::= ALTER TABLE nm DROP CONSTRAINT nm DOT nm CHECK */
#line 1764 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  pParse->initiateTTrans = true;
  sql_drop_field_check(pParse, &yymsp[-6].minor.yy0, &yymsp[-3].minor.yy0, &yymsp[-1].minor.yy0);
}
#line 4210 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 262: /* with ::= */
#line 1775 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{yymsp[1].minor.yy163 = 0;}
#line 4215 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 263: /* with ::= WITH wqlist */
#line 1776 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[-1].minor.yy163 = yymsp[0].minor.yy163; }
#line 4220 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 264: /* with ::= WITH RECURSIVE wqlist */
#line 1777 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[-2].minor.yy163 = yymsp[0].minor.yy163; }
#line 4225 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 265: /* wqlist ::= nm eidlist_opt AS LP select RP */
#line 1779 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-5].minor.yy163 = sqlWithAdd(pParse, 0, &yymsp[-5].minor.yy0, yymsp[-4].minor.yy414, yymsp[-1].minor.yy247); /*A-overwrites-X*/
}
#line 4232 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 266: /* wqlist ::= wqlist COMMA nm eidlist_opt AS LP select RP */
#line 1782 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-7].minor.yy163 = sqlWithAdd(pParse, yymsp[-7].minor.yy163, &yymsp[-5].minor.yy0, yymsp[-4].minor.yy414, yymsp[-1].minor.yy247);
}
#line 4239 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 267: /* typedef ::= TEXT */
      case 268: /* typedef ::= STRING_KW */ yytestcase(yyruleno==268);
#line 1788 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_STRING; }
#line 4245 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 269: /* typedef ::= SCALAR */
#line 1790 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_SCALAR; }
#line 4250 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 270: /* typedef ::= BOOL */
      case 271: /* typedef ::= BOOLEAN */ yytestcase(yyruleno==271);
#line 1792 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_BOOLEAN; }
#line 4256 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 272: /* typedef ::= VARBINARY */
#line 1794 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_VARBINARY; }
#line 4261 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 273: /* typedef ::= UUID */
#line 1795 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_UUID; }
#line 4266 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 274: /* typedef ::= ANY */
#line 1796 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_ANY; }
#line 4271 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 275: /* typedef ::= ARRAY */
#line 1797 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_ARRAY; }
#line 4276 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 276: /* typedef ::= MAP */
#line 1798 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_MAP; }
#line 4281 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 277: /* typedef ::= DATETIME */
#line 1799 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_DATETIME; }
#line 4286 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 278: /* typedef ::= INTERVAL */
#line 1800 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_INTERVAL; }
#line 4291 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 279: /* char_len ::= LP INTEGER RP */
#line 1812 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  (void) yymsp[-2].minor.yy40;
  (void) yymsp[-1].minor.yy0;
}
#line 4299 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 280: /* typedef ::= VARCHAR char_len */
#line 1818 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{
  yymsp[-1].minor.yy226.type = FIELD_TYPE_STRING;
  (void) yymsp[0].minor.yy40;
}
#line 4307 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 281: /* number_typedef ::= NUMBER */
#line 1825 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_NUMBER; }
#line 4312 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 282: /* number_typedef ::= DOUBLE */
#line 1826 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_DOUBLE; }
#line 4317 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 283: /* number_typedef ::= INT|INTEGER_KW */
#line 1827 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_INTEGER; }
#line 4322 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 284: /* number_typedef ::= UNSIGNED */
#line 1828 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_UNSIGNED; }
#line 4327 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      case 285: /* number_typedef ::= DECIMAL */
#line 1829 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"
{ yymsp[0].minor.yy226.type = FIELD_TYPE_DECIMAL; }
#line 4332 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
        break;
      default:
      /* (286) input ::= ecmd */ yytestcase(yyruleno==286);
      /* (287) ecmd ::= explain cmdx SEMI */ yytestcase(yyruleno==287);
      /* (288) explain ::= */ yytestcase(yyruleno==288);
      /* (289) cmdx ::= cmd (OPTIMIZED OUT) */ assert(yyruleno!=289);
      /* (290) savepoint_opt ::= SAVEPOINT */ yytestcase(yyruleno==290);
      /* (291) savepoint_opt ::= */ yytestcase(yyruleno==291);
      /* (292) cmd ::= create_table create_table_args with_opts create_table_end */ yytestcase(yyruleno==292);
      /* (293) create_table_args ::= LP columnlist RP */ yytestcase(yyruleno==293);
      /* (294) with_opts ::= WITH engine_opts */ yytestcase(yyruleno==294);
      /* (295) with_opts ::= */ yytestcase(yyruleno==295);
      /* (296) columnlist ::= columnlist COMMA tcons */ yytestcase(yyruleno==296);
      /* (297) columnlist ::= columnlist COMMA column_def create_column_end */ yytestcase(yyruleno==297);
      /* (298) columnlist ::= column_def create_column_end */ yytestcase(yyruleno==298);
      /* (299) column_def ::= column_name_and_type carglist */ yytestcase(yyruleno==299);
      /* (300) columnlist ::= tcons (OPTIMIZED OUT) */ assert(yyruleno!=300);
      /* (301) carglist ::= carglist ccons */ yytestcase(yyruleno==301);
      /* (302) carglist ::= */ yytestcase(yyruleno==302);
      /* (303) resolvetype ::= raisetype (OPTIMIZED OUT) */ assert(yyruleno!=303);
      /* (304) selectnowith ::= oneselect (OPTIMIZED OUT) */ assert(yyruleno!=304);
      /* (305) oneselect ::= values */ yytestcase(yyruleno==305);
      /* (306) sclp ::= selcollist COMMA */ yytestcase(yyruleno==306);
      /* (307) as ::= ID|STRING */ yytestcase(yyruleno==307);
      /* (308) join_nm ::= ID|INDEXED */ yytestcase(yyruleno==308);
      /* (309) join_nm ::= JOIN_KW */ yytestcase(yyruleno==309);
      /* (310) expr ::= term (OPTIMIZED OUT) */ assert(yyruleno!=310);
      /* (311) term ::= number (OPTIMIZED OUT) */ assert(yyruleno!=311);
      /* (312) maplist ::= nmaplist */ yytestcase(yyruleno==312);
      /* (313) type_func ::= CHAR */ yytestcase(yyruleno==313);
      /* (314) exprlist ::= nexprlist */ yytestcase(yyruleno==314);
      /* (315) foreach_clause ::= FOR EACH ROW */ yytestcase(yyruleno==315);
      /* (316) trnm ::= nm */ yytestcase(yyruleno==316);
      /* (317) tridxby ::= */ yytestcase(yyruleno==317);
      /* (318) cmd ::= alter_column_def carglist create_column_end */ yytestcase(yyruleno==318);
      /* (319) typedef ::= number_typedef (OPTIMIZED OUT) */ assert(yyruleno!=319);
        break;
/********** End reduce actions ************************************************/
  };
  assert( yyruleno<sizeof(yyRuleInfo)/sizeof(yyRuleInfo[0]) );
  yygoto = yyRuleInfo[yyruleno].lhs;
  yysize = yyRuleInfo[yyruleno].nrhs;
  yyact = yy_find_reduce_action(yymsp[-yysize].stateno,(YYCODETYPE)yygoto);
  if( yyact <= YY_MAX_SHIFTREDUCE ){
    if( yyact>YY_MAX_SHIFT ){
      yyact += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
    }
    yymsp -= yysize-1;
    yypParser->yytos = yymsp;
    yymsp->stateno = (YYACTIONTYPE)yyact;
    yymsp->major = (YYCODETYPE)yygoto;
    yyTraceShift(yypParser, yyact);
  }else{
    assert( yyact == YY_ACCEPT_ACTION );
    yypParser->yytos -= yysize;
    yy_accept(yypParser);
  }
}

/*
** The following code executes when the parse fails
*/
#ifndef YYNOERRORRECOVERY
static void yy_parse_failed(
  yyParser *yypParser           /* The parser */
){
  sqlParserARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sFail!\n",yyTracePrompt);
  }
#endif
  while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser fails */
/************ Begin %parse_failure code ***************************************/
/************ End %parse_failure code *****************************************/
  sqlParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}
#endif /* YYNOERRORRECOVERY */

/*
** The following code executes when a syntax error first occurs.
*/
static void yy_syntax_error(
  yyParser *yypParser,           /* The parser */
  int yymajor,                   /* The major type of the error token */
  sqlParserTOKENTYPE yyminor         /* The minor type of the error token */
){
  sqlParserARG_FETCH;
#define TOKEN yyminor
/************ Begin %syntax_error code ****************************************/
#line 32 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.y"

  UNUSED_PARAMETER(yymajor);  /* Silence some compiler warnings */
  assert( TOKEN.z[0] );  /* The tokenizer always gives us a token */
  if (yypParser->is_fallback_failed && TOKEN.isReserved) {
    diag_set(ClientError, ER_SQL_KEYWORD_IS_RESERVED, pParse->line_count,
             pParse->line_pos, TOKEN.n, TOKEN.z, TOKEN.n, TOKEN.z);
  } else {
    diag_set(ClientError, ER_SQL_SYNTAX_NEAR_TOKEN, pParse->line_count, TOKEN.n,
             TOKEN.z);
  }
  pParse->is_aborted = true;
#line 4437 "/usr/src/RPM/BUILD/tarantool-3.0.1/src/box/sql/parse.c"
/************ End %syntax_error code ******************************************/
  sqlParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/*
** The following is executed when the parser accepts
*/
static void yy_accept(
  yyParser *yypParser           /* The parser */
){
  sqlParserARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sAccept!\n",yyTracePrompt);
  }
#endif
#ifndef YYNOERRORRECOVERY
  yypParser->yyerrcnt = -1;
#endif
  assert( yypParser->yytos==yypParser->yystack );
  /* Here code is inserted which will be executed whenever the
  ** parser accepts */
/*********** Begin %parse_accept code *****************************************/
/*********** End %parse_accept code *******************************************/
  sqlParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/* The main parser program.
** The first argument is a pointer to a structure obtained from
** "sqlParserAlloc" which describes the current state of the parser.
** The second argument is the major token number.  The third is
** the minor token.  The fourth optional argument is whatever the
** user wants (and specified in the grammar) and is available for
** use by the action routines.
**
** Inputs:
** <ul>
** <li> A pointer to the parser (an opaque structure.)
** <li> The major token number.
** <li> The minor token number.
** <li> An option argument of a grammar-specified type.
** </ul>
**
** Outputs:
** None.
*/
void sqlParser(
  void *yyp,                   /* The parser */
  int yymajor,                 /* The major token code number */
  sqlParserTOKENTYPE yyminor       /* The value for the token */
  sqlParserARG_PDECL               /* Optional %extra_argument parameter */
){
  YYMINORTYPE yyminorunion;
  unsigned int yyact;   /* The parser action. */
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  int yyendofinput;     /* True if we are at the end of input */
#endif
#ifdef YYERRORSYMBOL
  int yyerrorhit = 0;   /* True if yymajor has invoked an error */
#endif
  yyParser *yypParser;  /* The parser */

  yypParser = (yyParser*)yyp;
  assert( yypParser->yytos!=0 );
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  yyendofinput = (yymajor==0);
#endif
  sqlParserARG_STORE;

#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sInput '%s'\n",yyTracePrompt,yyTokenName[yymajor]);
  }
#endif

  do{
    yyact = yy_find_shift_action(yypParser,(YYCODETYPE)yymajor);
    if( yyact <= YY_MAX_SHIFTREDUCE ){
      yy_shift(yypParser,yyact,yymajor,yyminor);
#ifndef YYNOERRORRECOVERY
      yypParser->yyerrcnt--;
#endif
      yymajor = YYNOCODE;
    }else if( yyact <= YY_MAX_REDUCE ){
      yy_reduce(yypParser,yyact-YY_MIN_REDUCE);
    }else{
      assert( yyact == YY_ERROR_ACTION );
      yyminorunion.yy0 = yyminor;
#ifdef YYERRORSYMBOL
      int yymx;
#endif
#ifndef NDEBUG
      if( yyTraceFILE ){
        fprintf(yyTraceFILE,"%sSyntax Error!\n",yyTracePrompt);
      }
#endif
#ifdef YYERRORSYMBOL
      /* A syntax error has occurred.
      ** The response to an error depends upon whether or not the
      ** grammar defines an error token "ERROR".  
      **
      ** This is what we do if the grammar does define ERROR:
      **
      **  * Call the %syntax_error function.
      **
      **  * Begin popping the stack until we enter a state where
      **    it is legal to shift the error symbol, then shift
      **    the error symbol.
      **
      **  * Set the error count to three.
      **
      **  * Begin accepting and shifting new tokens.  No new error
      **    processing will occur until three tokens have been
      **    shifted successfully.
      **
      */
      if( yypParser->yyerrcnt<0 ){
        yy_syntax_error(yypParser,yymajor,yyminor);
      }
      yymx = yypParser->yytos->major;
      if( yymx==YYERRORSYMBOL || yyerrorhit ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE,"%sDiscard input token %s\n",
             yyTracePrompt,yyTokenName[yymajor]);
        }
#endif
        yy_destructor((YYCODETYPE)yymajor, &yyminorunion);
        yymajor = YYNOCODE;
      }else{
        while( yypParser->yytos >= yypParser->yystack
            && yymx != YYERRORSYMBOL
            && (yyact = yy_find_reduce_action(
                        yypParser->yytos->stateno,
                        YYERRORSYMBOL)) >= YY_MIN_REDUCE
        ){
          yy_pop_parser_stack(yypParser);
        }
        if( yypParser->yytos < yypParser->yystack || yymajor==0 ){
          yy_destructor((YYCODETYPE)yymajor, &yyminorunion);
          yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
          yypParser->yyerrcnt = -1;
#endif
          yymajor = YYNOCODE;
        }else if( yymx!=YYERRORSYMBOL ){
          yy_shift(yypParser,yyact,YYERRORSYMBOL,yyminor);
        }
      }
      yypParser->yyerrcnt = 3;
      yyerrorhit = 1;
#elif defined(YYNOERRORRECOVERY)
      /* If the YYNOERRORRECOVERY macro is defined, then do not attempt to
      ** do any kind of error recovery.  Instead, simply invoke the syntax
      ** error routine and continue going as if nothing had happened.
      **
      ** Applications can set this macro (for example inside %include) if
      ** they intend to abandon the parse upon the first syntax error seen.
      */
      yy_syntax_error(yypParser,yymajor, yyminor);
      yy_destructor((YYCODETYPE)yymajor, &yyminorunion);
      yymajor = YYNOCODE;
      
#else  /* YYERRORSYMBOL is not defined */
      /* This is what we do if the grammar does not define ERROR:
      **
      **  * Report an error message, and throw away the input token.
      **
      **  * If the input token is $, then fail the parse.
      **
      ** As before, subsequent error messages are suppressed until
      ** three input tokens have been successfully shifted.
      */
      if( yypParser->yyerrcnt<=0 ){
        yy_syntax_error(yypParser,yymajor, yyminor);
      }
      yypParser->yyerrcnt = 3;
      yy_destructor((YYCODETYPE)yymajor, &yyminorunion);
      if( yyendofinput ){
        yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
        yypParser->yyerrcnt = -1;
#endif
      }
      yymajor = YYNOCODE;
#endif
    }
  }while( yymajor!=YYNOCODE && yypParser->yytos>yypParser->yystack
         PARSER_ERROR_CHECK);
#ifndef NDEBUG
  if( yyTraceFILE ){
    yyStackEntry *i;
    char cDiv = '[';
    fprintf(yyTraceFILE,"%sReturn. Stack=",yyTracePrompt);
    for(i=&yypParser->yystack[1]; i<=yypParser->yytos; i++){
      fprintf(yyTraceFILE,"%c%s", cDiv, yyTokenName[i->major]);
      cDiv = ' ';
    }
    fprintf(yyTraceFILE,"]\n");
  }
#endif
  return;
}
